
#line 1 "./zscan.rl"
/* Copyright © 2012 Brandon L Black <blblack@gmail.com> and Jay Reitz <jreitz@gmail.com>
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "zscan.h"

#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "conf.h"
#include "ltree.h"
#include "ltarena.h"
#include "gdnsd-misc.h"

#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif

#define parse_error(_fmt, ...) \
    log_fatal("Zonefile parse error at line %u of %s: " _fmt,z->lcount,z->curfn,__VA_ARGS__)

#define parse_error_noargs(_fmt) \
    log_fatal("Zonefile parse error at line %u of %s: " _fmt,z->lcount,z->curfn)

typedef struct {
    uint8_t  ipv6[16];
    uint32_t ipv4;
    bool     in_paren;
    bool     zn_err_detect;
    bool     lhs_is_ooz;
    unsigned n_subzones;
    unsigned lcount;
    unsigned num_texts;
    unsigned def_ttl;
    unsigned uval;
    unsigned ttl;
    unsigned uv_1;
    unsigned uv_2;
    unsigned uv_3;
    unsigned uv_4;
    unsigned uv_5;
    unsigned rfc3597_data_len;
    unsigned rfc3597_data_written;
    unsigned limit_v4;
    unsigned limit_v6;
    uint8_t* rfc3597_data;
    const uint8_t** subzones;
    const char* zones_dir;
    const char* curfn;
    const char* tstart;
    uint8_t* include_filename;
    uint8_t  zroot[256];
    uint8_t  origin[256];
    uint8_t  lhs_dname[256];
    uint8_t  rhs_dname[256];
    uint8_t  eml_dname[256];
    uint8_t** texts;
} zscan_t;

F_NONNULL
static void scanner(zscan_t* z, int fd);

/******** IP Addresses ********/

F_NONNULL
static void set_ipv4(zscan_t* z, const char* end) {
    dmn_assert(z); dmn_assert(end);
    char txt[16];
    unsigned len = end - z->tstart;
    memcpy(txt, z->tstart, len);
    txt[len] = 0;
    z->tstart = NULL;
    struct in_addr addr;
    int status = inet_pton(AF_INET, txt, &addr);
    if(status > 0)
        z->ipv4 = addr.s_addr;
    else
        parse_error("IPv4 address '%s' invalid", txt);
}

F_NONNULL
static void set_ipv6(zscan_t* z, const char* end) {
    dmn_assert(z);
    dmn_assert(end);
    char txt[INET6_ADDRSTRLEN + 1];
    unsigned len = end - z->tstart;
    memcpy(txt, z->tstart, len);
    txt[len] = 0;
    z->tstart = NULL;
    struct in6_addr* v6a = malloc(sizeof(struct in6_addr));
    int status = inet_pton(AF_INET6, txt, v6a);
    if(status > 0)
        memcpy(z->ipv6, v6a->s6_addr, 16);
    else
        parse_error("IPv6 address '%s' invalid", txt);

    free(v6a);
}

F_NONNULL
static void validate_dname_in_zone(const zscan_t* z, const uint8_t* dname) {
    dmn_assert(z); dmn_assert(z->zroot); dmn_assert(dname);
    if(!dname_isinzone(z->zroot, dname))
        parse_error("domainname %s is not within this zonefile's zone (%s)", logf_dname(dname), logf_dname(z->zroot));
}

F_NONNULL
static void validate_lhs_not_ooz(const zscan_t* z) {
    dmn_assert(z);
    if(z->lhs_is_ooz)
        parse_error("domainname %s is not within this zonefile's zone (%s)", logf_dname(z->lhs_dname), logf_dname(z->zroot));
}

F_NONNULL
static void dname_set(zscan_t* z, uint8_t* dname, unsigned len, bool lhs) {
    dmn_assert(z); dmn_assert(dname); dmn_assert(z->zroot); dmn_assert(z->origin);
    dname_status_t catstat;
    dname_status_t status = dname_from_string(dname, (const uint8_t*)z->tstart, len);
    switch(status) {
        case DNAME_INVALID:
            parse_error_noargs("unparseable domainname");
            break;
        case DNAME_VALID:
            if(lhs) z->lhs_is_ooz = !dname_isinzone(z->zroot, dname);
            break;
        case DNAME_PARTIAL:
            if(lhs) z->lhs_is_ooz = false;
            catstat = dname_cat(dname, z->origin);
            if(catstat == DNAME_INVALID)
                parse_error_noargs("illegal domainname");
            break;
    }
}

F_NONNULL
static zscan_t* zscan_init(zscan_t* z, const char* zones_dir, const uint8_t* zname, const uint8_t* origin, const char* fn, const unsigned def_ttl_arg, const unsigned limit_v4, const unsigned limit_v6, const unsigned n_subzones, const uint8_t** subzones) {
    dmn_assert(z); dmn_assert(zname); dmn_assert(origin); dmn_assert(fn);
    memset(z, 0, sizeof(zscan_t));
    z->lcount = 1;
    z->def_ttl = def_ttl_arg;
    z->curfn = fn;
    z->n_subzones = n_subzones;
    z->subzones = subzones;
    z->zones_dir = zones_dir;
    z->limit_v4 = limit_v4;
    z->limit_v6 = limit_v6;
    dname_copy(z->zroot, zname);
    dname_copy(z->origin, origin);
    dname_copy(z->lhs_dname, origin);

    return z;
}

/********** TXT/SPF ******************/

F_NONNULL
static void text_start(zscan_t* z) {
    dmn_assert(z);
    z->num_texts = 0;
    z->texts = NULL;
}

// Only if we didn't add_rec, as normally
//  add_rec consumes the allocation.
F_NONNULL
static void texts_free(zscan_t* z) {
    for(unsigned i = 0; i < z->num_texts; i++)
        free(z->texts[i]);
}

F_NONNULL
static void text_add_tok(zscan_t* z, const unsigned len, const bool big_ok) {
    dmn_assert(z);

    uint8_t text_temp[len + 1];

    const unsigned newlen = dns_unescape(text_temp, (const uint8_t*)z->tstart, len);
    dmn_assert(newlen <= len);

    if(newlen > 255) {
        if(!big_ok || gconfig.disable_text_autosplit)
            parse_error_noargs("Text chunk too long (>255 unescaped)");
        if(newlen > 65500) parse_error_noargs("Text chunk too long (>65500 unescaped)");
        unsigned remainder = newlen % 255;
        unsigned num_whole_chunks = (newlen - remainder) / 255;
        const uint8_t* zptr = text_temp;
        const unsigned new_alloc = 1 + z->num_texts + num_whole_chunks + (remainder ? 1 : 0);
        z->texts = realloc(z->texts, new_alloc * sizeof(uint8_t*));
        for(unsigned i = 0; i < num_whole_chunks; i++) {
            uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(256);
            *chunk++ = 255;
            memcpy(chunk, zptr, 255);
            zptr += 255;
        }
        if(remainder) {
            uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(remainder + 1);
            *chunk++ = remainder;
            memcpy(chunk, zptr, remainder);
        }
        z->texts[z->num_texts] = NULL;
    }
    else {
        z->texts = realloc(z->texts, (z->num_texts + 2) * sizeof(uint8_t*));
        uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(newlen + 1);
        *chunk++ = newlen;
        memcpy(chunk, text_temp, newlen);
        z->texts[z->num_texts] = NULL;
    }

    z->tstart = NULL;
}

F_NONNULL
static void set_filename(zscan_t* z, const unsigned len) {
    dmn_assert(z);
    uint8_t* fn = malloc(len + 1);
    const unsigned newlen = dns_unescape(fn, (const uint8_t*)z->tstart, len);
    dmn_assert(newlen <= len);
    z->include_filename = fn = realloc(fn, newlen + 1);
    fn[newlen] = 0;
    z->tstart = NULL;
}

// Input must have two bytes of text constrained to [0-9A-Fa-f]
F_NONNULL
static unsigned hexbyte(const char* intxt) {
    dmn_assert(intxt);
    dmn_assert(
        (intxt[0] >= '0' && intxt[0] <= '9')
        || (intxt[0] >= 'A' && intxt[0] <= 'F')
        || (intxt[0] >= 'a' && intxt[0] <= 'f')
    );
    dmn_assert(
        (intxt[1] >= '0' && intxt[1] <= '9')
        || (intxt[1] >= 'A' && intxt[1] <= 'F')
        || (intxt[1] >= 'a' && intxt[1] <= 'f')
    );

    unsigned out;

    if(intxt[0] <= '9')
        out = (intxt[0] - '0') << 4;
    else
        out = ((intxt[0] | 0x20) - ('a' - 10)) << 4;

    if(intxt[1] <= '9')
        out |= (intxt[1] - '0');
    else
        out |= ((intxt[1] | 0x20) - ('a' - 10));

    return out;
}

F_NONNULL
static bool lhs_subzones_ok(zscan_t* z) {
    for(unsigned i = 0; i < z->n_subzones; i++) {
        if(dname_isinzone(z->subzones[i], z->lhs_dname)) {
            log_warn("Ignoring RRs for name '%s' in zone '%s' because they overlap subzone '%s'",
                logf_dname(z->lhs_dname), logf_dname(z->zroot), logf_dname(z->subzones[i]));
            return false;
        }
    }

    return true;
}

F_NONNULL
static void process_include(zscan_t* z) {
    dmn_assert(z);
    dmn_assert(z->include_filename);
    char* zfn = gdnsd_make_abs_fn(z->zones_dir, (char*)z->include_filename);
    free(z->include_filename);
    z->include_filename = NULL;
    validate_dname_in_zone(z, z->rhs_dname);
    zscan_t* znew = malloc(sizeof(zscan_t));
    zscan_init(znew, z->zones_dir, z->zroot, z->rhs_dname, zfn, z->def_ttl, z->limit_v4, z->limit_v6, z->n_subzones, z->subzones);
    int newfd = open(zfn, O_RDONLY);
    if(newfd < 0)
        parse_error("Cannot open $INCLUDE file '%s' for reading: %s", zfn, logf_errno());
    scanner(znew, newfd);
    if(close(newfd))
        parse_error("Cannot close $INCLUDE file '%s': %s", zfn, logf_errno());
    free(zfn);
    free(znew);
}

F_NONNULL
static void mult_uval(zscan_t* z, int fc) {
    dmn_assert(z);
    fc |= 0x20;
    switch(fc) {
        case 'm': z->uval *= 60; break;
        case 'h': z->uval *= 3600; break;
        case 'd': z->uval *= 86400; break;
        case 'w': z->uval *= 604800; break;
    }
}

F_NONNULL
static void set_dyna(zscan_t* z, const char* fpc) {
    dmn_assert(z);
    unsigned dlen = fpc - z->tstart;
    if(dlen > 255)
        parse_error_noargs("DYNA/DYNC plugin!resource string cannot exceed 255 chars");
    memcpy(z->eml_dname, z->tstart, dlen);
    z->eml_dname[dlen] = 0;
    z->tstart = NULL;
}

F_NONNULL
static void rec_soa(zscan_t* z) {
    dmn_assert(z);
    if(dname_cmp(z->lhs_dname, z->zroot))
        parse_error_noargs("SOA record can only be defined for the root of the zone");
    ltree_add_rec_soa(z->lhs_dname, z->rhs_dname, z->eml_dname, z->ttl, z->uv_1, z->uv_2, z->uv_3, z->uv_4, z->uv_5);
}

F_NONNULL
static void rec_a(zscan_t* z) {
    dmn_assert(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_a(z->lhs_dname, z->ipv4, z->ttl, z->limit_v4, z->lhs_is_ooz ? z->zroot : NULL);
}

F_NONNULL
static void rec_aaaa(zscan_t* z) {
    dmn_assert(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_aaaa(z->lhs_dname, z->ipv6, z->ttl, z->limit_v6, z->lhs_is_ooz ? z->zroot : NULL);
}

F_NONNULL
static void rec_ns(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_ns(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_cname(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_cname(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_ptr(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_ptr(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_mx(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_mx(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1);
}

F_NONNULL
static void rec_srv(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_srv(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1, z->uv_2, z->uv_3);
}

F_NONNULL
static void rec_naptr(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_naptr(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1, z->uv_2, z->num_texts, z->texts);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_txt(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_txt(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_spf(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_spf(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_spftxt(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_spftxt(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_dyna(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_dynaddr(z->lhs_dname, z->eml_dname, z->ttl, z->limit_v4, z->limit_v6);
}

F_NONNULL
static void rec_dyncname(zscan_t* z) {
    dmn_assert(z);

    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z)) {
        ltree_add_rec_dyncname(z->lhs_dname, z->eml_dname, z->origin, z->ttl);
    }
}

F_NONNULL
static void rec_rfc3597(zscan_t* z) {
    dmn_assert(z);
    if(z->rfc3597_data_written < z->rfc3597_data_len)
        parse_error("RFC3597 generic RR claimed rdata length of %u, but only %u bytes of data present", z->rfc3597_data_len, z->rfc3597_data_written);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_rfc3597(z->lhs_dname, z->uv_1, z->ttl, z->rfc3597_data_len, z->rfc3597_data);
    else
        free(z->rfc3597_data);
}


#line 467 "./zscan_lm.c"
static const char _zone_actions[] = {
	0, 1, 0, 1, 1, 1, 2, 1, 
	3, 1, 4, 1, 5, 1, 6, 1, 
	7, 1, 8, 1, 10, 1, 11, 1, 
	13, 1, 14, 1, 15, 1, 16, 1, 
	17, 1, 22, 1, 23, 1, 24, 1, 
	25, 1, 26, 1, 27, 1, 28, 1, 
	42, 1, 43, 1, 44, 1, 49, 1, 
	50, 1, 51, 1, 52, 1, 53, 2, 
	1, 49, 2, 1, 50, 2, 1, 51, 
	2, 2, 49, 2, 2, 50, 2, 2, 
	51, 2, 3, 49, 2, 3, 50, 2, 
	3, 51, 2, 4, 9, 2, 4, 12, 
	2, 4, 36, 2, 4, 37, 2, 4, 
	38, 2, 4, 39, 2, 4, 40, 2, 
	4, 41, 2, 4, 49, 2, 4, 50, 
	2, 4, 51, 2, 5, 9, 2, 5, 
	12, 2, 5, 36, 2, 5, 37, 2, 
	5, 38, 2, 5, 39, 2, 5, 40, 
	2, 5, 41, 2, 5, 49, 2, 5, 
	50, 2, 5, 51, 2, 6, 9, 2, 
	6, 12, 2, 6, 36, 2, 6, 37, 
	2, 6, 38, 2, 6, 39, 2, 6, 
	40, 2, 6, 41, 2, 6, 49, 2, 
	6, 50, 2, 6, 51, 2, 7, 49, 
	2, 7, 50, 2, 7, 51, 2, 8, 
	49, 2, 8, 50, 2, 8, 51, 2, 
	10, 12, 2, 10, 49, 2, 10, 50, 
	2, 10, 51, 2, 11, 12, 2, 11, 
	49, 2, 11, 50, 2, 11, 51, 2, 
	13, 0, 2, 14, 0, 2, 14, 42, 
	2, 14, 43, 2, 14, 44, 2, 14, 
	49, 2, 14, 50, 2, 14, 51, 2, 
	15, 0, 2, 15, 42, 2, 15, 43, 
	2, 15, 44, 2, 15, 49, 2, 15, 
	50, 2, 15, 51, 2, 16, 49, 2, 
	16, 50, 2, 16, 51, 2, 17, 49, 
	2, 17, 50, 2, 17, 51, 2, 18, 
	34, 2, 19, 35, 2, 20, 21, 2, 
	20, 22, 2, 20, 23, 2, 20, 25, 
	2, 20, 26, 2, 20, 27, 2, 20, 
	28, 2, 20, 30, 2, 20, 31, 2, 
	20, 48, 2, 22, 49, 2, 22, 50, 
	2, 22, 51, 2, 23, 49, 2, 23, 
	50, 2, 23, 51, 2, 24, 0, 2, 
	25, 49, 2, 25, 50, 2, 25, 51, 
	2, 26, 49, 2, 26, 50, 2, 26, 
	51, 2, 27, 49, 2, 27, 50, 2, 
	27, 51, 2, 28, 49, 2, 28, 50, 
	2, 28, 51, 2, 29, 33, 2, 32, 
	45, 2, 32, 46, 2, 42, 51, 2, 
	43, 51, 2, 44, 51, 2, 49, 13, 
	2, 50, 13, 2, 51, 13, 2, 52, 
	2, 2, 52, 5, 2, 52, 7, 2, 
	52, 10, 2, 52, 14, 2, 52, 16, 
	2, 52, 51, 2, 53, 0, 2, 53, 
	47, 2, 53, 49, 2, 53, 50, 2, 
	53, 51, 3, 4, 9, 49, 3, 4, 
	9, 50, 3, 4, 9, 51, 3, 4, 
	12, 49, 3, 4, 12, 50, 3, 4, 
	12, 51, 3, 4, 36, 49, 3, 4, 
	36, 50, 3, 4, 36, 51, 3, 4, 
	37, 49, 3, 4, 37, 50, 3, 4, 
	37, 51, 3, 4, 38, 49, 3, 4, 
	38, 50, 3, 4, 38, 51, 3, 4, 
	39, 49, 3, 4, 39, 50, 3, 4, 
	39, 51, 3, 4, 40, 49, 3, 4, 
	40, 50, 3, 4, 40, 51, 3, 4, 
	41, 49, 3, 4, 41, 50, 3, 4, 
	41, 51, 3, 5, 9, 49, 3, 5, 
	9, 50, 3, 5, 9, 51, 3, 5, 
	12, 49, 3, 5, 12, 50, 3, 5, 
	12, 51, 3, 5, 36, 49, 3, 5, 
	36, 50, 3, 5, 36, 51, 3, 5, 
	37, 49, 3, 5, 37, 50, 3, 5, 
	37, 51, 3, 5, 38, 49, 3, 5, 
	38, 50, 3, 5, 38, 51, 3, 5, 
	39, 49, 3, 5, 39, 50, 3, 5, 
	39, 51, 3, 5, 40, 49, 3, 5, 
	40, 50, 3, 5, 40, 51, 3, 5, 
	41, 49, 3, 5, 41, 50, 3, 5, 
	41, 51, 3, 6, 9, 49, 3, 6, 
	9, 50, 3, 6, 9, 51, 3, 6, 
	12, 49, 3, 6, 12, 50, 3, 6, 
	12, 51, 3, 6, 36, 49, 3, 6, 
	36, 50, 3, 6, 36, 51, 3, 6, 
	37, 49, 3, 6, 37, 50, 3, 6, 
	37, 51, 3, 6, 38, 49, 3, 6, 
	38, 50, 3, 6, 38, 51, 3, 6, 
	39, 49, 3, 6, 39, 50, 3, 6, 
	39, 51, 3, 6, 40, 49, 3, 6, 
	40, 50, 3, 6, 40, 51, 3, 6, 
	41, 49, 3, 6, 41, 50, 3, 6, 
	41, 51, 3, 10, 12, 51, 3, 11, 
	12, 51, 3, 14, 42, 51, 3, 14, 
	43, 51, 3, 14, 44, 51, 3, 15, 
	42, 51, 3, 15, 43, 51, 3, 15, 
	44, 51, 3, 18, 34, 49, 3, 18, 
	34, 50, 3, 18, 34, 51, 3, 19, 
	35, 49, 3, 19, 35, 50, 3, 19, 
	35, 51, 3, 20, 22, 49, 3, 20, 
	22, 50, 3, 20, 22, 51, 3, 20, 
	23, 49, 3, 20, 23, 50, 3, 20, 
	23, 51, 3, 20, 25, 49, 3, 20, 
	25, 50, 3, 20, 25, 51, 3, 20, 
	26, 49, 3, 20, 26, 50, 3, 20, 
	26, 51, 3, 20, 27, 49, 3, 20, 
	27, 50, 3, 20, 27, 51, 3, 20, 
	28, 49, 3, 20, 28, 50, 3, 20, 
	28, 51, 3, 20, 29, 33, 3, 20, 
	30, 49, 3, 20, 30, 50, 3, 20, 
	30, 51, 3, 20, 31, 49, 3, 20, 
	31, 50, 3, 20, 31, 51, 3, 20, 
	48, 47, 3, 20, 48, 49, 3, 20, 
	48, 50, 3, 20, 48, 51, 3, 29, 
	33, 49, 3, 29, 33, 50, 3, 29, 
	33, 51, 3, 32, 45, 49, 3, 32, 
	45, 50, 3, 32, 45, 51, 3, 32, 
	46, 49, 3, 32, 46, 50, 3, 32, 
	46, 51, 3, 52, 2, 49, 3, 52, 
	2, 50, 3, 52, 2, 51, 3, 52, 
	5, 9, 3, 52, 5, 12, 3, 52, 
	5, 36, 3, 52, 5, 37, 3, 52, 
	5, 38, 3, 52, 5, 39, 3, 52, 
	5, 40, 3, 52, 5, 41, 3, 52, 
	5, 49, 3, 52, 5, 50, 3, 52, 
	5, 51, 3, 52, 7, 49, 3, 52, 
	7, 50, 3, 52, 7, 51, 3, 52, 
	10, 12, 3, 52, 10, 49, 3, 52, 
	10, 50, 3, 52, 10, 51, 3, 52, 
	14, 0, 3, 52, 14, 42, 3, 52, 
	14, 43, 3, 52, 14, 44, 3, 52, 
	14, 49, 3, 52, 14, 50, 3, 52, 
	14, 51, 3, 52, 16, 49, 3, 52, 
	16, 50, 3, 52, 16, 51, 3, 52, 
	32, 45, 3, 52, 32, 46, 3, 53, 
	47, 51, 4, 20, 29, 33, 49, 4, 
	20, 29, 33, 50, 4, 20, 29, 33, 
	51, 4, 20, 48, 47, 51, 4, 52, 
	5, 9, 49, 4, 52, 5, 9, 50, 
	4, 52, 5, 9, 51, 4, 52, 5, 
	12, 49, 4, 52, 5, 12, 50, 4, 
	52, 5, 12, 51, 4, 52, 5, 36, 
	49, 4, 52, 5, 36, 50, 4, 52, 
	5, 36, 51, 4, 52, 5, 37, 49, 
	4, 52, 5, 37, 50, 4, 52, 5, 
	37, 51, 4, 52, 5, 38, 49, 4, 
	52, 5, 38, 50, 4, 52, 5, 38, 
	51, 4, 52, 5, 39, 49, 4, 52, 
	5, 39, 50, 4, 52, 5, 39, 51, 
	4, 52, 5, 40, 49, 4, 52, 5, 
	40, 50, 4, 52, 5, 40, 51, 4, 
	52, 5, 41, 49, 4, 52, 5, 41, 
	50, 4, 52, 5, 41, 51, 4, 52, 
	10, 12, 51, 4, 52, 14, 42, 51, 
	4, 52, 14, 43, 51, 4, 52, 14, 
	44, 51, 4, 52, 32, 45, 49, 4, 
	52, 32, 45, 50, 4, 52, 32, 45, 
	51, 4, 52, 32, 46, 49, 4, 52, 
	32, 46, 50, 4, 52, 32, 46, 51
	
};

static const short _zone_cond_offsets[] = {
	0, 0, 2, 4, 6, 8, 10, 12, 
	12, 12, 12, 12, 12, 12, 14, 16, 
	19, 22, 24, 26, 26, 26, 26, 26, 
	26, 26, 29, 29, 29, 31, 33, 35, 
	37, 39, 39, 39, 39, 39, 39, 41, 
	43, 45, 45, 45, 45, 45, 47, 49, 
	52, 52, 52, 52, 52, 54, 56, 58, 
	58, 60, 60, 60, 60, 62, 62, 62, 
	62, 62, 64, 67, 67, 67, 67, 69, 
	71, 73, 73, 75, 75, 77, 77, 77, 
	77, 79, 79, 79, 82, 84, 86, 88, 
	88, 90, 90, 92, 92, 92, 92, 94, 
	94, 94, 97, 97, 99, 101, 101, 103, 
	105, 107, 109, 111, 111, 113, 113, 113, 
	113, 115, 115, 115, 115, 115, 117, 120, 
	123, 123, 123, 123, 123, 125, 127, 129, 
	131, 133, 135, 137, 139, 141, 143, 145, 
	147, 149, 149, 151, 151, 151, 151, 153, 
	153, 153, 153, 153, 155, 158, 158, 160, 
	160, 160, 160, 162, 162, 162, 162, 162, 
	165, 165, 167, 167, 167, 167, 169, 169, 
	169, 169, 169, 172, 172, 174, 174, 174, 
	174, 176, 176, 176, 176, 176, 179, 182, 
	185, 187, 189, 191, 191, 193, 193, 193, 
	193, 195, 195, 195, 195, 195, 197, 200, 
	200, 200, 202, 204, 206, 206, 208, 208, 
	208, 208, 210, 210, 210, 210, 210, 212, 
	215, 215, 215, 217, 219, 221, 223, 225, 
	227, 229, 231, 233, 235, 237, 239, 241, 
	243, 245, 247, 250, 252, 255, 257, 260, 
	262, 265, 267, 270, 270, 272, 272, 272, 
	272, 274, 274, 274, 274, 274, 277, 277, 
	279, 279, 279, 279, 281, 281, 281, 281, 
	281, 283, 286, 286, 288, 290, 292, 294, 
	294, 296, 296, 298, 301, 301, 301, 301, 
	301, 301, 301, 303, 306, 308, 310, 312, 
	314, 314, 316, 316, 318, 321, 321, 321, 
	321, 321, 321, 321, 323, 326, 329, 332, 
	332, 334, 336, 338, 340, 342, 344, 346, 
	348, 350, 350, 352, 352, 352, 352, 354, 
	354, 354, 354, 354, 356, 359, 362, 365, 
	368, 368, 368, 370, 372, 374, 376, 376, 
	378, 378, 380, 383, 383, 383, 383, 383, 
	383, 383, 385, 388, 391, 391, 391, 391, 
	393, 395, 395, 397, 399, 401, 403, 403, 
	405, 408, 411, 414, 417, 420, 422, 422, 
	424, 426, 428, 430, 433, 436, 436, 438, 
	438, 438, 440, 443, 443, 445, 445, 445, 
	445, 445, 445, 445, 445, 445, 445, 445, 
	445, 445, 445, 445, 445, 445, 445, 447, 
	449, 451, 454, 456, 458, 460, 463, 463, 
	463, 463, 463, 463, 463, 465, 467, 469, 
	471, 473, 473, 475, 475, 475, 475, 477, 
	477, 477, 477, 477, 479, 482, 482, 484, 
	484, 484, 484, 486, 486, 486, 486, 486, 
	489, 489, 489, 489, 489, 489, 491, 493, 
	495, 495, 497, 497, 497, 497, 499, 499, 
	499, 499, 499, 501, 504, 504, 504, 506, 
	508, 510, 512, 515, 517
};

static const char _zone_cond_lengths[] = {
	0, 2, 2, 2, 2, 2, 2, 0, 
	0, 0, 0, 0, 0, 2, 2, 3, 
	3, 2, 2, 0, 0, 0, 0, 0, 
	0, 3, 0, 0, 2, 2, 2, 2, 
	2, 0, 0, 0, 0, 0, 2, 2, 
	2, 0, 0, 0, 0, 2, 2, 3, 
	0, 0, 0, 0, 2, 2, 2, 0, 
	2, 0, 0, 0, 2, 0, 0, 0, 
	0, 2, 3, 0, 0, 0, 2, 2, 
	2, 0, 2, 0, 2, 0, 0, 0, 
	2, 0, 0, 3, 2, 2, 2, 0, 
	2, 0, 2, 0, 0, 0, 2, 0, 
	0, 3, 0, 2, 2, 0, 2, 2, 
	2, 2, 2, 0, 2, 0, 0, 0, 
	2, 0, 0, 0, 0, 2, 3, 3, 
	0, 0, 0, 0, 2, 2, 2, 2, 
	2, 2, 2, 2, 2, 2, 2, 2, 
	2, 0, 2, 0, 0, 0, 2, 0, 
	0, 0, 0, 2, 3, 0, 2, 0, 
	0, 0, 2, 0, 0, 0, 0, 3, 
	0, 2, 0, 0, 0, 2, 0, 0, 
	0, 0, 3, 0, 2, 0, 0, 0, 
	2, 0, 0, 0, 0, 3, 3, 3, 
	2, 2, 2, 0, 2, 0, 0, 0, 
	2, 0, 0, 0, 0, 2, 3, 0, 
	0, 2, 2, 2, 0, 2, 0, 0, 
	0, 2, 0, 0, 0, 0, 2, 3, 
	0, 0, 2, 2, 2, 2, 2, 2, 
	2, 2, 2, 2, 2, 2, 2, 2, 
	2, 2, 3, 2, 3, 2, 3, 2, 
	3, 2, 3, 0, 2, 0, 0, 0, 
	2, 0, 0, 0, 0, 3, 0, 2, 
	0, 0, 0, 2, 0, 0, 0, 0, 
	2, 3, 0, 2, 2, 2, 2, 0, 
	2, 0, 2, 3, 0, 0, 0, 0, 
	0, 0, 2, 3, 2, 2, 2, 2, 
	0, 2, 0, 2, 3, 0, 0, 0, 
	0, 0, 0, 2, 3, 3, 3, 0, 
	2, 2, 2, 2, 2, 2, 2, 2, 
	2, 0, 2, 0, 0, 0, 2, 0, 
	0, 0, 0, 2, 3, 3, 3, 3, 
	0, 0, 2, 2, 2, 2, 0, 2, 
	0, 2, 3, 0, 0, 0, 0, 0, 
	0, 2, 3, 3, 0, 0, 0, 2, 
	2, 0, 2, 2, 2, 2, 0, 2, 
	3, 3, 3, 3, 3, 2, 0, 2, 
	2, 2, 2, 3, 3, 0, 2, 0, 
	0, 2, 3, 0, 2, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 2, 2, 
	2, 3, 2, 2, 2, 3, 0, 0, 
	0, 0, 0, 0, 2, 2, 2, 2, 
	2, 0, 2, 0, 0, 0, 2, 0, 
	0, 0, 0, 2, 3, 0, 2, 0, 
	0, 0, 2, 0, 0, 0, 0, 3, 
	0, 0, 0, 0, 0, 2, 2, 2, 
	0, 2, 0, 0, 0, 2, 0, 0, 
	0, 0, 2, 3, 0, 0, 2, 2, 
	2, 2, 3, 2, 2
};

static const short _zone_cond_keys[] = {
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, -128, 9, 
	10, 10, 11, 127, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, -128, 9, 10, 10, 
	11, 127, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, -128, 9, 
	10, 10, 11, 127, -128, 9, 10, 10, 
	11, 127, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	-128, 9, 10, 10, 11, 127, -128, 9, 
	10, 10, 11, 127, -128, 9, 10, 10, 
	11, 127, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, -128, 9, 
	10, 10, 11, 127, 10, 10, 59, 59, 
	10, 10, 59, 59, -128, 9, 10, 10, 
	11, 127, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, 10, 10, 59, 59, 10, 10, 
	59, 59, -128, 9, 10, 10, 11, 127, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	10, 10, 59, 59, 10, 10, 59, 59, 
	-128, 9, 10, 10, 11, 127, 10, 10, 
	59, 59, 10, 10, 59, 59, 0
};

static const char _zone_cond_spaces[] = {
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0
};

static const short _zone_key_offsets[] = {
	0, 0, 15, 41, 57, 81, 89, 97, 
	100, 102, 105, 107, 110, 112, 122, 130, 
	133, 136, 146, 154, 157, 158, 161, 162, 
	165, 166, 169, 171, 173, 179, 191, 206, 
	221, 237, 239, 242, 244, 247, 249, 259, 
	269, 277, 280, 281, 284, 285, 301, 317, 
	320, 322, 324, 326, 328, 334, 351, 368, 
	373, 390, 392, 394, 397, 405, 410, 413, 
	415, 417, 425, 428, 430, 432, 436, 442, 
	457, 475, 483, 500, 505, 522, 524, 526, 
	531, 549, 551, 553, 556, 562, 577, 595, 
	603, 620, 625, 642, 644, 646, 651, 669, 
	671, 673, 676, 678, 684, 706, 708, 714, 
	722, 730, 747, 764, 769, 786, 788, 790, 
	793, 801, 806, 809, 811, 813, 821, 824, 
	827, 831, 833, 835, 837, 843, 851, 859, 
	867, 875, 889, 904, 918, 933, 947, 962, 
	979, 996, 1001, 1018, 1020, 1022, 1025, 1033, 
	1038, 1041, 1043, 1045, 1053, 1056, 1061, 1076, 
	1078, 1080, 1083, 1089, 1094, 1097, 1099, 1101, 
	1104, 1109, 1124, 1126, 1128, 1131, 1137, 1142, 
	1145, 1147, 1149, 1152, 1157, 1172, 1174, 1176, 
	1179, 1185, 1190, 1193, 1195, 1197, 1200, 1203, 
	1206, 1212, 1229, 1246, 1251, 1268, 1270, 1272, 
	1275, 1283, 1288, 1291, 1293, 1295, 1303, 1306, 
	1308, 1310, 1316, 1333, 1350, 1355, 1372, 1374, 
	1376, 1379, 1387, 1392, 1395, 1397, 1399, 1407, 
	1410, 1416, 1418, 1424, 1441, 1456, 1474, 1489, 
	1497, 1513, 1521, 1537, 1545, 1561, 1569, 1585, 
	1593, 1611, 1619, 1622, 1628, 1631, 1637, 1640, 
	1646, 1649, 1655, 1658, 1663, 1678, 1680, 1682, 
	1685, 1691, 1696, 1699, 1701, 1703, 1706, 1711, 
	1726, 1728, 1730, 1733, 1739, 1744, 1747, 1749, 
	1751, 1757, 1760, 1762, 1769, 1783, 1799, 1815, 
	1818, 1834, 1839, 1855, 1858, 1860, 1862, 1867, 
	1870, 1872, 1874, 1890, 1893, 1899, 1913, 1929, 
	1945, 1948, 1964, 1969, 1985, 1988, 1990, 1992, 
	1997, 2000, 2002, 2004, 2020, 2023, 2026, 2029, 
	2031, 2037, 2045, 2053, 2061, 2069, 2077, 2085, 
	2102, 2119, 2124, 2141, 2143, 2145, 2148, 2156, 
	2161, 2164, 2166, 2168, 2176, 2179, 2182, 2185, 
	2188, 2192, 2194, 2200, 2214, 2230, 2246, 2249, 
	2265, 2270, 2286, 2289, 2291, 2293, 2298, 2301, 
	2303, 2305, 2321, 2324, 2327, 2329, 2331, 2333, 
	2341, 2348, 2349, 2355, 2363, 2373, 2385, 2391, 
	2405, 2408, 2411, 2414, 2417, 2420, 2426, 2428, 
	2434, 2458, 2474, 2480, 2483, 2486, 2491, 2506, 
	2508, 2510, 2538, 2541, 2544, 2550, 2555, 2558, 
	2560, 2562, 2570, 2572, 2574, 2576, 2577, 2579, 
	2581, 2583, 2585, 2587, 2588, 2590, 2592, 2598, 
	2606, 2616, 2619, 2625, 2633, 2643, 2646, 2648, 
	2650, 2652, 2654, 2656, 2658, 2664, 2678, 2695, 
	2712, 2729, 2734, 2751, 2753, 2755, 2758, 2766, 
	2771, 2774, 2776, 2778, 2786, 2789, 2794, 2811, 
	2813, 2815, 2818, 2826, 2831, 2834, 2836, 2838, 
	2841, 2843, 2845, 2847, 2849, 2851, 2857, 2874, 
	2891, 2896, 2913, 2915, 2917, 2920, 2928, 2933, 
	2936, 2938, 2940, 2948, 2951, 2953, 2955, 2961, 
	2969, 2987, 2995, 2998, 3004
};

static const short _zone_trans_keys[] = {
	9, 32, 40, 41, 92, 522, 571, -128, 
	8, 11, 33, 35, 58, 60, 127, 9, 
	32, 40, 41, 65, 67, 68, 73, 77, 
	78, 80, 83, 84, 97, 99, 100, 105, 
	109, 110, 112, 115, 116, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 522, 571, 48, 
	57, 9, 32, 40, 41, 65, 67, 68, 
	73, 77, 78, 80, 83, 84, 97, 99, 
	100, 105, 109, 110, 112, 115, 116, 522, 
	571, 9, 32, 40, 41, 65, 97, 522, 
	571, 9, 32, 40, 41, 522, 571, 48, 
	57, 46, 48, 57, 48, 57, 46, 48, 
	57, 48, 57, 46, 48, 57, 48, 57, 
	9, 32, 40, 41, 266, 315, 522, 571, 
	48, 57, 9, 32, 40, 41, 266, 315, 
	522, 571, 266, 128, 383, 522, 384, 639, 
	9, 32, 40, 41, 266, 315, 522, 571, 
	48, 57, 9, 32, 40, 41, 266, 315, 
	522, 571, 46, 48, 57, 46, 46, 48, 
	57, 46, 46, 48, 57, 46, 522, 384, 
	639, 65, 97, 65, 97, 9, 32, 40, 
	41, 522, 571, 9, 32, 40, 41, 522, 
	571, 48, 58, 65, 70, 97, 102, 9, 
	32, 40, 41, 58, 266, 315, 522, 571, 
	48, 57, 65, 70, 97, 102, 9, 32, 
	40, 41, 58, 266, 315, 522, 571, 48, 
	57, 65, 70, 97, 102, 9, 32, 40, 
	41, 46, 58, 266, 315, 522, 571, 48, 
	57, 65, 70, 97, 102, 48, 57, 46, 
	48, 57, 48, 57, 46, 48, 57, 48, 
	57, 9, 32, 40, 41, 266, 315, 522, 
	571, 48, 57, 9, 32, 40, 41, 266, 
	315, 522, 571, 48, 57, 9, 32, 40, 
	41, 266, 315, 522, 571, 46, 48, 57, 
	46, 46, 48, 57, 46, 9, 32, 40, 
	41, 46, 58, 266, 315, 522, 571, 48, 
	57, 65, 70, 97, 102, 9, 32, 40, 
	41, 46, 58, 266, 315, 522, 571, 48, 
	57, 65, 70, 97, 102, 522, 384, 639, 
	78, 110, 65, 97, 77, 109, 69, 101, 
	9, 32, 40, 41, 522, 571, 9, 32, 
	34, 40, 41, 64, 92, 522, 571, -128, 
	8, 11, 35, 37, 58, 60, 127, 9, 
	32, 40, 41, 92, 266, 315, 522, 571, 
	-128, 8, 11, 33, 35, 58, 60, 127, 
	10, 48, 49, 50, 57, 9, 32, 40, 
	41, 92, 266, 315, 522, 571, -128, 8, 
	11, 33, 35, 58, 60, 127, 48, 57, 
	48, 57, 10, 34, 92, 9, 32, 40, 
	41, 266, 315, 522, 571, 10, 48, 49, 
	50, 57, 10, 34, 92, 48, 57, 48, 
	57, 9, 32, 40, 41, 266, 315, 522, 
	571, 522, 384, 639, 89, 121, 78, 110, 
	65, 67, 97, 99, 9, 32, 40, 41, 
	522, 571, 9, 32, 40, 41, 92, 522, 
	571, -128, 8, 11, 31, 35, 58, 60, 
	127, 9, 32, 33, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 31, 35, 
	58, 60, 127, 59, 92, 9, 10, 32, 
	34, 40, 41, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 31, 
	35, 58, 60, 127, 10, 48, 49, 50, 
	57, 9, 32, 40, 41, 92, 266, 315, 
	522, 571, -128, 8, 11, 31, 35, 58, 
	60, 127, 48, 57, 48, 57, 10, 48, 
	49, 50, 57, 9, 32, 33, 40, 41, 
	92, 266, 315, 522, 571, -128, 8, 11, 
	31, 35, 58, 60, 127, 48, 57, 48, 
	57, 522, 384, 639, 9, 32, 40, 41, 
	522, 571, 9, 32, 40, 41, 92, 522, 
	571, -128, 8, 11, 31, 35, 58, 60, 
	127, 9, 32, 33, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 31, 35, 
	58, 60, 127, 59, 92, 9, 10, 32, 
	34, 40, 41, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 31, 
	35, 58, 60, 127, 10, 48, 49, 50, 
	57, 9, 32, 40, 41, 92, 266, 315, 
	522, 571, -128, 8, 11, 31, 35, 58, 
	60, 127, 48, 57, 48, 57, 10, 48, 
	49, 50, 57, 9, 32, 33, 40, 41, 
	92, 266, 315, 522, 571, -128, 8, 11, 
	31, 35, 58, 60, 127, 48, 57, 48, 
	57, 522, 384, 639, 78, 110, 9, 32, 
	40, 41, 522, 571, 9, 32, 40, 41, 
	65, 67, 68, 77, 78, 80, 83, 84, 
	97, 99, 100, 109, 110, 112, 115, 116, 
	522, 571, 88, 120, 9, 32, 40, 41, 
	522, 571, 9, 32, 40, 41, 522, 571, 
	48, 57, 9, 32, 40, 41, 522, 571, 
	48, 57, 9, 32, 34, 40, 41, 64, 
	92, 522, 571, -128, 8, 11, 35, 37, 
	58, 60, 127, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 10, 48, 49, 50, 
	57, 9, 32, 40, 41, 92, 266, 315, 
	522, 571, -128, 8, 11, 33, 35, 58, 
	60, 127, 48, 57, 48, 57, 10, 34, 
	92, 9, 32, 40, 41, 266, 315, 522, 
	571, 10, 48, 49, 50, 57, 10, 34, 
	92, 48, 57, 48, 57, 9, 32, 40, 
	41, 266, 315, 522, 571, 522, 384, 639, 
	522, 384, 639, 65, 83, 97, 115, 80, 
	112, 84, 116, 82, 114, 9, 32, 40, 
	41, 522, 571, 9, 32, 40, 41, 522, 
	571, 48, 57, 9, 32, 40, 41, 522, 
	571, 48, 57, 9, 32, 40, 41, 522, 
	571, 48, 57, 9, 32, 40, 41, 522, 
	571, 48, 57, 9, 32, 34, 40, 41, 
	92, 522, 571, -128, 8, 11, 58, 60, 
	127, 9, 32, 40, 41, 92, 522, 571, 
	-128, 8, 11, 33, 35, 58, 60, 127, 
	9, 32, 34, 40, 41, 92, 522, 571, 
	-128, 8, 11, 58, 60, 127, 9, 32, 
	40, 41, 92, 522, 571, -128, 8, 11, 
	33, 35, 58, 60, 127, 9, 32, 34, 
	40, 41, 92, 522, 571, -128, 8, 11, 
	58, 60, 127, 9, 32, 40, 41, 92, 
	522, 571, -128, 8, 11, 33, 35, 58, 
	60, 127, 9, 32, 34, 40, 41, 64, 
	92, 522, 571, -128, 8, 11, 35, 37, 
	58, 60, 127, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 10, 48, 49, 50, 
	57, 9, 32, 40, 41, 92, 266, 315, 
	522, 571, -128, 8, 11, 33, 35, 58, 
	60, 127, 48, 57, 48, 57, 10, 34, 
	92, 9, 32, 40, 41, 266, 315, 522, 
	571, 10, 48, 49, 50, 57, 10, 34, 
	92, 48, 57, 48, 57, 9, 32, 40, 
	41, 266, 315, 522, 571, 522, 384, 639, 
	10, 48, 49, 50, 57, 9, 32, 40, 
	41, 92, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 48, 57, 48, 57, 
	10, 34, 92, 9, 32, 40, 41, 522, 
	571, 10, 48, 49, 50, 57, 10, 34, 
	92, 48, 57, 48, 57, 522, 384, 639, 
	10, 48, 49, 50, 57, 9, 32, 40, 
	41, 92, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 48, 57, 48, 57, 
	10, 34, 92, 9, 32, 40, 41, 522, 
	571, 10, 48, 49, 50, 57, 10, 34, 
	92, 48, 57, 48, 57, 522, 384, 639, 
	10, 48, 49, 50, 57, 9, 32, 40, 
	41, 92, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 48, 57, 48, 57, 
	10, 34, 92, 9, 32, 40, 41, 522, 
	571, 10, 48, 49, 50, 57, 10, 34, 
	92, 48, 57, 48, 57, 522, 384, 639, 
	522, 384, 639, 522, 384, 639, 9, 32, 
	40, 41, 522, 571, 9, 32, 34, 40, 
	41, 64, 92, 522, 571, -128, 8, 11, 
	35, 37, 58, 60, 127, 9, 32, 40, 
	41, 92, 266, 315, 522, 571, -128, 8, 
	11, 33, 35, 58, 60, 127, 10, 48, 
	49, 50, 57, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 48, 57, 48, 57, 
	10, 34, 92, 9, 32, 40, 41, 266, 
	315, 522, 571, 10, 48, 49, 50, 57, 
	10, 34, 92, 48, 57, 48, 57, 9, 
	32, 40, 41, 266, 315, 522, 571, 522, 
	384, 639, 84, 116, 82, 114, 9, 32, 
	40, 41, 522, 571, 9, 32, 34, 40, 
	41, 64, 92, 522, 571, -128, 8, 11, 
	35, 37, 58, 60, 127, 9, 32, 40, 
	41, 92, 266, 315, 522, 571, -128, 8, 
	11, 33, 35, 58, 60, 127, 10, 48, 
	49, 50, 57, 9, 32, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 33, 
	35, 58, 60, 127, 48, 57, 48, 57, 
	10, 34, 92, 9, 32, 40, 41, 266, 
	315, 522, 571, 10, 48, 49, 50, 57, 
	10, 34, 92, 48, 57, 48, 57, 9, 
	32, 40, 41, 266, 315, 522, 571, 522, 
	384, 639, 79, 80, 82, 111, 112, 114, 
	65, 97, 9, 32, 40, 41, 522, 571, 
	9, 32, 34, 40, 41, 64, 92, 522, 
	571, -128, 8, 11, 35, 37, 58, 60, 
	127, 9, 32, 40, 41, 92, 522, 571, 
	-128, 8, 11, 33, 35, 58, 60, 127, 
	9, 32, 34, 40, 41, 92, 522, 571, 
	-128, 8, 11, 35, 37, 58, 60, 63, 
	65, 127, 9, 32, 40, 41, 92, 522, 
	571, -128, 8, 11, 33, 35, 58, 60, 
	127, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 522, 571, 48, 
	57, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 522, 571, 48, 
	57, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 522, 571, 48, 
	57, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 522, 571, 48, 
	57, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 266, 315, 522, 
	571, 48, 57, 9, 32, 40, 41, 266, 
	315, 522, 571, 522, 384, 639, 9, 32, 
	40, 41, 522, 571, 522, 384, 639, 9, 
	32, 40, 41, 522, 571, 522, 384, 639, 
	9, 32, 40, 41, 522, 571, 522, 384, 
	639, 9, 32, 40, 41, 522, 571, 522, 
	384, 639, 10, 48, 49, 50, 57, 9, 
	32, 40, 41, 92, 522, 571, -128, 8, 
	11, 33, 35, 58, 60, 127, 48, 57, 
	48, 57, 10, 34, 92, 9, 32, 40, 
	41, 522, 571, 10, 48, 49, 50, 57, 
	10, 34, 92, 48, 57, 48, 57, 522, 
	384, 639, 10, 48, 49, 50, 57, 9, 
	32, 40, 41, 92, 522, 571, -128, 8, 
	11, 33, 35, 58, 60, 127, 48, 57, 
	48, 57, 10, 34, 92, 9, 32, 40, 
	41, 522, 571, 10, 48, 49, 50, 57, 
	10, 34, 92, 48, 57, 48, 57, 9, 
	32, 40, 41, 522, 571, 522, 384, 639, 
	70, 102, 9, 32, 40, 41, 43, 522, 
	571, 9, 32, 34, 40, 41, 92, 522, 
	571, -128, 8, 11, 58, 60, 127, 9, 
	32, 34, 40, 41, 92, 266, 315, 522, 
	571, -128, 8, 11, 58, 60, 127, 9, 
	32, 34, 40, 41, 92, 266, 315, 522, 
	571, -128, 8, 11, 58, 60, 127, 10, 
	34, 92, 9, 32, 34, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 58, 
	60, 127, 10, 48, 49, 50, 57, 9, 
	32, 34, 40, 41, 92, 266, 315, 522, 
	571, -128, 8, 11, 58, 60, 127, 522, 
	384, 639, 48, 57, 48, 57, 10, 48, 
	49, 50, 57, 10, 34, 92, 48, 57, 
	48, 57, 9, 32, 34, 40, 41, 92, 
	266, 315, 522, 571, -128, 8, 11, 58, 
	60, 127, 522, 384, 639, 9, 32, 40, 
	41, 522, 571, 9, 32, 34, 40, 41, 
	92, 522, 571, -128, 8, 11, 58, 60, 
	127, 9, 32, 34, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 58, 60, 
	127, 9, 32, 34, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 58, 60, 
	127, 10, 34, 92, 9, 32, 34, 40, 
	41, 92, 266, 315, 522, 571, -128, 8, 
	11, 58, 60, 127, 10, 48, 49, 50, 
	57, 9, 32, 34, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 58, 60, 
	127, 522, 384, 639, 48, 57, 48, 57, 
	10, 48, 49, 50, 57, 10, 34, 92, 
	48, 57, 48, 57, 9, 32, 34, 40, 
	41, 92, 266, 315, 522, 571, -128, 8, 
	11, 58, 60, 127, 522, 384, 639, 522, 
	384, 639, 522, 384, 639, 86, 118, 9, 
	32, 40, 41, 522, 571, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 34, 
	40, 41, 64, 92, 522, 571, -128, 8, 
	11, 35, 37, 58, 60, 127, 9, 32, 
	40, 41, 92, 266, 315, 522, 571, -128, 
	8, 11, 33, 35, 58, 60, 127, 10, 
	48, 49, 50, 57, 9, 32, 40, 41, 
	92, 266, 315, 522, 571, -128, 8, 11, 
	33, 35, 58, 60, 127, 48, 57, 48, 
	57, 10, 34, 92, 9, 32, 40, 41, 
	266, 315, 522, 571, 10, 48, 49, 50, 
	57, 10, 34, 92, 48, 57, 48, 57, 
	9, 32, 40, 41, 266, 315, 522, 571, 
	522, 384, 639, 522, 384, 639, 522, 384, 
	639, 522, 384, 639, 88, 89, 120, 121, 
	84, 116, 9, 32, 40, 41, 522, 571, 
	9, 32, 34, 40, 41, 92, 522, 571, 
	-128, 8, 11, 58, 60, 127, 9, 32, 
	34, 40, 41, 92, 266, 315, 522, 571, 
	-128, 8, 11, 58, 60, 127, 9, 32, 
	34, 40, 41, 92, 266, 315, 522, 571, 
	-128, 8, 11, 58, 60, 127, 10, 34, 
	92, 9, 32, 34, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 58, 60, 
	127, 10, 48, 49, 50, 57, 9, 32, 
	34, 40, 41, 92, 266, 315, 522, 571, 
	-128, 8, 11, 58, 60, 127, 522, 384, 
	639, 48, 57, 48, 57, 10, 48, 49, 
	50, 57, 10, 34, 92, 48, 57, 48, 
	57, 9, 32, 34, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 58, 60, 
	127, 522, 384, 639, 522, 384, 639, 80, 
	112, 69, 101, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 9, 32, 40, 
	41, 92, 522, 571, 35, 9, 32, 40, 
	41, 522, 571, 9, 32, 40, 41, 522, 
	571, 48, 57, 9, 32, 40, 41, 266, 
	315, 522, 571, 48, 57, 9, 32, 40, 
	41, 522, 571, 48, 57, 65, 70, 97, 
	102, 48, 57, 65, 70, 97, 102, 9, 
	32, 40, 41, 266, 315, 522, 571, 48, 
	57, 65, 70, 97, 102, 522, 384, 639, 
	522, 384, 639, 522, 384, 639, 522, 384, 
	639, 522, 384, 639, 9, 32, 40, 41, 
	522, 571, 78, 110, 9, 32, 40, 41, 
	522, 571, 9, 32, 40, 41, 65, 67, 
	68, 77, 78, 80, 83, 84, 97, 99, 
	100, 109, 110, 112, 115, 116, 522, 571, 
	48, 57, 9, 32, 40, 41, 68, 72, 
	77, 87, 100, 104, 109, 119, 522, 571, 
	48, 57, 9, 32, 40, 41, 522, 571, 
	522, 384, 639, 522, 384, 639, 10, 48, 
	49, 50, 57, 9, 32, 40, 41, 92, 
	522, 571, -128, 8, 11, 33, 35, 58, 
	60, 127, 48, 57, 48, 57, 9, 32, 
	40, 41, 65, 67, 68, 73, 77, 78, 
	80, 83, 84, 97, 99, 100, 105, 109, 
	110, 112, 115, 116, 266, 315, 522, 571, 
	48, 57, 522, 384, 639, 10, 34, 92, 
	9, 32, 40, 41, 522, 571, 10, 48, 
	49, 50, 57, 10, 34, 92, 48, 57, 
	48, 57, 65, 73, 79, 84, 97, 105, 
	111, 116, 68, 100, 68, 100, 82, 114, 
	95, 76, 108, 73, 105, 77, 109, 73, 
	105, 84, 116, 95, 86, 118, 52, 54, 
	9, 32, 40, 41, 522, 571, 9, 32, 
	40, 41, 522, 571, 48, 57, 9, 32, 
	40, 41, 266, 315, 522, 571, 48, 57, 
	522, 384, 639, 9, 32, 40, 41, 522, 
	571, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 266, 315, 522, 
	571, 48, 57, 522, 384, 639, 78, 110, 
	67, 99, 76, 108, 85, 117, 68, 100, 
	69, 101, 9, 32, 40, 41, 522, 571, 
	9, 32, 34, 40, 41, 92, 522, 571, 
	-128, 8, 11, 58, 60, 127, 9, 32, 
	40, 41, 92, 266, 315, 522, 571, -128, 
	8, 11, 33, 35, 58, 60, 127, 9, 
	32, 34, 40, 41, 64, 92, 522, 571, 
	-128, 8, 11, 35, 37, 58, 60, 127, 
	9, 32, 40, 41, 92, 266, 315, 522, 
	571, -128, 8, 11, 33, 35, 58, 60, 
	127, 10, 48, 49, 50, 57, 9, 32, 
	40, 41, 92, 266, 315, 522, 571, -128, 
	8, 11, 33, 35, 58, 60, 127, 48, 
	57, 48, 57, 10, 34, 92, 9, 32, 
	40, 41, 266, 315, 522, 571, 10, 48, 
	49, 50, 57, 10, 34, 92, 48, 57, 
	48, 57, 9, 32, 40, 41, 266, 315, 
	522, 571, 522, 384, 639, 10, 48, 49, 
	50, 57, 9, 32, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 33, 35, 
	58, 60, 127, 48, 57, 48, 57, 10, 
	34, 92, 9, 32, 40, 41, 266, 315, 
	522, 571, 10, 48, 49, 50, 57, 10, 
	34, 92, 48, 57, 48, 57, 522, 384, 
	639, 82, 114, 73, 105, 71, 103, 73, 
	105, 78, 110, 9, 32, 40, 41, 522, 
	571, 9, 32, 34, 40, 41, 64, 92, 
	522, 571, -128, 8, 11, 35, 37, 58, 
	60, 127, 9, 32, 40, 41, 92, 266, 
	315, 522, 571, -128, 8, 11, 33, 35, 
	58, 60, 127, 10, 48, 49, 50, 57, 
	9, 32, 40, 41, 92, 266, 315, 522, 
	571, -128, 8, 11, 33, 35, 58, 60, 
	127, 48, 57, 48, 57, 10, 34, 92, 
	9, 32, 40, 41, 266, 315, 522, 571, 
	10, 48, 49, 50, 57, 10, 34, 92, 
	48, 57, 48, 57, 9, 32, 40, 41, 
	266, 315, 522, 571, 522, 384, 639, 84, 
	116, 76, 108, 9, 32, 40, 41, 522, 
	571, 9, 32, 40, 41, 522, 571, 48, 
	57, 9, 32, 40, 41, 68, 72, 77, 
	87, 100, 104, 109, 119, 266, 315, 522, 
	571, 48, 57, 9, 32, 40, 41, 266, 
	315, 522, 571, 522, 384, 639, 9, 32, 
	40, 41, 522, 571, 9, 32, 34, 36, 
	40, 41, 64, 92, 266, 315, 522, 571, 
	-128, 8, 11, 58, 60, 127, 0
};

static const char _zone_single_lengths[] = {
	0, 7, 24, 14, 24, 8, 6, 1, 
	0, 1, 0, 1, 0, 8, 8, 1, 
	1, 8, 8, 1, 1, 1, 1, 1, 
	1, 1, 2, 2, 6, 6, 9, 9, 
	10, 0, 1, 0, 1, 0, 8, 8, 
	8, 1, 1, 1, 1, 10, 10, 1, 
	2, 2, 2, 2, 6, 9, 9, 1, 
	9, 0, 0, 3, 8, 1, 3, 0, 
	0, 8, 1, 2, 2, 4, 6, 7, 
	10, 2, 9, 1, 9, 0, 0, 1, 
	10, 0, 0, 1, 6, 7, 10, 2, 
	9, 1, 9, 0, 0, 1, 10, 0, 
	0, 1, 2, 6, 22, 2, 6, 6, 
	6, 9, 9, 1, 9, 0, 0, 3, 
	8, 1, 3, 0, 0, 8, 1, 1, 
	4, 2, 2, 2, 6, 6, 6, 6, 
	6, 8, 7, 8, 7, 8, 7, 9, 
	9, 1, 9, 0, 0, 3, 8, 1, 
	3, 0, 0, 8, 1, 1, 7, 0, 
	0, 3, 6, 1, 3, 0, 0, 1, 
	1, 7, 0, 0, 3, 6, 1, 3, 
	0, 0, 1, 1, 7, 0, 0, 3, 
	6, 1, 3, 0, 0, 1, 1, 1, 
	6, 9, 9, 1, 9, 0, 0, 3, 
	8, 1, 3, 0, 0, 8, 1, 2, 
	2, 6, 9, 9, 1, 9, 0, 0, 
	3, 8, 1, 3, 0, 0, 8, 1, 
	6, 2, 6, 9, 7, 8, 7, 6, 
	14, 6, 14, 6, 14, 6, 14, 6, 
	16, 8, 1, 6, 1, 6, 1, 6, 
	1, 6, 1, 1, 7, 0, 0, 3, 
	6, 1, 3, 0, 0, 1, 1, 7, 
	0, 0, 3, 6, 1, 3, 0, 0, 
	6, 1, 2, 7, 8, 10, 10, 3, 
	10, 1, 10, 1, 0, 0, 1, 3, 
	0, 0, 10, 1, 6, 8, 10, 10, 
	3, 10, 1, 10, 1, 0, 0, 1, 
	3, 0, 0, 10, 1, 1, 1, 2, 
	6, 6, 6, 6, 6, 6, 6, 9, 
	9, 1, 9, 0, 0, 3, 8, 1, 
	3, 0, 0, 8, 1, 1, 1, 1, 
	4, 2, 6, 8, 10, 10, 3, 10, 
	1, 10, 1, 0, 0, 1, 3, 0, 
	0, 10, 1, 1, 2, 2, 0, 6, 
	7, 1, 6, 6, 8, 6, 0, 8, 
	1, 1, 1, 1, 1, 6, 2, 6, 
	22, 14, 6, 1, 1, 1, 7, 0, 
	0, 26, 1, 3, 6, 1, 3, 0, 
	0, 8, 2, 2, 2, 1, 2, 2, 
	2, 2, 2, 1, 2, 2, 6, 6, 
	8, 1, 6, 6, 8, 1, 2, 2, 
	2, 2, 2, 2, 6, 8, 9, 9, 
	9, 1, 9, 0, 0, 3, 8, 1, 
	3, 0, 0, 8, 1, 1, 9, 0, 
	0, 3, 8, 1, 3, 0, 0, 1, 
	2, 2, 2, 2, 2, 6, 9, 9, 
	1, 9, 0, 0, 3, 8, 1, 3, 
	0, 0, 8, 1, 2, 2, 6, 6, 
	16, 8, 1, 6, 12
};

static const char _zone_range_lengths[] = {
	0, 4, 1, 1, 0, 0, 1, 1, 
	1, 1, 1, 1, 1, 1, 0, 1, 
	1, 1, 0, 1, 0, 1, 0, 1, 
	0, 1, 0, 0, 0, 3, 3, 3, 
	3, 1, 1, 1, 1, 1, 1, 1, 
	0, 1, 0, 1, 0, 3, 3, 1, 
	0, 0, 0, 0, 0, 4, 4, 2, 
	4, 1, 1, 0, 0, 2, 0, 1, 
	1, 0, 1, 0, 0, 0, 0, 4, 
	4, 3, 4, 2, 4, 1, 1, 2, 
	4, 1, 1, 1, 0, 4, 4, 3, 
	4, 2, 4, 1, 1, 2, 4, 1, 
	1, 1, 0, 0, 0, 0, 0, 1, 
	1, 4, 4, 2, 4, 1, 1, 0, 
	0, 2, 0, 1, 1, 0, 1, 1, 
	0, 0, 0, 0, 0, 1, 1, 1, 
	1, 3, 4, 3, 4, 3, 4, 4, 
	4, 2, 4, 1, 1, 0, 0, 2, 
	0, 1, 1, 0, 1, 2, 4, 1, 
	1, 0, 0, 2, 0, 1, 1, 1, 
	2, 4, 1, 1, 0, 0, 2, 0, 
	1, 1, 1, 2, 4, 1, 1, 0, 
	0, 2, 0, 1, 1, 1, 1, 1, 
	0, 4, 4, 2, 4, 1, 1, 0, 
	0, 2, 0, 1, 1, 0, 1, 0, 
	0, 0, 4, 4, 2, 4, 1, 1, 
	0, 0, 2, 0, 1, 1, 0, 1, 
	0, 0, 0, 4, 4, 5, 4, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 0, 1, 0, 1, 0, 1, 0, 
	1, 0, 1, 2, 4, 1, 1, 0, 
	0, 2, 0, 1, 1, 1, 2, 4, 
	1, 1, 0, 0, 2, 0, 1, 1, 
	0, 1, 0, 0, 3, 3, 3, 0, 
	3, 2, 3, 1, 1, 1, 2, 0, 
	1, 1, 3, 1, 0, 3, 3, 3, 
	0, 3, 2, 3, 1, 1, 1, 2, 
	0, 1, 1, 3, 1, 1, 1, 0, 
	0, 1, 1, 1, 1, 1, 1, 4, 
	4, 2, 4, 1, 1, 0, 0, 2, 
	0, 1, 1, 0, 1, 1, 1, 1, 
	0, 0, 0, 3, 3, 3, 0, 3, 
	2, 3, 1, 1, 1, 2, 0, 1, 
	1, 3, 1, 1, 0, 0, 1, 1, 
	0, 0, 0, 1, 1, 3, 3, 3, 
	1, 1, 1, 1, 1, 0, 0, 0, 
	1, 1, 0, 1, 1, 2, 4, 1, 
	1, 1, 1, 0, 0, 2, 0, 1, 
	1, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 1, 
	1, 1, 0, 1, 1, 1, 0, 0, 
	0, 0, 0, 0, 0, 3, 4, 4, 
	4, 2, 4, 1, 1, 0, 0, 2, 
	0, 1, 1, 0, 1, 2, 4, 1, 
	1, 0, 0, 2, 0, 1, 1, 1, 
	0, 0, 0, 0, 0, 0, 4, 4, 
	2, 4, 1, 1, 0, 0, 2, 0, 
	1, 1, 0, 1, 0, 0, 0, 1, 
	1, 0, 1, 0, 3
};

static const short _zone_index_offsets[] = {
	0, 0, 12, 38, 54, 79, 88, 96, 
	99, 101, 104, 106, 109, 111, 121, 130, 
	133, 136, 146, 155, 158, 160, 163, 165, 
	168, 170, 173, 176, 179, 186, 196, 209, 
	222, 236, 238, 241, 243, 246, 248, 258, 
	268, 277, 280, 282, 285, 287, 301, 315, 
	318, 321, 324, 327, 330, 337, 351, 365, 
	369, 383, 385, 387, 391, 400, 404, 408, 
	410, 412, 421, 424, 427, 430, 435, 442, 
	454, 469, 475, 489, 493, 507, 509, 511, 
	515, 530, 532, 534, 537, 544, 556, 571, 
	577, 591, 595, 609, 611, 613, 617, 632, 
	634, 636, 639, 642, 649, 672, 675, 682, 
	690, 698, 712, 726, 730, 744, 746, 748, 
	752, 761, 765, 769, 771, 773, 782, 785, 
	788, 793, 796, 799, 802, 809, 817, 825, 
	833, 841, 853, 865, 877, 889, 901, 913, 
	927, 941, 945, 959, 961, 963, 967, 976, 
	980, 984, 986, 988, 997, 1000, 1004, 1016, 
	1018, 1020, 1024, 1031, 1035, 1039, 1041, 1043, 
	1046, 1050, 1062, 1064, 1066, 1070, 1077, 1081, 
	1085, 1087, 1089, 1092, 1096, 1108, 1110, 1112, 
	1116, 1123, 1127, 1131, 1133, 1135, 1138, 1141, 
	1144, 1151, 1165, 1179, 1183, 1197, 1199, 1201, 
	1205, 1214, 1218, 1222, 1224, 1226, 1235, 1238, 
	1241, 1244, 1251, 1265, 1279, 1283, 1297, 1299, 
	1301, 1305, 1314, 1318, 1322, 1324, 1326, 1335, 
	1338, 1345, 1348, 1355, 1369, 1381, 1395, 1407, 
	1415, 1431, 1439, 1455, 1463, 1479, 1487, 1503, 
	1511, 1529, 1538, 1541, 1548, 1551, 1558, 1561, 
	1568, 1571, 1578, 1581, 1585, 1597, 1599, 1601, 
	1605, 1612, 1616, 1620, 1622, 1624, 1627, 1631, 
	1643, 1645, 1647, 1651, 1658, 1662, 1666, 1668, 
	1670, 1677, 1680, 1683, 1691, 1703, 1717, 1731, 
	1735, 1749, 1753, 1767, 1770, 1772, 1774, 1778, 
	1782, 1784, 1786, 1800, 1803, 1810, 1822, 1836, 
	1850, 1854, 1868, 1872, 1886, 1889, 1891, 1893, 
	1897, 1901, 1903, 1905, 1919, 1922, 1925, 1928, 
	1931, 1938, 1946, 1954, 1962, 1970, 1978, 1986, 
	2000, 2014, 2018, 2032, 2034, 2036, 2040, 2049, 
	2053, 2057, 2059, 2061, 2070, 2073, 2076, 2079, 
	2082, 2087, 2090, 2097, 2109, 2123, 2137, 2141, 
	2155, 2159, 2173, 2176, 2178, 2180, 2184, 2188, 
	2190, 2192, 2206, 2209, 2212, 2215, 2218, 2220, 
	2228, 2236, 2238, 2245, 2253, 2263, 2273, 2277, 
	2289, 2292, 2295, 2298, 2301, 2304, 2311, 2314, 
	2321, 2345, 2361, 2368, 2371, 2374, 2378, 2390, 
	2392, 2394, 2422, 2425, 2429, 2436, 2440, 2444, 
	2446, 2448, 2457, 2460, 2463, 2466, 2468, 2471, 
	2474, 2477, 2480, 2483, 2485, 2488, 2491, 2498, 
	2506, 2516, 2519, 2526, 2534, 2544, 2547, 2550, 
	2553, 2556, 2559, 2562, 2565, 2572, 2584, 2598, 
	2612, 2626, 2630, 2644, 2646, 2648, 2652, 2661, 
	2665, 2669, 2671, 2673, 2682, 2685, 2689, 2703, 
	2705, 2707, 2711, 2720, 2724, 2728, 2730, 2732, 
	2735, 2738, 2741, 2744, 2747, 2750, 2757, 2771, 
	2785, 2789, 2803, 2805, 2807, 2811, 2820, 2824, 
	2828, 2830, 2832, 2841, 2844, 2847, 2850, 2857, 
	2865, 2883, 2892, 2895, 2902
};

static const short _zone_trans_targs[] = {
	2, 2, 2, 2, 373, 2, 372, 1, 
	1, 1, 1, 0, 2, 2, 2, 2, 
	5, 48, 67, 366, 101, 120, 199, 216, 
	328, 5, 48, 67, 366, 101, 120, 199, 
	216, 328, 2, 372, 3, 0, 4, 4, 
	4, 4, 365, 365, 365, 365, 365, 365, 
	365, 365, 4, 364, 3, 0, 4, 4, 
	4, 4, 5, 48, 67, 98, 101, 120, 
	199, 216, 328, 5, 48, 67, 98, 101, 
	120, 199, 216, 328, 4, 364, 0, 6, 
	6, 6, 6, 26, 26, 6, 25, 0, 
	6, 6, 6, 6, 6, 25, 7, 0, 
	8, 23, 0, 9, 0, 10, 21, 0, 
	11, 0, 12, 19, 0, 13, 0, 14, 
	14, 14, 14, 468, 15, 14, 16, 17, 
	0, 14, 14, 14, 14, 468, 15, 14, 
	16, 0, 468, 15, 0, 14, 16, 0, 
	14, 14, 14, 14, 468, 15, 14, 16, 
	18, 0, 14, 14, 14, 14, 468, 15, 
	14, 16, 0, 12, 20, 0, 12, 0, 
	10, 22, 0, 10, 0, 8, 24, 0, 
	8, 0, 6, 25, 0, 27, 27, 0, 
	28, 28, 0, 29, 29, 29, 29, 29, 
	47, 0, 29, 29, 29, 29, 29, 47, 
	30, 30, 30, 0, 14, 14, 14, 14, 
	31, 468, 15, 14, 16, 30, 30, 30, 
	0, 14, 14, 14, 14, 31, 468, 15, 
	14, 16, 32, 30, 30, 0, 14, 14, 
	14, 14, 33, 31, 468, 15, 14, 16, 
	45, 30, 30, 0, 34, 0, 35, 43, 
	0, 36, 0, 37, 41, 0, 38, 0, 
	14, 14, 14, 14, 468, 15, 14, 16, 
	39, 0, 14, 14, 14, 14, 468, 15, 
	14, 16, 40, 0, 14, 14, 14, 14, 
	468, 15, 14, 16, 0, 37, 42, 0, 
	37, 0, 35, 44, 0, 35, 0, 14, 
	14, 14, 14, 33, 31, 468, 15, 14, 
	16, 46, 30, 30, 0, 14, 14, 14, 
	14, 33, 31, 468, 15, 14, 16, 30, 
	30, 30, 0, 29, 47, 0, 49, 49, 
	0, 50, 50, 0, 51, 51, 0, 52, 
	52, 0, 53, 53, 53, 53, 53, 66, 
	0, 53, 53, 59, 53, 53, 65, 55, 
	53, 66, 54, 54, 54, 54, 0, 14, 
	14, 14, 14, 55, 468, 15, 14, 16, 
	54, 54, 54, 54, 0, 56, 57, 0, 
	54, 14, 14, 14, 14, 55, 468, 15, 
	14, 16, 54, 54, 54, 54, 0, 58, 
	0, 54, 0, 59, 60, 61, 59, 14, 
	14, 14, 14, 468, 15, 14, 16, 0, 
	62, 63, 0, 59, 59, 60, 61, 59, 
	64, 0, 59, 0, 14, 14, 14, 14, 
	468, 15, 14, 16, 0, 53, 66, 0, 
	68, 68, 0, 69, 69, 0, 70, 84, 
	70, 84, 0, 71, 71, 71, 71, 71, 
	83, 0, 71, 71, 71, 71, 79, 71, 
	83, 72, 72, 72, 72, 0, 14, 14, 
	73, 14, 14, 79, 468, 15, 14, 16, 
	72, 72, 72, 72, 0, 0, 75, 0, 
	0, 0, 74, 14, 14, 14, 14, 75, 
	468, 15, 14, 16, 74, 74, 74, 74, 
	0, 76, 77, 0, 74, 14, 14, 14, 
	14, 75, 468, 15, 14, 16, 74, 74, 
	74, 74, 0, 78, 0, 74, 0, 80, 
	81, 0, 72, 14, 14, 73, 14, 14, 
	79, 468, 15, 14, 16, 72, 72, 72, 
	72, 0, 82, 0, 72, 0, 71, 83, 
	0, 85, 85, 85, 85, 85, 97, 0, 
	85, 85, 85, 85, 93, 85, 97, 86, 
	86, 86, 86, 0, 14, 14, 87, 14, 
	14, 93, 468, 15, 14, 16, 86, 86, 
	86, 86, 0, 0, 89, 0, 0, 0, 
	88, 14, 14, 14, 14, 89, 468, 15, 
	14, 16, 88, 88, 88, 88, 0, 90, 
	91, 0, 88, 14, 14, 14, 14, 89, 
	468, 15, 14, 16, 88, 88, 88, 88, 
	0, 92, 0, 88, 0, 94, 95, 0, 
	86, 14, 14, 87, 14, 14, 93, 468, 
	15, 14, 16, 86, 86, 86, 86, 0, 
	96, 0, 86, 0, 85, 97, 0, 99, 
	99, 0, 100, 100, 100, 100, 100, 363, 
	0, 100, 100, 100, 100, 5, 48, 67, 
	101, 120, 199, 216, 328, 5, 48, 67, 
	101, 120, 199, 216, 328, 100, 363, 0, 
	102, 102, 0, 103, 103, 103, 103, 103, 
	119, 0, 103, 103, 103, 103, 103, 119, 
	104, 0, 105, 105, 105, 105, 105, 118, 
	104, 0, 105, 105, 111, 105, 105, 117, 
	107, 105, 118, 106, 106, 106, 106, 0, 
	14, 14, 14, 14, 107, 468, 15, 14, 
	16, 106, 106, 106, 106, 0, 108, 109, 
	0, 106, 14, 14, 14, 14, 107, 468, 
	15, 14, 16, 106, 106, 106, 106, 0, 
	110, 0, 106, 0, 111, 112, 113, 111, 
	14, 14, 14, 14, 468, 15, 14, 16, 
	0, 114, 115, 0, 111, 111, 112, 113, 
	111, 116, 0, 111, 0, 14, 14, 14, 
	14, 468, 15, 14, 16, 0, 105, 118, 
	0, 103, 119, 0, 121, 184, 121, 184, 
	0, 122, 122, 0, 123, 123, 0, 124, 
	124, 0, 125, 125, 125, 125, 125, 183, 
	0, 125, 125, 125, 125, 125, 183, 126, 
	0, 127, 127, 127, 127, 127, 182, 126, 
	0, 127, 127, 127, 127, 127, 182, 128, 
	0, 129, 129, 129, 129, 129, 181, 128, 
	0, 129, 129, 175, 129, 129, 171, 129, 
	181, 130, 130, 130, 0, 131, 131, 131, 
	131, 171, 131, 170, 130, 130, 130, 130, 
	0, 131, 131, 164, 131, 131, 160, 131, 
	170, 132, 132, 132, 0, 133, 133, 133, 
	133, 160, 133, 159, 132, 132, 132, 132, 
	0, 133, 133, 153, 133, 133, 149, 133, 
	159, 134, 134, 134, 0, 135, 135, 135, 
	135, 149, 135, 148, 134, 134, 134, 134, 
	0, 135, 135, 141, 135, 135, 147, 137, 
	135, 148, 136, 136, 136, 136, 0, 14, 
	14, 14, 14, 137, 468, 15, 14, 16, 
	136, 136, 136, 136, 0, 138, 139, 0, 
	136, 14, 14, 14, 14, 137, 468, 15, 
	14, 16, 136, 136, 136, 136, 0, 140, 
	0, 136, 0, 141, 142, 143, 141, 14, 
	14, 14, 14, 468, 15, 14, 16, 0, 
	144, 145, 0, 141, 141, 142, 143, 141, 
	146, 0, 141, 0, 14, 14, 14, 14, 
	468, 15, 14, 16, 0, 135, 148, 0, 
	150, 151, 0, 134, 135, 135, 135, 135, 
	149, 135, 148, 134, 134, 134, 134, 0, 
	152, 0, 134, 0, 153, 154, 155, 153, 
	135, 135, 135, 135, 135, 148, 0, 156, 
	157, 0, 153, 153, 154, 155, 153, 158, 
	0, 153, 0, 133, 159, 0, 161, 162, 
	0, 132, 133, 133, 133, 133, 160, 133, 
	159, 132, 132, 132, 132, 0, 163, 0, 
	132, 0, 164, 165, 166, 164, 133, 133, 
	133, 133, 133, 159, 0, 167, 168, 0, 
	164, 164, 165, 166, 164, 169, 0, 164, 
	0, 131, 170, 0, 172, 173, 0, 130, 
	131, 131, 131, 131, 171, 131, 170, 130, 
	130, 130, 130, 0, 174, 0, 130, 0, 
	175, 176, 177, 175, 131, 131, 131, 131, 
	131, 170, 0, 178, 179, 0, 175, 175, 
	176, 177, 175, 180, 0, 175, 0, 129, 
	181, 0, 127, 182, 0, 125, 183, 0, 
	185, 185, 185, 185, 185, 198, 0, 185, 
	185, 191, 185, 185, 197, 187, 185, 198, 
	186, 186, 186, 186, 0, 14, 14, 14, 
	14, 187, 468, 15, 14, 16, 186, 186, 
	186, 186, 0, 188, 189, 0, 186, 14, 
	14, 14, 14, 187, 468, 15, 14, 16, 
	186, 186, 186, 186, 0, 190, 0, 186, 
	0, 191, 192, 193, 191, 14, 14, 14, 
	14, 468, 15, 14, 16, 0, 194, 195, 
	0, 191, 191, 192, 193, 191, 196, 0, 
	191, 0, 14, 14, 14, 14, 468, 15, 
	14, 16, 0, 185, 198, 0, 200, 200, 
	0, 201, 201, 0, 202, 202, 202, 202, 
	202, 215, 0, 202, 202, 208, 202, 202, 
	214, 204, 202, 215, 203, 203, 203, 203, 
	0, 14, 14, 14, 14, 204, 468, 15, 
	14, 16, 203, 203, 203, 203, 0, 205, 
	206, 0, 203, 14, 14, 14, 14, 204, 
	468, 15, 14, 16, 203, 203, 203, 203, 
	0, 207, 0, 203, 0, 208, 209, 210, 
	208, 14, 14, 14, 14, 468, 15, 14, 
	16, 0, 211, 212, 0, 208, 208, 209, 
	210, 208, 213, 0, 208, 0, 14, 14, 
	14, 14, 468, 15, 14, 16, 0, 202, 
	215, 0, 217, 266, 303, 217, 266, 303, 
	0, 218, 218, 0, 219, 219, 219, 219, 
	219, 265, 0, 219, 219, 258, 219, 219, 
	264, 254, 219, 265, 220, 220, 220, 220, 
	0, 221, 221, 221, 221, 254, 221, 253, 
	220, 220, 220, 220, 0, 221, 221, 247, 
	221, 221, 243, 221, 253, 222, 222, 222, 
	222, 222, 0, 223, 223, 223, 223, 243, 
	223, 242, 222, 222, 222, 222, 0, 223, 
	223, 223, 223, 223, 242, 224, 0, 225, 
	225, 225, 225, 241, 241, 241, 241, 241, 
	241, 241, 241, 225, 240, 224, 0, 225, 
	225, 225, 225, 225, 240, 226, 0, 227, 
	227, 227, 227, 239, 239, 239, 239, 239, 
	239, 239, 239, 227, 238, 226, 0, 227, 
	227, 227, 227, 227, 238, 228, 0, 229, 
	229, 229, 229, 237, 237, 237, 237, 237, 
	237, 237, 237, 229, 236, 228, 0, 229, 
	229, 229, 229, 229, 236, 230, 0, 231, 
	231, 231, 231, 235, 235, 235, 235, 235, 
	235, 235, 235, 231, 234, 230, 0, 231, 
	231, 231, 231, 231, 234, 232, 0, 14, 
	14, 14, 14, 233, 233, 233, 233, 233, 
	233, 233, 233, 468, 15, 14, 16, 232, 
	0, 14, 14, 14, 14, 468, 15, 14, 
	16, 0, 231, 234, 0, 231, 231, 231, 
	231, 231, 234, 0, 229, 236, 0, 229, 
	229, 229, 229, 229, 236, 0, 227, 238, 
	0, 227, 227, 227, 227, 227, 238, 0, 
	225, 240, 0, 225, 225, 225, 225, 225, 
	240, 0, 223, 242, 0, 244, 245, 0, 
	222, 223, 223, 223, 223, 243, 223, 242, 
	222, 222, 222, 222, 0, 246, 0, 222, 
	0, 247, 248, 249, 247, 223, 223, 223, 
	223, 223, 242, 0, 250, 251, 0, 247, 
	247, 248, 249, 247, 252, 0, 247, 0, 
	221, 253, 0, 255, 256, 0, 220, 221, 
	221, 221, 221, 254, 221, 253, 220, 220, 
	220, 220, 0, 257, 0, 220, 0, 258, 
	259, 260, 258, 221, 221, 221, 221, 221, 
	253, 0, 261, 262, 0, 258, 258, 259, 
	260, 258, 263, 0, 258, 0, 221, 221, 
	221, 221, 221, 253, 0, 219, 265, 0, 
	267, 267, 0, 268, 268, 268, 268, 284, 
	268, 302, 0, 282, 282, 271, 282, 282, 
	273, 282, 283, 269, 269, 269, 0, 270, 
	270, 271, 270, 270, 273, 468, 15, 270, 
	275, 269, 269, 269, 0, 270, 270, 271, 
	270, 270, 273, 468, 15, 270, 275, 269, 
	269, 269, 0, 271, 272, 278, 271, 270, 
	270, 271, 270, 270, 273, 468, 15, 270, 
	275, 269, 269, 269, 0, 274, 276, 0, 
	269, 270, 270, 271, 270, 270, 273, 468, 
	15, 270, 275, 269, 269, 269, 0, 270, 
	275, 0, 277, 0, 269, 0, 279, 280, 
	0, 271, 271, 272, 278, 271, 281, 0, 
	271, 0, 282, 282, 271, 282, 282, 273, 
	468, 15, 282, 283, 269, 269, 269, 0, 
	282, 283, 0, 285, 285, 285, 285, 285, 
	301, 0, 299, 299, 288, 299, 299, 290, 
	299, 300, 286, 286, 286, 0, 287, 287, 
	288, 287, 287, 290, 468, 15, 287, 292, 
	286, 286, 286, 0, 287, 287, 288, 287, 
	287, 290, 468, 15, 287, 292, 286, 286, 
	286, 0, 288, 289, 295, 288, 287, 287, 
	288, 287, 287, 290, 468, 15, 287, 292, 
	286, 286, 286, 0, 291, 293, 0, 286, 
	287, 287, 288, 287, 287, 290, 468, 15, 
	287, 292, 286, 286, 286, 0, 287, 292, 
	0, 294, 0, 286, 0, 296, 297, 0, 
	288, 288, 289, 295, 288, 298, 0, 288, 
	0, 299, 299, 288, 299, 299, 290, 468, 
	15, 299, 300, 286, 286, 286, 0, 299, 
	300, 0, 285, 301, 0, 268, 302, 0, 
	304, 304, 0, 305, 305, 305, 305, 305, 
	327, 0, 305, 305, 305, 305, 305, 327, 
	306, 0, 307, 307, 307, 307, 307, 326, 
	306, 0, 307, 307, 307, 307, 307, 326, 
	308, 0, 309, 309, 309, 309, 309, 325, 
	308, 0, 309, 309, 309, 309, 309, 325, 
	310, 0, 311, 311, 311, 311, 311, 324, 
	310, 0, 311, 311, 317, 311, 311, 323, 
	313, 311, 324, 312, 312, 312, 312, 0, 
	14, 14, 14, 14, 313, 468, 15, 14, 
	16, 312, 312, 312, 312, 0, 314, 315, 
	0, 312, 14, 14, 14, 14, 313, 468, 
	15, 14, 16, 312, 312, 312, 312, 0, 
	316, 0, 312, 0, 317, 318, 319, 317, 
	14, 14, 14, 14, 468, 15, 14, 16, 
	0, 320, 321, 0, 317, 317, 318, 319, 
	317, 322, 0, 317, 0, 14, 14, 14, 
	14, 468, 15, 14, 16, 0, 311, 324, 
	0, 309, 325, 0, 307, 326, 0, 305, 
	327, 0, 329, 348, 329, 348, 0, 330, 
	330, 0, 331, 331, 331, 331, 331, 347, 
	0, 345, 345, 334, 345, 345, 336, 345, 
	346, 332, 332, 332, 0, 333, 333, 334, 
	333, 333, 336, 468, 15, 333, 338, 332, 
	332, 332, 0, 333, 333, 334, 333, 333, 
	336, 468, 15, 333, 338, 332, 332, 332, 
	0, 334, 335, 341, 334, 333, 333, 334, 
	333, 333, 336, 468, 15, 333, 338, 332, 
	332, 332, 0, 337, 339, 0, 332, 333, 
	333, 334, 333, 333, 336, 468, 15, 333, 
	338, 332, 332, 332, 0, 333, 338, 0, 
	340, 0, 332, 0, 342, 343, 0, 334, 
	334, 335, 341, 334, 344, 0, 334, 0, 
	345, 345, 334, 345, 345, 336, 468, 15, 
	345, 346, 332, 332, 332, 0, 345, 346, 
	0, 331, 347, 0, 349, 349, 0, 350, 
	350, 0, 351, 0, 352, 352, 352, 352, 
	352, 362, 351, 0, 352, 352, 352, 352, 
	353, 352, 362, 0, 354, 0, 355, 355, 
	355, 355, 355, 361, 0, 355, 355, 355, 
	355, 355, 361, 356, 0, 357, 357, 357, 
	357, 468, 15, 357, 360, 356, 0, 357, 
	357, 357, 357, 357, 360, 358, 358, 358, 
	0, 359, 359, 359, 0, 357, 357, 357, 
	357, 468, 15, 357, 360, 358, 358, 358, 
	0, 357, 360, 0, 355, 361, 0, 352, 
	362, 0, 100, 363, 0, 4, 364, 0, 
	4, 4, 4, 4, 4, 364, 0, 367, 
	367, 0, 368, 368, 368, 368, 368, 371, 
	0, 368, 368, 368, 368, 5, 48, 67, 
	101, 120, 199, 216, 328, 5, 48, 67, 
	101, 120, 199, 216, 328, 368, 371, 369, 
	0, 100, 100, 100, 100, 370, 370, 370, 
	370, 370, 370, 370, 370, 100, 363, 369, 
	0, 100, 100, 100, 100, 100, 363, 0, 
	368, 371, 0, 2, 372, 0, 374, 375, 
	0, 1, 2, 2, 2, 2, 373, 2, 
	372, 1, 1, 1, 1, 0, 376, 0, 
	1, 0, 377, 377, 377, 377, 5, 48, 
	67, 366, 101, 120, 199, 216, 328, 5, 
	48, 67, 366, 101, 120, 199, 216, 328, 
	468, 15, 377, 378, 3, 0, 377, 378, 
	0, 379, 380, 381, 379, 2, 2, 2, 
	2, 2, 372, 0, 382, 383, 0, 379, 
	379, 380, 381, 379, 384, 0, 379, 0, 
	386, 406, 440, 460, 386, 406, 440, 460, 
	0, 387, 387, 0, 388, 388, 0, 389, 
	389, 0, 390, 0, 391, 391, 0, 392, 
	392, 0, 393, 393, 0, 394, 394, 0, 
	395, 395, 0, 396, 0, 397, 397, 0, 
	398, 402, 0, 399, 399, 399, 399, 399, 
	401, 0, 399, 399, 399, 399, 399, 401, 
	400, 0, 14, 14, 14, 14, 468, 15, 
	14, 16, 400, 0, 399, 401, 0, 403, 
	403, 403, 403, 403, 405, 0, 403, 403, 
	403, 403, 403, 405, 404, 0, 14, 14, 
	14, 14, 468, 15, 14, 16, 404, 0, 
	403, 405, 0, 407, 407, 0, 408, 408, 
	0, 409, 409, 0, 410, 410, 0, 411, 
	411, 0, 412, 412, 0, 413, 413, 413, 
	413, 413, 439, 0, 413, 413, 433, 413, 
	413, 429, 413, 439, 414, 414, 414, 0, 
	415, 415, 415, 415, 429, 468, 15, 415, 
	428, 414, 414, 414, 414, 0, 415, 415, 
	421, 415, 415, 427, 417, 415, 428, 416, 
	416, 416, 416, 0, 14, 14, 14, 14, 
	417, 468, 15, 14, 16, 416, 416, 416, 
	416, 0, 418, 419, 0, 416, 14, 14, 
	14, 14, 417, 468, 15, 14, 16, 416, 
	416, 416, 416, 0, 420, 0, 416, 0, 
	421, 422, 423, 421, 14, 14, 14, 14, 
	468, 15, 14, 16, 0, 424, 425, 0, 
	421, 421, 422, 423, 421, 426, 0, 421, 
	0, 14, 14, 14, 14, 468, 15, 14, 
	16, 0, 415, 428, 0, 430, 431, 0, 
	414, 415, 415, 415, 415, 429, 468, 15, 
	415, 428, 414, 414, 414, 414, 0, 432, 
	0, 414, 0, 433, 434, 435, 433, 415, 
	415, 415, 415, 468, 15, 415, 428, 0, 
	436, 437, 0, 433, 433, 434, 435, 433, 
	438, 0, 433, 0, 413, 439, 0, 441, 
	441, 0, 442, 442, 0, 443, 443, 0, 
	444, 444, 0, 445, 445, 0, 446, 446, 
	446, 446, 446, 459, 0, 446, 446, 452, 
	446, 446, 458, 448, 446, 459, 447, 447, 
	447, 447, 0, 14, 14, 14, 14, 448, 
	468, 15, 14, 16, 447, 447, 447, 447, 
	0, 449, 450, 0, 447, 14, 14, 14, 
	14, 448, 468, 15, 14, 16, 447, 447, 
	447, 447, 0, 451, 0, 447, 0, 452, 
	453, 454, 452, 14, 14, 14, 14, 468, 
	15, 14, 16, 0, 455, 456, 0, 452, 
	452, 453, 454, 452, 457, 0, 452, 0, 
	14, 14, 14, 14, 468, 15, 14, 16, 
	0, 446, 459, 0, 461, 461, 0, 462, 
	462, 0, 463, 463, 463, 463, 463, 466, 
	0, 463, 463, 463, 463, 463, 466, 464, 
	0, 14, 14, 14, 14, 465, 465, 465, 
	465, 465, 465, 465, 465, 468, 15, 14, 
	16, 464, 0, 14, 14, 14, 14, 468, 
	15, 14, 16, 0, 463, 466, 0, 2, 
	2, 2, 2, 2, 372, 0, 377, 377, 
	379, 385, 377, 377, 467, 373, 468, 15, 
	377, 378, 1, 1, 1, 0, 0
};

static const short _zone_trans_actions[] = {
	5, 5, 72, 75, 0, 78, 5, 0, 
	0, 0, 0, 0, 0, 0, 53, 55, 
	37, 37, 37, 37, 37, 37, 37, 37, 
	37, 37, 37, 37, 37, 37, 37, 37, 
	37, 37, 57, 0, 348, 0, 303, 303, 
	794, 798, 300, 300, 300, 300, 300, 300, 
	300, 300, 802, 303, 0, 0, 0, 0, 
	53, 55, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	0, 53, 55, 0, 0, 57, 0, 0, 
	0, 0, 53, 55, 57, 0, 1, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 294, 
	294, 770, 774, 778, 294, 778, 294, 0, 
	0, 0, 0, 53, 55, 57, 0, 57, 
	0, 0, 57, 0, 0, 57, 0, 0, 
	294, 294, 770, 774, 778, 294, 778, 294, 
	0, 0, 294, 294, 770, 774, 778, 294, 
	778, 294, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 57, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 1, 1, 0, 297, 297, 782, 786, 
	0, 790, 297, 790, 297, 0, 0, 0, 
	0, 297, 297, 782, 786, 0, 790, 297, 
	790, 297, 0, 0, 0, 0, 297, 297, 
	782, 786, 0, 0, 790, 297, 790, 297, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	297, 297, 782, 786, 790, 297, 790, 297, 
	0, 0, 297, 297, 782, 786, 790, 297, 
	790, 297, 0, 0, 297, 297, 782, 786, 
	790, 297, 790, 297, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 297, 
	297, 782, 786, 0, 0, 790, 297, 790, 
	297, 0, 0, 0, 0, 297, 297, 782, 
	786, 0, 0, 790, 297, 790, 297, 0, 
	0, 0, 0, 57, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 0, 0, 1, 53, 55, 1, 1, 
	57, 0, 1, 1, 1, 1, 0, 132, 
	132, 582, 586, 0, 590, 132, 590, 132, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 970, 970, 1147, 1152, 59, 1157, 970, 
	1157, 970, 59, 59, 59, 59, 0, 0, 
	0, 0, 0, 57, 0, 0, 0, 165, 
	165, 678, 682, 686, 165, 686, 165, 0, 
	0, 0, 0, 0, 432, 59, 59, 59, 
	0, 0, 0, 0, 99, 99, 486, 490, 
	494, 99, 494, 99, 0, 57, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 0, 0, 53, 55, 1, 57, 
	0, 1, 1, 1, 1, 0, 390, 390, 
	0, 922, 926, 0, 930, 390, 930, 390, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 390, 390, 922, 926, 0, 
	930, 390, 930, 390, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 1070, 1070, 1242, 
	1247, 59, 1252, 1070, 1252, 1070, 59, 59, 
	59, 59, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 1070, 1070, 59, 1242, 1247, 
	59, 1252, 1070, 1252, 1070, 59, 59, 59, 
	59, 0, 0, 0, 0, 0, 57, 0, 
	0, 0, 0, 53, 55, 57, 0, 0, 
	0, 0, 53, 55, 1, 57, 0, 1, 
	1, 1, 1, 0, 393, 393, 0, 934, 
	938, 0, 942, 393, 942, 393, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 393, 393, 934, 938, 0, 942, 393, 
	942, 393, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 1074, 1074, 1257, 1262, 59, 
	1267, 1074, 1267, 1074, 59, 59, 59, 59, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 1074, 1074, 59, 1257, 1262, 59, 1267, 
	1074, 1267, 1074, 59, 59, 59, 59, 0, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 0, 0, 53, 55, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 57, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 0, 309, 309, 818, 822, 826, 309, 
	0, 0, 0, 0, 1, 53, 55, 1, 
	1, 57, 0, 1, 1, 1, 1, 0, 
	138, 138, 606, 610, 0, 614, 138, 614, 
	138, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 978, 978, 1177, 1182, 59, 1187, 
	978, 1187, 978, 59, 59, 59, 59, 0, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	171, 171, 702, 706, 710, 171, 710, 171, 
	0, 0, 0, 0, 0, 432, 59, 59, 
	59, 0, 0, 0, 0, 105, 105, 510, 
	514, 518, 105, 518, 105, 0, 57, 0, 
	0, 57, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 0, 0, 53, 55, 57, 0, 1, 
	0, 309, 309, 818, 822, 826, 309, 0, 
	0, 0, 0, 53, 55, 57, 0, 1, 
	0, 312, 312, 830, 834, 838, 312, 0, 
	0, 0, 0, 231, 53, 55, 231, 57, 
	0, 231, 231, 231, 0, 29, 29, 276, 
	279, 0, 282, 29, 0, 0, 0, 0, 
	0, 0, 0, 1, 53, 55, 1, 57, 
	0, 1, 1, 1, 0, 29, 29, 276, 
	279, 0, 282, 29, 0, 0, 0, 0, 
	0, 0, 0, 1, 53, 55, 1, 57, 
	0, 1, 1, 1, 0, 29, 29, 276, 
	279, 0, 282, 29, 0, 0, 0, 0, 
	0, 0, 0, 1, 53, 55, 1, 1, 
	57, 0, 1, 1, 1, 1, 0, 144, 
	144, 630, 634, 0, 638, 144, 638, 144, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 986, 986, 1207, 1212, 59, 1217, 986, 
	1217, 986, 59, 59, 59, 59, 0, 0, 
	0, 0, 0, 57, 0, 0, 0, 177, 
	177, 726, 730, 734, 177, 734, 177, 0, 
	0, 0, 0, 0, 432, 59, 59, 59, 
	0, 0, 0, 0, 111, 111, 534, 538, 
	542, 111, 542, 111, 0, 57, 0, 0, 
	0, 0, 0, 0, 429, 429, 1058, 1062, 
	59, 1066, 429, 59, 59, 59, 59, 0, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	31, 31, 285, 288, 291, 31, 0, 0, 
	0, 0, 0, 432, 59, 59, 59, 0, 
	0, 0, 0, 57, 0, 0, 0, 0, 
	0, 0, 429, 429, 1058, 1062, 59, 1066, 
	429, 59, 59, 59, 59, 0, 0, 0, 
	0, 0, 57, 0, 0, 0, 31, 31, 
	285, 288, 291, 31, 0, 0, 0, 0, 
	0, 432, 59, 59, 59, 0, 0, 0, 
	0, 57, 0, 0, 0, 0, 0, 0, 
	429, 429, 1058, 1062, 59, 1066, 429, 59, 
	59, 59, 59, 0, 0, 0, 0, 0, 
	57, 0, 0, 0, 31, 31, 285, 288, 
	291, 31, 0, 0, 0, 0, 0, 432, 
	59, 59, 59, 0, 0, 0, 0, 57, 
	0, 0, 57, 0, 0, 57, 0, 0, 
	0, 0, 53, 55, 57, 0, 0, 0, 
	0, 1, 53, 55, 1, 1, 57, 0, 
	1, 1, 1, 1, 0, 129, 129, 570, 
	574, 0, 578, 129, 578, 129, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 966, 
	966, 1132, 1137, 59, 1142, 966, 1142, 966, 
	59, 59, 59, 59, 0, 0, 0, 0, 
	0, 57, 0, 0, 0, 162, 162, 666, 
	670, 674, 162, 674, 162, 0, 0, 0, 
	0, 0, 432, 59, 59, 59, 0, 0, 
	0, 0, 96, 96, 474, 478, 482, 96, 
	482, 96, 0, 57, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 53, 55, 
	57, 0, 0, 0, 0, 1, 53, 55, 
	1, 1, 57, 0, 1, 1, 1, 1, 
	0, 135, 135, 594, 598, 0, 602, 135, 
	602, 135, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 974, 974, 1162, 1167, 59, 
	1172, 974, 1172, 974, 59, 59, 59, 59, 
	0, 0, 0, 0, 0, 57, 0, 0, 
	0, 168, 168, 690, 694, 698, 168, 698, 
	168, 0, 0, 0, 0, 0, 432, 59, 
	59, 59, 0, 0, 0, 0, 102, 102, 
	498, 502, 506, 102, 506, 102, 0, 57, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 53, 55, 
	57, 0, 0, 0, 0, 1, 53, 55, 
	1, 1, 57, 0, 1, 1, 1, 1, 
	0, 11, 11, 147, 150, 0, 153, 11, 
	0, 0, 0, 0, 0, 0, 0, 1, 
	53, 55, 1, 57, 0, 1, 1, 1, 
	1, 1, 0, 15, 15, 189, 192, 0, 
	195, 15, 0, 0, 0, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 0, 309, 
	309, 818, 822, 300, 300, 300, 300, 300, 
	300, 300, 300, 826, 309, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 0, 312, 
	312, 830, 834, 300, 300, 300, 300, 300, 
	300, 300, 300, 838, 312, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 0, 315, 
	315, 842, 846, 300, 300, 300, 300, 300, 
	300, 300, 300, 850, 315, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 0, 318, 
	318, 854, 858, 300, 300, 300, 300, 300, 
	300, 300, 300, 862, 318, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 0, 866, 
	866, 1082, 1087, 300, 300, 300, 300, 300, 
	300, 300, 300, 1092, 866, 1092, 866, 0, 
	0, 387, 387, 910, 914, 918, 387, 918, 
	387, 0, 57, 0, 0, 45, 45, 378, 
	381, 384, 45, 0, 57, 0, 0, 43, 
	43, 369, 372, 375, 43, 0, 57, 0, 
	0, 41, 41, 360, 363, 366, 41, 0, 
	57, 0, 0, 39, 39, 351, 354, 357, 
	39, 0, 57, 0, 0, 0, 0, 0, 
	0, 420, 420, 1002, 1006, 59, 1010, 420, 
	59, 59, 59, 59, 0, 0, 0, 0, 
	0, 57, 0, 0, 0, 17, 17, 198, 
	201, 204, 17, 0, 0, 0, 0, 0, 
	432, 59, 59, 59, 0, 0, 0, 0, 
	57, 0, 0, 0, 0, 0, 0, 417, 
	417, 990, 994, 59, 998, 417, 59, 59, 
	59, 59, 0, 0, 0, 0, 0, 57, 
	0, 0, 0, 13, 13, 180, 183, 186, 
	13, 0, 0, 0, 0, 0, 432, 59, 
	59, 59, 0, 0, 0, 0, 9, 9, 
	114, 117, 120, 9, 0, 57, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 0, 
	57, 0, 0, 23, 23, 231, 405, 408, 
	231, 411, 23, 231, 231, 231, 0, 25, 
	25, 234, 246, 249, 0, 750, 240, 252, 
	25, 0, 0, 0, 0, 0, 0, 1, 
	53, 55, 1, 399, 49, 57, 0, 1, 
	1, 1, 0, 57, 0, 0, 0, 27, 
	27, 255, 267, 270, 255, 762, 261, 273, 
	27, 255, 255, 255, 0, 0, 0, 0, 
	0, 426, 426, 1030, 1046, 1050, 59, 1232, 
	1038, 1054, 426, 59, 59, 59, 0, 57, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 432, 59, 59, 59, 0, 0, 
	0, 0, 23, 23, 231, 405, 408, 231, 
	399, 49, 411, 23, 231, 231, 231, 0, 
	57, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 23, 23, 231, 405, 408, 231, 
	411, 23, 231, 231, 231, 0, 25, 25, 
	234, 246, 249, 0, 754, 243, 252, 25, 
	0, 0, 0, 0, 0, 0, 1, 53, 
	55, 1, 402, 51, 57, 0, 1, 1, 
	1, 0, 57, 0, 0, 0, 27, 27, 
	255, 267, 270, 255, 766, 264, 273, 27, 
	255, 255, 255, 0, 0, 0, 0, 0, 
	426, 426, 1030, 1046, 1050, 59, 1237, 1042, 
	1054, 426, 59, 59, 59, 0, 57, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 432, 59, 59, 59, 0, 0, 0, 
	0, 23, 23, 231, 405, 408, 231, 402, 
	51, 411, 23, 231, 231, 231, 0, 57, 
	0, 0, 57, 0, 0, 57, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 0, 309, 309, 818, 822, 826, 309, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 0, 312, 312, 830, 834, 838, 312, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 0, 315, 315, 842, 846, 850, 315, 
	0, 0, 0, 0, 1, 53, 55, 1, 
	1, 57, 0, 1, 1, 1, 1, 0, 
	141, 141, 618, 622, 0, 626, 141, 626, 
	141, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 982, 982, 1192, 1197, 59, 1202, 
	982, 1202, 982, 59, 59, 59, 59, 0, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	174, 174, 714, 718, 722, 174, 722, 174, 
	0, 0, 0, 0, 0, 432, 59, 59, 
	59, 0, 0, 0, 0, 108, 108, 522, 
	526, 530, 108, 530, 108, 0, 57, 0, 
	0, 57, 0, 0, 57, 0, 0, 57, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 23, 23, 231, 405, 408, 231, 411, 
	23, 231, 231, 231, 0, 25, 25, 234, 
	246, 249, 0, 746, 237, 252, 25, 0, 
	0, 0, 0, 0, 0, 1, 53, 55, 
	1, 396, 47, 57, 0, 1, 1, 1, 
	0, 57, 0, 0, 0, 27, 27, 255, 
	267, 270, 255, 758, 258, 273, 27, 255, 
	255, 255, 0, 0, 0, 0, 0, 426, 
	426, 1030, 1046, 1050, 59, 1227, 1034, 1054, 
	426, 59, 59, 59, 0, 57, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	432, 59, 59, 59, 0, 0, 0, 0, 
	23, 23, 231, 405, 408, 231, 396, 47, 
	411, 23, 231, 231, 231, 0, 57, 0, 
	0, 57, 0, 0, 0, 0, 0, 0, 
	0, 0, 1, 0, 309, 309, 818, 822, 
	826, 309, 0, 0, 0, 0, 53, 55, 
	0, 57, 0, 0, 0, 0, 0, 0, 
	53, 55, 57, 0, 0, 0, 0, 53, 
	55, 57, 0, 1, 0, 327, 327, 898, 
	902, 1097, 894, 906, 327, 0, 0, 0, 
	0, 53, 55, 57, 0, 1, 1, 1, 
	0, 0, 0, 0, 0, 61, 61, 441, 
	444, 1078, 438, 447, 61, 435, 435, 435, 
	0, 57, 0, 0, 57, 0, 0, 57, 
	0, 0, 57, 0, 0, 57, 0, 0, 
	33, 33, 330, 333, 336, 33, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 0, 0, 53, 55, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 57, 0, 1, 
	0, 303, 303, 794, 798, 300, 300, 300, 
	300, 300, 300, 300, 300, 802, 303, 0, 
	0, 33, 33, 330, 333, 336, 33, 0, 
	57, 0, 0, 57, 0, 0, 0, 0, 
	0, 0, 414, 414, 946, 950, 59, 954, 
	414, 59, 59, 59, 59, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 37, 37, 
	37, 37, 37, 37, 37, 37, 37, 37, 
	37, 37, 37, 37, 37, 37, 37, 37, 
	57, 0, 57, 0, 348, 0, 57, 0, 
	0, 57, 0, 0, 0, 7, 7, 81, 
	84, 87, 7, 0, 0, 0, 0, 0, 
	432, 59, 59, 59, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 53, 55, 57, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	1, 0, 321, 321, 870, 874, 878, 321, 
	878, 321, 0, 0, 57, 0, 0, 0, 
	0, 53, 55, 57, 0, 0, 0, 0, 
	53, 55, 57, 0, 1, 0, 324, 324, 
	882, 886, 890, 324, 890, 324, 0, 0, 
	57, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 53, 
	55, 57, 0, 0, 0, 0, 1, 53, 
	55, 1, 57, 0, 1, 1, 1, 0, 
	19, 19, 210, 213, 0, 738, 207, 216, 
	19, 0, 0, 0, 0, 0, 0, 0, 
	1, 53, 55, 1, 1, 57, 0, 1, 
	1, 1, 1, 0, 126, 126, 558, 562, 
	0, 566, 126, 566, 126, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 962, 962, 
	1117, 1122, 59, 1127, 962, 1127, 962, 59, 
	59, 59, 59, 0, 0, 0, 0, 0, 
	57, 0, 0, 0, 159, 159, 654, 658, 
	662, 159, 662, 159, 0, 0, 0, 0, 
	0, 432, 59, 59, 59, 0, 0, 0, 
	0, 93, 93, 462, 466, 470, 93, 470, 
	93, 0, 57, 0, 0, 0, 0, 0, 
	0, 423, 423, 1018, 1022, 59, 1222, 1014, 
	1026, 423, 59, 59, 59, 59, 0, 0, 
	0, 0, 0, 57, 0, 0, 0, 21, 
	21, 222, 225, 742, 219, 228, 21, 0, 
	0, 0, 0, 0, 432, 59, 59, 59, 
	0, 0, 0, 0, 57, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 0, 0, 0, 
	53, 55, 57, 0, 0, 0, 0, 1, 
	53, 55, 1, 1, 57, 0, 1, 1, 
	1, 1, 0, 123, 123, 546, 550, 0, 
	554, 123, 554, 123, 0, 0, 0, 0, 
	0, 0, 0, 0, 0, 958, 958, 1102, 
	1107, 59, 1112, 958, 1112, 958, 59, 59, 
	59, 59, 0, 0, 0, 0, 0, 57, 
	0, 0, 0, 156, 156, 642, 646, 650, 
	156, 650, 156, 0, 0, 0, 0, 0, 
	432, 59, 59, 59, 0, 0, 0, 0, 
	90, 90, 450, 454, 458, 90, 458, 90, 
	0, 57, 0, 0, 0, 0, 0, 0, 
	0, 0, 0, 0, 53, 55, 57, 0, 
	0, 0, 0, 53, 55, 57, 0, 1, 
	0, 306, 306, 806, 810, 300, 300, 300, 
	300, 300, 300, 300, 300, 814, 306, 814, 
	306, 0, 0, 35, 35, 339, 342, 345, 
	35, 345, 35, 0, 57, 0, 0, 3, 
	3, 63, 66, 69, 3, 0, 0, 0, 
	1, 0, 53, 55, 1, 1, 57, 0, 
	57, 0, 1, 1, 1, 0, 0
};

static const int zone_start = 468;
static const int zone_error = 0;

static const int zone_en_main = 468;


#line 692 "./zscan.rl"


/*
 * The longest possible tstart-based token is a maximum-length
 *  quoted TXT string, which given autosplit, can be up to
 *  roughly 64K, which means buffer size has to be that big
 *  to accomodate that.  There probably wouldn't be any bugs
 *  going down to something reasonable like 4K, but it would
 *  cause parse errors if anyone tried to use longer TXT strings.
 * Another important thing: for integers, we use atoi() directly
 *  on the buffer itself.  In the normal case this works because
 *  there is always some non-integer after it in the buffer, which
 *  halts atoi().  The corner case is if the last digit of an
 *  integer happened to be the last byte of the buffer.  This
 *  is why we allocate one extra buffer byte and set it to zero.
 */
#define BUFSIZE 65536

F_NONNULL
static void scanner(zscan_t* z, int fd) {
    dmn_assert(z);

#ifdef HAVE_POSIX_FADVISE
    {
        struct stat fdstat;
        if(fstat(fd, &fdstat))
            log_fatal("fstat() of '%s' failed: %s", z->curfn, logf_errno());
        posix_fadvise(fd, 0, fdstat.st_size, POSIX_FADV_SEQUENTIAL);
    }
#endif

    char* buf = malloc(BUFSIZE + 1);
    char* read_at;
    buf[BUFSIZE] = 0;

    const char* pe = NULL;
    const char* eof = NULL;
    int cs = zone_start;

    while(!eof) {
        unsigned have = 0;
        if(z->tstart != NULL) {
            dmn_assert(pe);
            dmn_assert(z->tstart < pe);
            dmn_assert(z->tstart != buf);
            have = pe - z->tstart;
            memmove(buf, z->tstart, have);
            z->tstart = buf;
        }

        const int space = BUFSIZE - have;
        const char* p = read_at = buf + have;

        const int len = read(fd, read_at, space);
        if(len < 0)
            parse_error("read() failed: %s", logf_errno());

        pe = p + len;

        if(len < space)
            eof = pe;

        
#line 2391 "./zscan_lm.c"
	{
	int _klen;
	unsigned int _trans;
	short _widec;
	const char *_acts;
	unsigned int _nacts;
	const short *_keys;

	if ( p == pe )
		goto _test_eof;
	if ( cs == 0 )
		goto _out;
_resume:
	_widec = (*p);
	_klen = _zone_cond_lengths[cs];
	_keys = _zone_cond_keys + (_zone_cond_offsets[cs]*2);
	if ( _klen > 0 ) {
		const short *_lower = _keys;
		const short *_mid;
		const short *_upper = _keys + (_klen<<1) - 2;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + (((_upper-_lower) >> 1) & ~1);
			if ( _widec < _mid[0] )
				_upper = _mid - 2;
			else if ( _widec > _mid[1] )
				_lower = _mid + 2;
			else {
				switch ( _zone_cond_spaces[_zone_cond_offsets[cs] + ((_mid - _keys)>>1)] ) {
	case 0: {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		break;
	}
				}
				break;
			}
		}
	}

	_keys = _zone_trans_keys + _zone_key_offsets[cs];
	_trans = _zone_index_offsets[cs];

	_klen = _zone_single_lengths[cs];
	if ( _klen > 0 ) {
		const short *_lower = _keys;
		const short *_mid;
		const short *_upper = _keys + _klen - 1;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + ((_upper-_lower) >> 1);
			if ( _widec < *_mid )
				_upper = _mid - 1;
			else if ( _widec > *_mid )
				_lower = _mid + 1;
			else {
				_trans += (unsigned int)(_mid - _keys);
				goto _match;
			}
		}
		_keys += _klen;
		_trans += _klen;
	}

	_klen = _zone_range_lengths[cs];
	if ( _klen > 0 ) {
		const short *_lower = _keys;
		const short *_mid;
		const short *_upper = _keys + (_klen<<1) - 2;
		while (1) {
			if ( _upper < _lower )
				break;

			_mid = _lower + (((_upper-_lower) >> 1) & ~1);
			if ( _widec < _mid[0] )
				_upper = _mid - 2;
			else if ( _widec > _mid[1] )
				_lower = _mid + 2;
			else {
				_trans += (unsigned int)((_mid - _keys)>>1);
				goto _match;
			}
		}
		_trans += _klen;
	}

_match:
	cs = _zone_trans_targs[_trans];

	if ( _zone_trans_actions[_trans] == 0 )
		goto _again;

	_acts = _zone_actions + _zone_trans_actions[_trans];
	_nacts = (unsigned int) *_acts++;
	while ( _nacts-- > 0 )
	{
		switch ( *_acts++ )
		{
	case 0:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	break;
	case 1:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
	break;
	case 2:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
	break;
	case 3:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
	break;
	case 4:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
	break;
	case 5:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
	break;
	case 6:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
	break;
	case 7:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
	break;
	case 8:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
	break;
	case 9:
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	break;
	case 10:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
	break;
	case 11:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
	break;
	case 12:
#line 485 "./zscan.rl"
	{ process_include(z); }
	break;
	case 13:
#line 487 "./zscan.rl"
	{ text_start(z); }
	break;
	case 14:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	break;
	case 15:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	break;
	case 16:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	break;
	case 17:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	break;
	case 18:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
	break;
	case 19:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
	break;
	case 20:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
	break;
	case 21:
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	break;
	case 22:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	break;
	case 23:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	break;
	case 24:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	break;
	case 25:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	break;
	case 26:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	break;
	case 27:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	break;
	case 28:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
	break;
	case 29:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
	break;
	case 30:
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
	break;
	case 31:
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
	break;
	case 32:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
	break;
	case 33:
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	break;
	case 34:
#line 523 "./zscan.rl"
	{ rec_a(z); }
	break;
	case 35:
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
	break;
	case 36:
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	break;
	case 37:
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	break;
	case 38:
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	break;
	case 39:
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	break;
	case 40:
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	break;
	case 41:
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	break;
	case 42:
#line 531 "./zscan.rl"
	{ rec_txt(z); }
	break;
	case 43:
#line 532 "./zscan.rl"
	{ rec_spf(z); }
	break;
	case 44:
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
	break;
	case 45:
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	break;
	case 46:
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	break;
	case 47:
#line 536 "./zscan.rl"
	{ rec_rfc3597(z); }
	break;
	case 48:
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
	break;
	case 49:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	break;
	case 50:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	break;
	case 51:
#line 546 "./zscan.rl"
	{z->lcount++;}
	break;
	case 52:
#line 562 "./zscan.rl"
	{z->lcount++;}
	break;
	case 53:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
	break;
#line 2727 "./zscan_lm.c"
		}
	}

_again:
	if ( cs == 0 )
		goto _out;
	if ( ++p != pe )
		goto _resume;
	_test_eof: {}
	_out: {}
	}

#line 756 "./zscan.rl"


        if(cs == zone_error)
            parse_error_noargs("unparseable");
    }

    free(buf);
}

void scan_zone(const zoneinfo_t* zone) {
    dmn_assert(zone);

    log_debug("Scanning zone '%s'", logf_dname(zone->dname));

    zscan_t* z = malloc(sizeof(zscan_t));
    zscan_init(z, zone->zones_dir, zone->dname, zone->dname, zone->file, zone->def_ttl, 0, 0, zone->n_subzones, zone->subzones);
    int fd = open(zone->file, O_RDONLY);
    if(fd < 0)
        log_fatal("Cannot open zone file '%s' for reading: %s", zone->file, logf_errno());
    scanner(z, fd);
    if(close(fd))
        log_fatal("Cannot close zone file '%s': %s", zone->file, logf_errno());
    free(z);
}
