
#line 1 "./zscan.rl"
/* Copyright © 2012 Brandon L Black <blblack@gmail.com> and Jay Reitz <jreitz@gmail.com>
 *
 * This file is part of gdnsd.
 *
 * gdnsd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "zscan.h"

#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "conf.h"
#include "ltree.h"
#include "ltarena.h"
#include "gdnsd-misc.h"

#ifndef INET6_ADDRSTRLEN
#define INET6_ADDRSTRLEN 46
#endif

#define parse_error(_fmt, ...) \
    log_fatal("Zonefile parse error at line %u of %s: " _fmt,z->lcount,z->curfn,__VA_ARGS__)

#define parse_error_noargs(_fmt) \
    log_fatal("Zonefile parse error at line %u of %s: " _fmt,z->lcount,z->curfn)

typedef struct {
    uint8_t  ipv6[16];
    uint32_t ipv4;
    bool     in_paren;
    bool     zn_err_detect;
    bool     lhs_is_ooz;
    unsigned n_subzones;
    unsigned lcount;
    unsigned num_texts;
    unsigned def_ttl;
    unsigned uval;
    unsigned ttl;
    unsigned uv_1;
    unsigned uv_2;
    unsigned uv_3;
    unsigned uv_4;
    unsigned uv_5;
    unsigned rfc3597_data_len;
    unsigned rfc3597_data_written;
    unsigned limit_v4;
    unsigned limit_v6;
    uint8_t* rfc3597_data;
    const uint8_t** subzones;
    const char* zones_dir;
    const char* curfn;
    const char* tstart;
    uint8_t* include_filename;
    uint8_t  zroot[256];
    uint8_t  origin[256];
    uint8_t  lhs_dname[256];
    uint8_t  rhs_dname[256];
    uint8_t  eml_dname[256];
    uint8_t** texts;
} zscan_t;

F_NONNULL
static void scanner(zscan_t* z, int fd);

/******** IP Addresses ********/

F_NONNULL
static void set_ipv4(zscan_t* z, const char* end) {
    dmn_assert(z); dmn_assert(end);
    char txt[16];
    unsigned len = end - z->tstart;
    memcpy(txt, z->tstart, len);
    txt[len] = 0;
    z->tstart = NULL;
    struct in_addr addr;
    int status = inet_pton(AF_INET, txt, &addr);
    if(status > 0)
        z->ipv4 = addr.s_addr;
    else
        parse_error("IPv4 address '%s' invalid", txt);
}

F_NONNULL
static void set_ipv6(zscan_t* z, const char* end) {
    dmn_assert(z);
    dmn_assert(end);
    char txt[INET6_ADDRSTRLEN + 1];
    unsigned len = end - z->tstart;
    memcpy(txt, z->tstart, len);
    txt[len] = 0;
    z->tstart = NULL;
    struct in6_addr* v6a = malloc(sizeof(struct in6_addr));
    int status = inet_pton(AF_INET6, txt, v6a);
    if(status > 0)
        memcpy(z->ipv6, v6a->s6_addr, 16);
    else
        parse_error("IPv6 address '%s' invalid", txt);

    free(v6a);
}

F_NONNULL
static void validate_dname_in_zone(const zscan_t* z, const uint8_t* dname) {
    dmn_assert(z); dmn_assert(z->zroot); dmn_assert(dname);
    if(!dname_isinzone(z->zroot, dname))
        parse_error("domainname %s is not within this zonefile's zone (%s)", logf_dname(dname), logf_dname(z->zroot));
}

F_NONNULL
static void validate_lhs_not_ooz(const zscan_t* z) {
    dmn_assert(z);
    if(z->lhs_is_ooz)
        parse_error("domainname %s is not within this zonefile's zone (%s)", logf_dname(z->lhs_dname), logf_dname(z->zroot));
}

F_NONNULL
static void dname_set(zscan_t* z, uint8_t* dname, unsigned len, bool lhs) {
    dmn_assert(z); dmn_assert(dname); dmn_assert(z->zroot); dmn_assert(z->origin);
    dname_status_t catstat;
    dname_status_t status = dname_from_string(dname, (const uint8_t*)z->tstart, len);
    switch(status) {
        case DNAME_INVALID:
            parse_error_noargs("unparseable domainname");
            break;
        case DNAME_VALID:
            if(lhs) z->lhs_is_ooz = !dname_isinzone(z->zroot, dname);
            break;
        case DNAME_PARTIAL:
            if(lhs) z->lhs_is_ooz = false;
            catstat = dname_cat(dname, z->origin);
            if(catstat == DNAME_INVALID)
                parse_error_noargs("illegal domainname");
            break;
    }
}

F_NONNULL
static zscan_t* zscan_init(zscan_t* z, const char* zones_dir, const uint8_t* zname, const uint8_t* origin, const char* fn, const unsigned def_ttl_arg, const unsigned limit_v4, const unsigned limit_v6, const unsigned n_subzones, const uint8_t** subzones) {
    dmn_assert(z); dmn_assert(zname); dmn_assert(origin); dmn_assert(fn);
    memset(z, 0, sizeof(zscan_t));
    z->lcount = 1;
    z->def_ttl = def_ttl_arg;
    z->curfn = fn;
    z->n_subzones = n_subzones;
    z->subzones = subzones;
    z->zones_dir = zones_dir;
    z->limit_v4 = limit_v4;
    z->limit_v6 = limit_v6;
    dname_copy(z->zroot, zname);
    dname_copy(z->origin, origin);
    dname_copy(z->lhs_dname, origin);

    return z;
}

/********** TXT/SPF ******************/

F_NONNULL
static void text_start(zscan_t* z) {
    dmn_assert(z);
    z->num_texts = 0;
    z->texts = NULL;
}

// Only if we didn't add_rec, as normally
//  add_rec consumes the allocation.
F_NONNULL
static void texts_free(zscan_t* z) {
    for(unsigned i = 0; i < z->num_texts; i++)
        free(z->texts[i]);
}

F_NONNULL
static void text_add_tok(zscan_t* z, const unsigned len, const bool big_ok) {
    dmn_assert(z);

    uint8_t text_temp[len + 1];

    const unsigned newlen = dns_unescape(text_temp, (const uint8_t*)z->tstart, len);
    dmn_assert(newlen <= len);

    if(newlen > 255) {
        if(!big_ok || gconfig.disable_text_autosplit)
            parse_error_noargs("Text chunk too long (>255 unescaped)");
        if(newlen > 65500) parse_error_noargs("Text chunk too long (>65500 unescaped)");
        unsigned remainder = newlen % 255;
        unsigned num_whole_chunks = (newlen - remainder) / 255;
        const uint8_t* zptr = text_temp;
        const unsigned new_alloc = 1 + z->num_texts + num_whole_chunks + (remainder ? 1 : 0);
        z->texts = realloc(z->texts, new_alloc * sizeof(uint8_t*));
        for(unsigned i = 0; i < num_whole_chunks; i++) {
            uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(256);
            *chunk++ = 255;
            memcpy(chunk, zptr, 255);
            zptr += 255;
        }
        if(remainder) {
            uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(remainder + 1);
            *chunk++ = remainder;
            memcpy(chunk, zptr, remainder);
        }
        z->texts[z->num_texts] = NULL;
    }
    else {
        z->texts = realloc(z->texts, (z->num_texts + 2) * sizeof(uint8_t*));
        uint8_t* chunk = z->texts[z->num_texts++] = lta_malloc_1(newlen + 1);
        *chunk++ = newlen;
        memcpy(chunk, text_temp, newlen);
        z->texts[z->num_texts] = NULL;
    }

    z->tstart = NULL;
}

F_NONNULL
static void set_filename(zscan_t* z, const unsigned len) {
    dmn_assert(z);
    uint8_t* fn = malloc(len + 1);
    const unsigned newlen = dns_unescape(fn, (const uint8_t*)z->tstart, len);
    dmn_assert(newlen <= len);
    z->include_filename = fn = realloc(fn, newlen + 1);
    fn[newlen] = 0;
    z->tstart = NULL;
}

// Input must have two bytes of text constrained to [0-9A-Fa-f]
F_NONNULL
static unsigned hexbyte(const char* intxt) {
    dmn_assert(intxt);
    dmn_assert(
        (intxt[0] >= '0' && intxt[0] <= '9')
        || (intxt[0] >= 'A' && intxt[0] <= 'F')
        || (intxt[0] >= 'a' && intxt[0] <= 'f')
    );
    dmn_assert(
        (intxt[1] >= '0' && intxt[1] <= '9')
        || (intxt[1] >= 'A' && intxt[1] <= 'F')
        || (intxt[1] >= 'a' && intxt[1] <= 'f')
    );

    unsigned out;

    if(intxt[0] <= '9')
        out = (intxt[0] - '0') << 4;
    else
        out = ((intxt[0] | 0x20) - ('a' - 10)) << 4;

    if(intxt[1] <= '9')
        out |= (intxt[1] - '0');
    else
        out |= ((intxt[1] | 0x20) - ('a' - 10));

    return out;
}

F_NONNULL
static bool lhs_subzones_ok(zscan_t* z) {
    for(unsigned i = 0; i < z->n_subzones; i++) {
        if(dname_isinzone(z->subzones[i], z->lhs_dname)) {
            log_warn("Ignoring RRs for name '%s' in zone '%s' because they overlap subzone '%s'",
                logf_dname(z->lhs_dname), logf_dname(z->zroot), logf_dname(z->subzones[i]));
            return false;
        }
    }

    return true;
}

F_NONNULL
static void process_include(zscan_t* z) {
    dmn_assert(z);
    dmn_assert(z->include_filename);
    char* zfn = gdnsd_make_abs_fn(z->zones_dir, (char*)z->include_filename);
    free(z->include_filename);
    z->include_filename = NULL;
    validate_dname_in_zone(z, z->rhs_dname);
    zscan_t* znew = malloc(sizeof(zscan_t));
    zscan_init(znew, z->zones_dir, z->zroot, z->rhs_dname, zfn, z->def_ttl, z->limit_v4, z->limit_v6, z->n_subzones, z->subzones);
    int newfd = open(zfn, O_RDONLY);
    if(newfd < 0)
        parse_error("Cannot open $INCLUDE file '%s' for reading: %s", zfn, logf_errno());
    scanner(znew, newfd);
    if(close(newfd))
        parse_error("Cannot close $INCLUDE file '%s': %s", zfn, logf_errno());
    free(zfn);
    free(znew);
}

F_NONNULL
static void mult_uval(zscan_t* z, int fc) {
    dmn_assert(z);
    fc |= 0x20;
    switch(fc) {
        case 'm': z->uval *= 60; break;
        case 'h': z->uval *= 3600; break;
        case 'd': z->uval *= 86400; break;
        case 'w': z->uval *= 604800; break;
    }
}

F_NONNULL
static void set_dyna(zscan_t* z, const char* fpc) {
    dmn_assert(z);
    unsigned dlen = fpc - z->tstart;
    if(dlen > 255)
        parse_error_noargs("DYNA/DYNC plugin!resource string cannot exceed 255 chars");
    memcpy(z->eml_dname, z->tstart, dlen);
    z->eml_dname[dlen] = 0;
    z->tstart = NULL;
}

F_NONNULL
static void rec_soa(zscan_t* z) {
    dmn_assert(z);
    if(dname_cmp(z->lhs_dname, z->zroot))
        parse_error_noargs("SOA record can only be defined for the root of the zone");
    ltree_add_rec_soa(z->lhs_dname, z->rhs_dname, z->eml_dname, z->ttl, z->uv_1, z->uv_2, z->uv_3, z->uv_4, z->uv_5);
}

F_NONNULL
static void rec_a(zscan_t* z) {
    dmn_assert(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_a(z->lhs_dname, z->ipv4, z->ttl, z->limit_v4, z->lhs_is_ooz ? z->zroot : NULL);
}

F_NONNULL
static void rec_aaaa(zscan_t* z) {
    dmn_assert(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_aaaa(z->lhs_dname, z->ipv6, z->ttl, z->limit_v6, z->lhs_is_ooz ? z->zroot : NULL);
}

F_NONNULL
static void rec_ns(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_ns(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_cname(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_cname(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_ptr(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_ptr(z->lhs_dname, z->rhs_dname, z->ttl);
}

F_NONNULL
static void rec_mx(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_mx(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1);
}

F_NONNULL
static void rec_srv(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_srv(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1, z->uv_2, z->uv_3);
}

F_NONNULL
static void rec_naptr(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_naptr(z->lhs_dname, z->rhs_dname, z->ttl, z->uv_1, z->uv_2, z->num_texts, z->texts);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_txt(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_txt(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_spf(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_spf(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_spftxt(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_spftxt(z->lhs_dname, z->num_texts, z->texts, z->ttl);
    else
        texts_free(z);
    free(z->texts);
}

F_NONNULL
static void rec_dyna(zscan_t* z) {
    dmn_assert(z);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_dynaddr(z->lhs_dname, z->eml_dname, z->ttl, z->limit_v4, z->limit_v6);
}

F_NONNULL
static void rec_dyncname(zscan_t* z) {
    dmn_assert(z);

    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z)) {
        ltree_add_rec_dyncname(z->lhs_dname, z->eml_dname, z->origin, z->ttl);
    }
}

F_NONNULL
static void rec_rfc3597(zscan_t* z) {
    dmn_assert(z);
    if(z->rfc3597_data_written < z->rfc3597_data_len)
        parse_error("RFC3597 generic RR claimed rdata length of %u, but only %u bytes of data present", z->rfc3597_data_len, z->rfc3597_data_written);
    validate_lhs_not_ooz(z);
    if(lhs_subzones_ok(z))
        ltree_add_rec_rfc3597(z->lhs_dname, z->uv_1, z->ttl, z->rfc3597_data_len, z->rfc3597_data);
    else
        free(z->rfc3597_data);
}


#line 467 "./zscan.c"
static const int zone_start = 468;
static const int zone_error = 0;

static const int zone_en_main = 468;


#line 692 "./zscan.rl"


/*
 * The longest possible tstart-based token is a maximum-length
 *  quoted TXT string, which given autosplit, can be up to
 *  roughly 64K, which means buffer size has to be that big
 *  to accomodate that.  There probably wouldn't be any bugs
 *  going down to something reasonable like 4K, but it would
 *  cause parse errors if anyone tried to use longer TXT strings.
 * Another important thing: for integers, we use atoi() directly
 *  on the buffer itself.  In the normal case this works because
 *  there is always some non-integer after it in the buffer, which
 *  halts atoi().  The corner case is if the last digit of an
 *  integer happened to be the last byte of the buffer.  This
 *  is why we allocate one extra buffer byte and set it to zero.
 */
#define BUFSIZE 65536

F_NONNULL
static void scanner(zscan_t* z, int fd) {
    dmn_assert(z);

#ifdef HAVE_POSIX_FADVISE
    {
        struct stat fdstat;
        if(fstat(fd, &fdstat))
            log_fatal("fstat() of '%s' failed: %s", z->curfn, logf_errno());
        posix_fadvise(fd, 0, fdstat.st_size, POSIX_FADV_SEQUENTIAL);
    }
#endif

    char* buf = malloc(BUFSIZE + 1);
    char* read_at;
    buf[BUFSIZE] = 0;

    const char* pe = NULL;
    const char* eof = NULL;
    int cs = zone_start;

    while(!eof) {
        unsigned have = 0;
        if(z->tstart != NULL) {
            dmn_assert(pe);
            dmn_assert(z->tstart < pe);
            dmn_assert(z->tstart != buf);
            have = pe - z->tstart;
            memmove(buf, z->tstart, have);
            z->tstart = buf;
        }

        const int space = BUFSIZE - have;
        const char* p = read_at = buf + have;

        const int len = read(fd, read_at, space);
        if(len < 0)
            parse_error("read() failed: %s", logf_errno());

        pe = p + len;

        if(len < space)
            eof = pe;

        
#line 538 "./zscan.c"
	{
	short _widec;
	if ( p == pe )
		goto _test_eof;
	switch ( cs )
	{
tr71:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr64:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr92:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr129:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr140:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr152:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr166:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr187:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr200:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr224:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr237:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr281:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr292:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr304:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr318:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr410:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr421:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr433:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr447:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr543:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr554:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr566:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr580:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr600:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr611:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr623:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr637:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr739:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr746:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr848:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr858:
#line 532 "./zscan.rl"
	{ rec_spf(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr872:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr884:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr917:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr927:
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr941:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr953:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1018:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1029:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1041:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1055:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1081:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1091:
#line 531 "./zscan.rl"
	{ rec_txt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1105:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1117:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1157:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
#line 536 "./zscan.rl"
	{ rec_rfc3597(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1172:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 536 "./zscan.rl"
	{ rec_rfc3597(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1258:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1272:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1295:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1313:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1324:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1336:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1350:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1361:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1373:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1403:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1414:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1426:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1440:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1457:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
tr1464:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st468;
st468:
	if ( ++p == pe )
		goto _test_eof468;
case 468:
#line 1089 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st377;
		case 32: goto st377;
		case 34: goto tr1474;
		case 36: goto st385;
		case 40: goto tr1211;
		case 41: goto tr1212;
		case 64: goto tr1476;
		case 92: goto tr1477;
		case 266: goto tr71;
		case 315: goto st15;
		case 522: goto tr1213;
		case 571: goto st378;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1473;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1473;
	} else
		goto tr1473;
	goto st0;
tr1473:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st1;
tr1202:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st1;
st1:
	if ( ++p == pe )
		goto _test_eof1;
case 1:
#line 1139 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1;
		case 32: goto tr1;
		case 40: goto tr3;
		case 41: goto tr4;
		case 92: goto st373;
		case 522: goto tr6;
		case 571: goto tr7;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st1;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st1;
		} else if ( _widec >= 35 )
			goto st1;
	} else
		goto st1;
	goto st0;
tr1:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
	goto st2;
tr3:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st2;
tr4:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st2;
tr6:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st2;
tr9:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st2;
tr10:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st2;
tr21:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st2;
tr1203:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
	goto st2;
tr1204:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st2;
tr1205:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st2;
tr1207:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st2;
tr1219:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
	goto st2;
tr1220:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st2;
tr1221:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st2;
tr1222:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st2;
tr1468:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
	goto st2;
tr1469:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st2;
tr1470:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st2;
tr1471:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
#line 1287 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st2;
		case 32: goto st2;
		case 40: goto tr9;
		case 41: goto tr10;
		case 65: goto tr12;
		case 67: goto tr13;
		case 68: goto tr14;
		case 73: goto tr15;
		case 77: goto tr16;
		case 78: goto tr17;
		case 80: goto tr18;
		case 83: goto tr19;
		case 84: goto tr20;
		case 97: goto tr12;
		case 99: goto tr13;
		case 100: goto tr14;
		case 105: goto tr15;
		case 109: goto tr16;
		case 110: goto tr17;
		case 112: goto tr18;
		case 115: goto tr19;
		case 116: goto tr20;
		case 522: goto tr21;
		case 571: goto st372;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr11;
	goto st0;
st0:
cs = 0;
	goto _out;
tr11:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
#line 1344 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr23;
		case 32: goto tr23;
		case 40: goto tr24;
		case 41: goto tr25;
		case 68: goto tr27;
		case 72: goto tr27;
		case 77: goto tr27;
		case 87: goto tr27;
		case 100: goto tr27;
		case 104: goto tr27;
		case 109: goto tr27;
		case 119: goto tr27;
		case 522: goto tr28;
		case 571: goto tr29;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st3;
	goto st0;
tr31:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st4;
tr32:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st4;
tr42:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st4;
tr23:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st4;
tr24:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st4;
tr25:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st4;
tr28:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st4;
tr1176:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st4;
tr1177:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st4;
tr1178:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st4;
tr1179:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
#line 1446 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st4;
		case 32: goto st4;
		case 40: goto tr31;
		case 41: goto tr32;
		case 65: goto st5;
		case 67: goto st48;
		case 68: goto st67;
		case 73: goto st98;
		case 77: goto st101;
		case 78: goto st120;
		case 80: goto st199;
		case 83: goto st216;
		case 84: goto st328;
		case 97: goto st5;
		case 99: goto st48;
		case 100: goto st67;
		case 105: goto st98;
		case 109: goto st101;
		case 110: goto st120;
		case 112: goto st199;
		case 115: goto st216;
		case 116: goto st328;
		case 522: goto tr42;
		case 571: goto st364;
	}
	goto st0;
tr12:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
#line 1496 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st6;
		case 32: goto st6;
		case 40: goto tr45;
		case 41: goto tr46;
		case 65: goto st26;
		case 97: goto st26;
		case 522: goto tr48;
		case 571: goto st25;
	}
	goto st0;
tr45:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st6;
tr46:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st6;
tr48:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
#line 1538 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st6;
		case 32: goto st6;
		case 40: goto tr45;
		case 41: goto tr46;
		case 522: goto tr48;
		case 571: goto st25;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr50;
	goto st0;
tr50:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
#line 1572 "./zscan.c"
	if ( (*p) == 46 )
		goto st8;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st23;
	goto st0;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st9;
	goto st0;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
	if ( (*p) == 46 )
		goto st10;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st21;
	goto st0;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st11;
	goto st0;
st11:
	if ( ++p == pe )
		goto _test_eof11;
case 11:
	if ( (*p) == 46 )
		goto st12;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st19;
	goto st0;
st12:
	if ( ++p == pe )
		goto _test_eof12;
case 12:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st13;
	goto st0;
st13:
	if ( ++p == pe )
		goto _test_eof13;
case 13:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr60;
		case 32: goto tr60;
		case 40: goto tr61;
		case 41: goto tr62;
		case 266: goto tr64;
		case 315: goto tr65;
		case 522: goto tr66;
		case 571: goto tr67;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st17;
	goto st0;
tr69:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr70:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr73:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr60:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
	goto st14;
tr61:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr62:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr66:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr87:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
	goto st14;
tr88:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr89:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr94:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr125:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st14;
tr126:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr127:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr131:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr136:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st14;
tr137:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr138:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr142:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr149:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st14;
tr150:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr151:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr154:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr163:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st14;
tr164:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr165:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr168:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr182:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st14;
tr184:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr185:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr189:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr196:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st14;
tr197:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr198:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr202:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr219:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st14;
tr221:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr222:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr226:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr233:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st14;
tr234:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr235:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr239:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr277:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st14;
tr278:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr279:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr283:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr288:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st14;
tr289:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr290:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr294:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr301:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st14;
tr302:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr303:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr306:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr315:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st14;
tr316:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr317:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr320:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr406:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st14;
tr407:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr408:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr412:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr417:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st14;
tr418:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr419:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr423:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr430:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st14;
tr431:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr432:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr435:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr444:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st14;
tr445:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr446:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr449:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr539:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st14;
tr540:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr541:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr545:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr550:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st14;
tr551:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr552:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr556:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr563:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st14;
tr564:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr565:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr568:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr577:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st14;
tr578:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr579:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr582:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr596:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st14;
tr597:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr598:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr602:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr607:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st14;
tr608:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr609:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr613:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr620:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st14;
tr621:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr622:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr625:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr634:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st14;
tr635:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr636:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr639:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr734:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st14;
tr735:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr736:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr741:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr743:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st14;
tr744:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr745:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr748:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1014:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st14;
tr1015:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1016:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1020:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1025:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st14;
tr1026:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1027:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1031:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1038:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st14;
tr1039:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1040:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1043:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1052:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st14;
tr1053:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1054:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1057:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1254:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
	goto st14;
tr1255:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1256:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1260:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1268:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
	goto st14;
tr1269:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1270:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1274:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1309:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st14;
tr1310:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1311:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1315:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1320:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st14;
tr1321:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1322:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1326:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1333:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st14;
tr1334:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1335:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1338:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1347:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st14;
tr1348:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1349:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1352:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1399:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st14;
tr1400:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1401:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1405:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1410:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st14;
tr1411:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1412:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1416:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1423:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st14;
tr1424:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1425:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1428:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1437:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st14;
tr1438:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1439:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1442:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1452:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st14;
tr1453:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1454:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1459:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
tr1461:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st14;
tr1462:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st14;
tr1463:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st14;
tr1466:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
#line 3144 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st14;
		case 32: goto st14;
		case 40: goto tr69;
		case 41: goto tr70;
		case 266: goto tr71;
		case 315: goto st15;
		case 522: goto tr73;
		case 571: goto st16;
	}
	goto st0;
tr65:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
	goto st15;
tr93:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
	goto st15;
tr130:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st15;
tr141:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st15;
tr153:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st15;
tr167:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st15;
tr188:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st15;
tr201:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st15;
tr225:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st15;
tr238:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st15;
tr282:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st15;
tr293:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st15;
tr305:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st15;
tr319:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st15;
tr411:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st15;
tr422:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st15;
tr434:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st15;
tr448:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st15;
tr544:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st15;
tr555:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st15;
tr567:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st15;
tr581:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st15;
tr601:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st15;
tr612:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st15;
tr624:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st15;
tr638:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st15;
tr740:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st15;
tr747:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st15;
tr849:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
	goto st15;
tr859:
#line 532 "./zscan.rl"
	{ rec_spf(z); }
	goto st15;
tr873:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
	goto st15;
tr885:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 532 "./zscan.rl"
	{ rec_spf(z); }
	goto st15;
tr918:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
	goto st15;
tr928:
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
	goto st15;
tr942:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
	goto st15;
tr954:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 533 "./zscan.rl"
	{ rec_spftxt(z); }
	goto st15;
tr1019:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st15;
tr1030:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st15;
tr1042:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st15;
tr1056:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st15;
tr1082:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
	goto st15;
tr1092:
#line 531 "./zscan.rl"
	{ rec_txt(z); }
	goto st15;
tr1106:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
	goto st15;
tr1118:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 531 "./zscan.rl"
	{ rec_txt(z); }
	goto st15;
tr1158:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
#line 536 "./zscan.rl"
	{ rec_rfc3597(z); }
	goto st15;
tr1173:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 536 "./zscan.rl"
	{ rec_rfc3597(z); }
	goto st15;
tr1259:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
	goto st15;
tr1273:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
	goto st15;
tr1296:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1314:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1325:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1337:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1351:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1362:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1374:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st15;
tr1404:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st15;
tr1415:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st15;
tr1427:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st15;
tr1441:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st15;
tr1458:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st15;
tr1465:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
#line 3588 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 266 )
		goto tr71;
	if ( 128 <= _widec && _widec <= 383 )
		goto st15;
	goto st0;
tr67:
#line 493 "./zscan.rl"
	{ set_ipv4(z, p); }
#line 523 "./zscan.rl"
	{ rec_a(z); }
	goto st16;
tr95:
#line 494 "./zscan.rl"
	{ set_ipv6(z, p); }
#line 524 "./zscan.rl"
	{ rec_aaaa(z); }
	goto st16;
tr132:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st16;
tr143:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st16;
tr155:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st16;
tr169:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 526 "./zscan.rl"
	{ rec_cname(z); }
	goto st16;
tr190:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st16;
tr203:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 534 "./zscan.rl"
	{ rec_dyna(z); }
	goto st16;
tr227:
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st16;
tr240:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 520 "./zscan.rl"
	{ set_dyna(z, p); }
#line 535 "./zscan.rl"
	{ rec_dyncname(z); }
	goto st16;
tr284:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st16;
tr295:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st16;
tr307:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st16;
tr321:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 528 "./zscan.rl"
	{ rec_mx(z); }
	goto st16;
tr413:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st16;
tr424:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st16;
tr436:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st16;
tr450:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 530 "./zscan.rl"
	{ rec_naptr(z); }
	goto st16;
tr546:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st16;
tr557:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st16;
tr569:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st16;
tr583:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 525 "./zscan.rl"
	{ rec_ns(z); }
	goto st16;
tr603:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st16;
tr614:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st16;
tr626:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st16;
tr640:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 527 "./zscan.rl"
	{ rec_ptr(z); }
	goto st16;
tr742:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st16;
tr749:
#line 505 "./zscan.rl"
	{ z->uv_5 = z->uval; }
#line 522 "./zscan.rl"
	{ rec_soa(z); }
	goto st16;
tr1021:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st16;
tr1032:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st16;
tr1044:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st16;
tr1058:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 529 "./zscan.rl"
	{ rec_srv(z); }
	goto st16;
tr1261:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 507 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V4 value %u out of range (0-65535)", z->uval);
        z->limit_v4 = z->uval;
    }
	goto st16;
tr1275:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 513 "./zscan.rl"
	{
        if(z->uval > 65535)
            parse_error("$ADDR_LIMIT_V6 value %u out of range (0-65535)", z->uval);
        z->limit_v6 = z->uval;
    }
	goto st16;
tr1316:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st16;
tr1327:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st16;
tr1339:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st16;
tr1353:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 485 "./zscan.rl"
	{ process_include(z); }
	goto st16;
tr1406:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st16;
tr1417:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st16;
tr1429:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st16;
tr1443:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 477 "./zscan.rl"
	{
        validate_dname_in_zone(z, z->rhs_dname);
        dname_copy(z->origin, z->rhs_dname);
    }
	goto st16;
tr1460:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st16;
tr1467:
#line 499 "./zscan.rl"
	{ z->def_ttl = z->uval; }
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
#line 3923 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr73;
	if ( 384 <= _widec && _widec <= 639 )
		goto st16;
	goto st0;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr60;
		case 32: goto tr60;
		case 40: goto tr61;
		case 41: goto tr62;
		case 266: goto tr64;
		case 315: goto tr65;
		case 522: goto tr66;
		case 571: goto tr67;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st18;
	goto st0;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr60;
		case 32: goto tr60;
		case 40: goto tr61;
		case 41: goto tr62;
		case 266: goto tr64;
		case 315: goto tr65;
		case 522: goto tr66;
		case 571: goto tr67;
	}
	goto st0;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
	if ( (*p) == 46 )
		goto st12;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st20;
	goto st0;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
	if ( (*p) == 46 )
		goto st12;
	goto st0;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
	if ( (*p) == 46 )
		goto st10;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st22;
	goto st0;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
	if ( (*p) == 46 )
		goto st10;
	goto st0;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
	if ( (*p) == 46 )
		goto st8;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st24;
	goto st0;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
	if ( (*p) == 46 )
		goto st8;
	goto st0;
st25:
	if ( ++p == pe )
		goto _test_eof25;
case 25:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr48;
	if ( 384 <= _widec && _widec <= 639 )
		goto st25;
	goto st0;
st26:
	if ( ++p == pe )
		goto _test_eof26;
case 26:
	switch( (*p) ) {
		case 65: goto st27;
		case 97: goto st27;
	}
	goto st0;
st27:
	if ( ++p == pe )
		goto _test_eof27;
case 27:
	switch( (*p) ) {
		case 65: goto st28;
		case 97: goto st28;
	}
	goto st0;
st28:
	if ( ++p == pe )
		goto _test_eof28;
case 28:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st29;
		case 32: goto st29;
		case 40: goto tr82;
		case 41: goto tr83;
		case 522: goto tr84;
		case 571: goto st47;
	}
	goto st0;
tr82:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st29;
tr83:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st29;
tr84:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st29;
st29:
	if ( ++p == pe )
		goto _test_eof29;
case 29:
#line 4149 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st29;
		case 32: goto st29;
		case 40: goto tr82;
		case 41: goto tr83;
		case 522: goto tr84;
		case 571: goto st47;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr86;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr86;
	} else
		goto tr86;
	goto st0;
tr86:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st30;
st30:
	if ( ++p == pe )
		goto _test_eof30;
case 30:
#line 4189 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 58: goto st31;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto st30;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto st30;
	} else
		goto st30;
	goto st0;
st31:
	if ( ++p == pe )
		goto _test_eof31;
case 31:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 58: goto st31;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto st32;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto st30;
	} else
		goto st30;
	goto st0;
st32:
	if ( ++p == pe )
		goto _test_eof32;
case 32:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 46: goto st33;
		case 58: goto st31;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto st45;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto st30;
	} else
		goto st30;
	goto st0;
st33:
	if ( ++p == pe )
		goto _test_eof33;
case 33:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st34;
	goto st0;
st34:
	if ( ++p == pe )
		goto _test_eof34;
case 34:
	if ( (*p) == 46 )
		goto st35;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st43;
	goto st0;
st35:
	if ( ++p == pe )
		goto _test_eof35;
case 35:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st36;
	goto st0;
st36:
	if ( ++p == pe )
		goto _test_eof36;
case 36:
	if ( (*p) == 46 )
		goto st37;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st41;
	goto st0;
st37:
	if ( ++p == pe )
		goto _test_eof37;
case 37:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st38;
	goto st0;
st38:
	if ( ++p == pe )
		goto _test_eof38;
case 38:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st39;
	goto st0;
st39:
	if ( ++p == pe )
		goto _test_eof39;
case 39:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st40;
	goto st0;
st40:
	if ( ++p == pe )
		goto _test_eof40;
case 40:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	goto st0;
st41:
	if ( ++p == pe )
		goto _test_eof41;
case 41:
	if ( (*p) == 46 )
		goto st37;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st42;
	goto st0;
st42:
	if ( ++p == pe )
		goto _test_eof42;
case 42:
	if ( (*p) == 46 )
		goto st37;
	goto st0;
st43:
	if ( ++p == pe )
		goto _test_eof43;
case 43:
	if ( (*p) == 46 )
		goto st35;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st44;
	goto st0;
st44:
	if ( ++p == pe )
		goto _test_eof44;
case 44:
	if ( (*p) == 46 )
		goto st35;
	goto st0;
st45:
	if ( ++p == pe )
		goto _test_eof45;
case 45:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 46: goto st33;
		case 58: goto st31;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto st46;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto st30;
	} else
		goto st30;
	goto st0;
st46:
	if ( ++p == pe )
		goto _test_eof46;
case 46:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr87;
		case 32: goto tr87;
		case 40: goto tr88;
		case 41: goto tr89;
		case 46: goto st33;
		case 58: goto st31;
		case 266: goto tr92;
		case 315: goto tr93;
		case 522: goto tr94;
		case 571: goto tr95;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto st30;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto st30;
	} else
		goto st30;
	goto st0;
st47:
	if ( ++p == pe )
		goto _test_eof47;
case 47:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr84;
	if ( 384 <= _widec && _widec <= 639 )
		goto st47;
	goto st0;
tr13:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st48;
st48:
	if ( ++p == pe )
		goto _test_eof48;
case 48:
#line 4579 "./zscan.c"
	switch( (*p) ) {
		case 78: goto st49;
		case 110: goto st49;
	}
	goto st0;
st49:
	if ( ++p == pe )
		goto _test_eof49;
case 49:
	switch( (*p) ) {
		case 65: goto st50;
		case 97: goto st50;
	}
	goto st0;
st50:
	if ( ++p == pe )
		goto _test_eof50;
case 50:
	switch( (*p) ) {
		case 77: goto st51;
		case 109: goto st51;
	}
	goto st0;
st51:
	if ( ++p == pe )
		goto _test_eof51;
case 51:
	switch( (*p) ) {
		case 69: goto st52;
		case 101: goto st52;
	}
	goto st0;
st52:
	if ( ++p == pe )
		goto _test_eof52;
case 52:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st53;
		case 32: goto st53;
		case 40: goto tr116;
		case 41: goto tr117;
		case 522: goto tr118;
		case 571: goto st66;
	}
	goto st0;
tr116:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st53;
tr117:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st53;
tr118:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st53;
st53:
	if ( ++p == pe )
		goto _test_eof53;
case 53:
#line 4655 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st53;
		case 32: goto st53;
		case 34: goto tr121;
		case 40: goto tr116;
		case 41: goto tr117;
		case 64: goto tr122;
		case 92: goto tr123;
		case 522: goto tr118;
		case 571: goto st66;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr120;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr120;
		} else if ( _widec >= 37 )
			goto tr120;
	} else
		goto tr120;
	goto st0;
tr120:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st54;
tr135:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st54;
st54:
	if ( ++p == pe )
		goto _test_eof54;
case 54:
#line 4705 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr125;
		case 32: goto tr125;
		case 40: goto tr126;
		case 41: goto tr127;
		case 92: goto st55;
		case 266: goto tr129;
		case 315: goto tr130;
		case 522: goto tr131;
		case 571: goto tr132;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st54;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st54;
		} else if ( _widec >= 35 )
			goto st54;
	} else
		goto st54;
	goto st0;
tr123:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st55;
tr139:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st55;
st55:
	if ( ++p == pe )
		goto _test_eof55;
case 55:
#line 4755 "./zscan.c"
	if ( (*p) == 10 )
		goto st56;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st57;
	goto st54;
st56:
	if ( ++p == pe )
		goto _test_eof56;
case 56:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr136;
		case 32: goto tr136;
		case 40: goto tr137;
		case 41: goto tr138;
		case 92: goto tr139;
		case 266: goto tr140;
		case 315: goto tr141;
		case 522: goto tr142;
		case 571: goto tr143;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr135;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr135;
		} else if ( _widec >= 35 )
			goto tr135;
	} else
		goto tr135;
	goto st0;
st57:
	if ( ++p == pe )
		goto _test_eof57;
case 57:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st58;
	goto st0;
st58:
	if ( ++p == pe )
		goto _test_eof58;
case 58:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st54;
	goto st0;
tr146:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st59;
tr121:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st59;
tr158:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st59;
tr159:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st59;
st59:
	if ( ++p == pe )
		goto _test_eof59;
case 59:
#line 4841 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr146;
		case 34: goto st60;
		case 92: goto st61;
	}
	goto st59;
tr160:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st60;
st60:
	if ( ++p == pe )
		goto _test_eof60;
case 60:
#line 4856 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr149;
		case 32: goto tr149;
		case 40: goto tr150;
		case 41: goto tr151;
		case 266: goto tr152;
		case 315: goto tr153;
		case 522: goto tr154;
		case 571: goto tr155;
	}
	goto st0;
tr161:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st61;
st61:
	if ( ++p == pe )
		goto _test_eof61;
case 61:
#line 4890 "./zscan.c"
	if ( (*p) == 10 )
		goto st62;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st63;
	goto st59;
st62:
	if ( ++p == pe )
		goto _test_eof62;
case 62:
	switch( (*p) ) {
		case 10: goto tr159;
		case 34: goto tr160;
		case 92: goto tr161;
	}
	goto tr158;
st63:
	if ( ++p == pe )
		goto _test_eof63;
case 63:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st64;
	goto st0;
st64:
	if ( ++p == pe )
		goto _test_eof64;
case 64:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st59;
	goto st0;
tr122:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st65;
st65:
	if ( ++p == pe )
		goto _test_eof65;
case 65:
#line 4931 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr163;
		case 32: goto tr163;
		case 40: goto tr164;
		case 41: goto tr165;
		case 266: goto tr166;
		case 315: goto tr167;
		case 522: goto tr168;
		case 571: goto tr169;
	}
	goto st0;
st66:
	if ( ++p == pe )
		goto _test_eof66;
case 66:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr118;
	if ( 384 <= _widec && _widec <= 639 )
		goto st66;
	goto st0;
tr14:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st67;
st67:
	if ( ++p == pe )
		goto _test_eof67;
case 67:
#line 4995 "./zscan.c"
	switch( (*p) ) {
		case 89: goto st68;
		case 121: goto st68;
	}
	goto st0;
st68:
	if ( ++p == pe )
		goto _test_eof68;
case 68:
	switch( (*p) ) {
		case 78: goto st69;
		case 110: goto st69;
	}
	goto st0;
st69:
	if ( ++p == pe )
		goto _test_eof69;
case 69:
	switch( (*p) ) {
		case 65: goto st70;
		case 67: goto st84;
		case 97: goto st70;
		case 99: goto st84;
	}
	goto st0;
st70:
	if ( ++p == pe )
		goto _test_eof70;
case 70:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st71;
		case 32: goto st71;
		case 40: goto tr175;
		case 41: goto tr176;
		case 522: goto tr177;
		case 571: goto st83;
	}
	goto st0;
tr175:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st71;
tr176:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st71;
tr177:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st71;
st71:
	if ( ++p == pe )
		goto _test_eof71;
case 71:
#line 5064 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st71;
		case 32: goto st71;
		case 40: goto tr175;
		case 41: goto tr176;
		case 92: goto tr180;
		case 522: goto tr177;
		case 571: goto st83;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr179;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr179;
		} else if ( _widec >= 35 )
			goto tr179;
	} else
		goto tr179;
	goto st0;
tr179:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st72;
tr207:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st72;
st72:
	if ( ++p == pe )
		goto _test_eof72;
case 72:
#line 5112 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr182;
		case 32: goto tr182;
		case 33: goto st73;
		case 40: goto tr184;
		case 41: goto tr185;
		case 92: goto st79;
		case 266: goto tr187;
		case 315: goto tr188;
		case 522: goto tr189;
		case 571: goto tr190;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st72;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st72;
		} else if ( _widec >= 35 )
			goto st72;
	} else
		goto st72;
	goto st0;
tr208:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st73;
st73:
	if ( ++p == pe )
		goto _test_eof73;
case 73:
#line 5159 "./zscan.c"
	switch( (*p) ) {
		case 59: goto st0;
		case 92: goto st75;
	}
	if ( (*p) < 32 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto st0;
	} else if ( (*p) > 34 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto st0;
	} else
		goto st0;
	goto st74;
tr195:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st74;
st74:
	if ( ++p == pe )
		goto _test_eof74;
case 74:
#line 5181 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr182;
		case 32: goto tr182;
		case 40: goto tr184;
		case 41: goto tr185;
		case 92: goto st75;
		case 266: goto tr187;
		case 315: goto tr188;
		case 522: goto tr189;
		case 571: goto tr190;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st74;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st74;
		} else if ( _widec >= 35 )
			goto st74;
	} else
		goto st74;
	goto st0;
tr199:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st75;
st75:
	if ( ++p == pe )
		goto _test_eof75;
case 75:
#line 5227 "./zscan.c"
	if ( (*p) == 10 )
		goto st76;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st77;
	goto st74;
st76:
	if ( ++p == pe )
		goto _test_eof76;
case 76:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr196;
		case 32: goto tr196;
		case 40: goto tr197;
		case 41: goto tr198;
		case 92: goto tr199;
		case 266: goto tr200;
		case 315: goto tr201;
		case 522: goto tr202;
		case 571: goto tr203;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr195;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr195;
		} else if ( _widec >= 35 )
			goto tr195;
	} else
		goto tr195;
	goto st0;
st77:
	if ( ++p == pe )
		goto _test_eof77;
case 77:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st78;
	goto st0;
st78:
	if ( ++p == pe )
		goto _test_eof78;
case 78:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st74;
	goto st0;
tr180:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st79;
tr209:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st79;
st79:
	if ( ++p == pe )
		goto _test_eof79;
case 79:
#line 5303 "./zscan.c"
	if ( (*p) == 10 )
		goto st80;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st81;
	goto st72;
st80:
	if ( ++p == pe )
		goto _test_eof80;
case 80:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr196;
		case 32: goto tr196;
		case 33: goto tr208;
		case 40: goto tr197;
		case 41: goto tr198;
		case 92: goto tr209;
		case 266: goto tr200;
		case 315: goto tr201;
		case 522: goto tr202;
		case 571: goto tr203;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr207;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr207;
		} else if ( _widec >= 35 )
			goto tr207;
	} else
		goto tr207;
	goto st0;
st81:
	if ( ++p == pe )
		goto _test_eof81;
case 81:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st82;
	goto st0;
st82:
	if ( ++p == pe )
		goto _test_eof82;
case 82:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st72;
	goto st0;
st83:
	if ( ++p == pe )
		goto _test_eof83;
case 83:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr177;
	if ( 384 <= _widec && _widec <= 639 )
		goto st83;
	goto st0;
st84:
	if ( ++p == pe )
		goto _test_eof84;
case 84:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st85;
		case 32: goto st85;
		case 40: goto tr212;
		case 41: goto tr213;
		case 522: goto tr214;
		case 571: goto st97;
	}
	goto st0;
tr212:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st85;
tr213:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st85;
tr214:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st85;
st85:
	if ( ++p == pe )
		goto _test_eof85;
case 85:
#line 5441 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st85;
		case 32: goto st85;
		case 40: goto tr212;
		case 41: goto tr213;
		case 92: goto tr217;
		case 522: goto tr214;
		case 571: goto st97;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr216;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr216;
		} else if ( _widec >= 35 )
			goto tr216;
	} else
		goto tr216;
	goto st0;
tr216:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st86;
tr244:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st86;
st86:
	if ( ++p == pe )
		goto _test_eof86;
case 86:
#line 5489 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr219;
		case 32: goto tr219;
		case 33: goto st87;
		case 40: goto tr221;
		case 41: goto tr222;
		case 92: goto st93;
		case 266: goto tr224;
		case 315: goto tr225;
		case 522: goto tr226;
		case 571: goto tr227;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st86;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st86;
		} else if ( _widec >= 35 )
			goto st86;
	} else
		goto st86;
	goto st0;
tr245:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st87;
st87:
	if ( ++p == pe )
		goto _test_eof87;
case 87:
#line 5536 "./zscan.c"
	switch( (*p) ) {
		case 59: goto st0;
		case 92: goto st89;
	}
	if ( (*p) < 32 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto st0;
	} else if ( (*p) > 34 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto st0;
	} else
		goto st0;
	goto st88;
tr232:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st88;
st88:
	if ( ++p == pe )
		goto _test_eof88;
case 88:
#line 5558 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr219;
		case 32: goto tr219;
		case 40: goto tr221;
		case 41: goto tr222;
		case 92: goto st89;
		case 266: goto tr224;
		case 315: goto tr225;
		case 522: goto tr226;
		case 571: goto tr227;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st88;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st88;
		} else if ( _widec >= 35 )
			goto st88;
	} else
		goto st88;
	goto st0;
tr236:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st89;
st89:
	if ( ++p == pe )
		goto _test_eof89;
case 89:
#line 5604 "./zscan.c"
	if ( (*p) == 10 )
		goto st90;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st91;
	goto st88;
st90:
	if ( ++p == pe )
		goto _test_eof90;
case 90:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr233;
		case 32: goto tr233;
		case 40: goto tr234;
		case 41: goto tr235;
		case 92: goto tr236;
		case 266: goto tr237;
		case 315: goto tr238;
		case 522: goto tr239;
		case 571: goto tr240;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr232;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr232;
		} else if ( _widec >= 35 )
			goto tr232;
	} else
		goto tr232;
	goto st0;
st91:
	if ( ++p == pe )
		goto _test_eof91;
case 91:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st92;
	goto st0;
st92:
	if ( ++p == pe )
		goto _test_eof92;
case 92:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st88;
	goto st0;
tr217:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st93;
tr246:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st93;
st93:
	if ( ++p == pe )
		goto _test_eof93;
case 93:
#line 5680 "./zscan.c"
	if ( (*p) == 10 )
		goto st94;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st95;
	goto st86;
st94:
	if ( ++p == pe )
		goto _test_eof94;
case 94:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr233;
		case 32: goto tr233;
		case 33: goto tr245;
		case 40: goto tr234;
		case 41: goto tr235;
		case 92: goto tr246;
		case 266: goto tr237;
		case 315: goto tr238;
		case 522: goto tr239;
		case 571: goto tr240;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr244;
	} else if ( _widec > 31 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr244;
		} else if ( _widec >= 35 )
			goto tr244;
	} else
		goto tr244;
	goto st0;
st95:
	if ( ++p == pe )
		goto _test_eof95;
case 95:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st96;
	goto st0;
st96:
	if ( ++p == pe )
		goto _test_eof96;
case 96:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st86;
	goto st0;
st97:
	if ( ++p == pe )
		goto _test_eof97;
case 97:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr214;
	if ( 384 <= _widec && _widec <= 639 )
		goto st97;
	goto st0;
st98:
	if ( ++p == pe )
		goto _test_eof98;
case 98:
	switch( (*p) ) {
		case 78: goto st99;
		case 110: goto st99;
	}
	goto st0;
st99:
	if ( ++p == pe )
		goto _test_eof99;
case 99:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st100;
		case 32: goto st100;
		case 40: goto tr250;
		case 41: goto tr251;
		case 522: goto tr252;
		case 571: goto st363;
	}
	goto st0;
tr250:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st100;
tr251:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st100;
tr252:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st100;
tr1188:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st100;
tr1189:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st100;
tr1190:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st100;
tr1193:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st100;
tr1195:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st100;
tr1196:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st100;
tr1197:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st100;
tr1198:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st100;
st100:
	if ( ++p == pe )
		goto _test_eof100;
case 100:
#line 5879 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st100;
		case 32: goto st100;
		case 40: goto tr250;
		case 41: goto tr251;
		case 65: goto st5;
		case 67: goto st48;
		case 68: goto st67;
		case 77: goto st101;
		case 78: goto st120;
		case 80: goto st199;
		case 83: goto st216;
		case 84: goto st328;
		case 97: goto st5;
		case 99: goto st48;
		case 100: goto st67;
		case 109: goto st101;
		case 110: goto st120;
		case 112: goto st199;
		case 115: goto st216;
		case 116: goto st328;
		case 522: goto tr252;
		case 571: goto st363;
	}
	goto st0;
tr16:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st101;
st101:
	if ( ++p == pe )
		goto _test_eof101;
case 101:
#line 5927 "./zscan.c"
	switch( (*p) ) {
		case 88: goto st102;
		case 120: goto st102;
	}
	goto st0;
st102:
	if ( ++p == pe )
		goto _test_eof102;
case 102:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st103;
		case 32: goto st103;
		case 40: goto tr256;
		case 41: goto tr257;
		case 522: goto tr258;
		case 571: goto st119;
	}
	goto st0;
tr256:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st103;
tr257:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st103;
tr258:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st103;
st103:
	if ( ++p == pe )
		goto _test_eof103;
case 103:
#line 5976 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st103;
		case 32: goto st103;
		case 40: goto tr256;
		case 41: goto tr257;
		case 522: goto tr258;
		case 571: goto st119;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr260;
	goto st0;
tr260:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st104;
st104:
	if ( ++p == pe )
		goto _test_eof104;
case 104:
#line 6010 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr261;
		case 32: goto tr261;
		case 40: goto tr262;
		case 41: goto tr263;
		case 522: goto tr265;
		case 571: goto tr266;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st104;
	goto st0;
tr270:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st105;
tr271:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st105;
tr274:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st105;
tr261:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st105;
tr262:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st105;
tr263:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st105;
tr265:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st105;
st105:
	if ( ++p == pe )
		goto _test_eof105;
case 105:
#line 6082 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st105;
		case 32: goto st105;
		case 34: goto tr269;
		case 40: goto tr270;
		case 41: goto tr271;
		case 64: goto tr272;
		case 92: goto tr273;
		case 522: goto tr274;
		case 571: goto st118;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr267;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr267;
		} else if ( _widec >= 37 )
			goto tr267;
	} else
		goto tr267;
	goto st0;
tr267:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st106;
tr287:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st106;
st106:
	if ( ++p == pe )
		goto _test_eof106;
case 106:
#line 6132 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr277;
		case 32: goto tr277;
		case 40: goto tr278;
		case 41: goto tr279;
		case 92: goto st107;
		case 266: goto tr281;
		case 315: goto tr282;
		case 522: goto tr283;
		case 571: goto tr284;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st106;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st106;
		} else if ( _widec >= 35 )
			goto st106;
	} else
		goto st106;
	goto st0;
tr273:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st107;
tr291:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st107;
st107:
	if ( ++p == pe )
		goto _test_eof107;
case 107:
#line 6182 "./zscan.c"
	if ( (*p) == 10 )
		goto st108;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st109;
	goto st106;
st108:
	if ( ++p == pe )
		goto _test_eof108;
case 108:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr288;
		case 32: goto tr288;
		case 40: goto tr289;
		case 41: goto tr290;
		case 92: goto tr291;
		case 266: goto tr292;
		case 315: goto tr293;
		case 522: goto tr294;
		case 571: goto tr295;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr287;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr287;
		} else if ( _widec >= 35 )
			goto tr287;
	} else
		goto tr287;
	goto st0;
st109:
	if ( ++p == pe )
		goto _test_eof109;
case 109:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st110;
	goto st0;
st110:
	if ( ++p == pe )
		goto _test_eof110;
case 110:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st106;
	goto st0;
tr298:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st111;
tr269:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st111;
tr310:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st111;
tr311:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st111;
st111:
	if ( ++p == pe )
		goto _test_eof111;
case 111:
#line 6268 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr298;
		case 34: goto st112;
		case 92: goto st113;
	}
	goto st111;
tr312:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st112;
st112:
	if ( ++p == pe )
		goto _test_eof112;
case 112:
#line 6283 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr301;
		case 32: goto tr301;
		case 40: goto tr302;
		case 41: goto tr303;
		case 266: goto tr304;
		case 315: goto tr305;
		case 522: goto tr306;
		case 571: goto tr307;
	}
	goto st0;
tr313:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st113;
st113:
	if ( ++p == pe )
		goto _test_eof113;
case 113:
#line 6317 "./zscan.c"
	if ( (*p) == 10 )
		goto st114;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st115;
	goto st111;
st114:
	if ( ++p == pe )
		goto _test_eof114;
case 114:
	switch( (*p) ) {
		case 10: goto tr311;
		case 34: goto tr312;
		case 92: goto tr313;
	}
	goto tr310;
st115:
	if ( ++p == pe )
		goto _test_eof115;
case 115:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st116;
	goto st0;
st116:
	if ( ++p == pe )
		goto _test_eof116;
case 116:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st111;
	goto st0;
tr272:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st117;
st117:
	if ( ++p == pe )
		goto _test_eof117;
case 117:
#line 6358 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr315;
		case 32: goto tr315;
		case 40: goto tr316;
		case 41: goto tr317;
		case 266: goto tr318;
		case 315: goto tr319;
		case 522: goto tr320;
		case 571: goto tr321;
	}
	goto st0;
tr266:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st118;
st118:
	if ( ++p == pe )
		goto _test_eof118;
case 118:
#line 6394 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr274;
	if ( 384 <= _widec && _widec <= 639 )
		goto st118;
	goto st0;
st119:
	if ( ++p == pe )
		goto _test_eof119;
case 119:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr258;
	if ( 384 <= _widec && _widec <= 639 )
		goto st119;
	goto st0;
tr17:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st120;
st120:
	if ( ++p == pe )
		goto _test_eof120;
case 120:
#line 6459 "./zscan.c"
	switch( (*p) ) {
		case 65: goto st121;
		case 83: goto st184;
		case 97: goto st121;
		case 115: goto st184;
	}
	goto st0;
st121:
	if ( ++p == pe )
		goto _test_eof121;
case 121:
	switch( (*p) ) {
		case 80: goto st122;
		case 112: goto st122;
	}
	goto st0;
st122:
	if ( ++p == pe )
		goto _test_eof122;
case 122:
	switch( (*p) ) {
		case 84: goto st123;
		case 116: goto st123;
	}
	goto st0;
st123:
	if ( ++p == pe )
		goto _test_eof123;
case 123:
	switch( (*p) ) {
		case 82: goto st124;
		case 114: goto st124;
	}
	goto st0;
st124:
	if ( ++p == pe )
		goto _test_eof124;
case 124:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st125;
		case 32: goto st125;
		case 40: goto tr328;
		case 41: goto tr329;
		case 522: goto tr330;
		case 571: goto st183;
	}
	goto st0;
tr328:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st125;
tr329:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st125;
tr330:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st125;
st125:
	if ( ++p == pe )
		goto _test_eof125;
case 125:
#line 6537 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st125;
		case 32: goto st125;
		case 40: goto tr328;
		case 41: goto tr329;
		case 522: goto tr330;
		case 571: goto st183;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr332;
	goto st0;
tr332:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st126;
st126:
	if ( ++p == pe )
		goto _test_eof126;
case 126:
#line 6571 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr333;
		case 32: goto tr333;
		case 40: goto tr334;
		case 41: goto tr335;
		case 522: goto tr337;
		case 571: goto tr338;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st126;
	goto st0;
tr340:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st127;
tr341:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st127;
tr343:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st127;
tr333:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st127;
tr334:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st127;
tr335:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st127;
tr337:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st127;
st127:
	if ( ++p == pe )
		goto _test_eof127;
case 127:
#line 6643 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st127;
		case 32: goto st127;
		case 40: goto tr340;
		case 41: goto tr341;
		case 522: goto tr343;
		case 571: goto st182;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr342;
	goto st0;
tr342:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st128;
st128:
	if ( ++p == pe )
		goto _test_eof128;
case 128:
#line 6677 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr345;
		case 32: goto tr345;
		case 40: goto tr346;
		case 41: goto tr347;
		case 522: goto tr349;
		case 571: goto tr350;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st128;
	goto st0;
tr354:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st129;
tr355:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st129;
tr357:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st129;
tr345:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st129;
tr346:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st129;
tr347:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st129;
tr349:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st129;
st129:
	if ( ++p == pe )
		goto _test_eof129;
case 129:
#line 6749 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st129;
		case 32: goto st129;
		case 34: goto tr353;
		case 40: goto tr354;
		case 41: goto tr355;
		case 92: goto tr356;
		case 522: goto tr357;
		case 571: goto st181;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr351;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr351;
	} else
		goto tr351;
	goto st0;
tr505:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st130;
tr351:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st130;
st130:
	if ( ++p == pe )
		goto _test_eof130;
case 130:
#line 6797 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr360;
		case 32: goto tr360;
		case 40: goto tr361;
		case 41: goto tr362;
		case 92: goto st171;
		case 522: goto tr364;
		case 571: goto tr365;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st130;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st130;
		} else if ( _widec >= 35 )
			goto st130;
	} else
		goto st130;
	goto st0;
tr369:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st131;
tr370:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st131;
tr372:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st131;
tr360:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st131;
tr361:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st131;
tr362:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st131;
tr364:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st131;
tr506:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st131;
tr507:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st131;
tr508:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st131;
tr510:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st131;
tr517:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st131;
tr518:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st131;
tr519:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st131;
tr520:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st131;
st131:
	if ( ++p == pe )
		goto _test_eof131;
case 131:
#line 6923 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st131;
		case 32: goto st131;
		case 34: goto tr368;
		case 40: goto tr369;
		case 41: goto tr370;
		case 92: goto tr371;
		case 522: goto tr372;
		case 571: goto st170;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr366;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr366;
	} else
		goto tr366;
	goto st0;
tr366:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st132;
tr479:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st132;
st132:
	if ( ++p == pe )
		goto _test_eof132;
case 132:
#line 6969 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr375;
		case 32: goto tr375;
		case 40: goto tr376;
		case 41: goto tr377;
		case 92: goto st160;
		case 522: goto tr379;
		case 571: goto tr380;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st132;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st132;
		} else if ( _widec >= 35 )
			goto st132;
	} else
		goto st132;
	goto st0;
tr384:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st133;
tr385:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st133;
tr387:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st133;
tr375:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st133;
tr376:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st133;
tr377:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st133;
tr379:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st133;
tr480:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st133;
tr481:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st133;
tr482:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st133;
tr484:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st133;
tr491:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st133;
tr492:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st133;
tr493:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st133;
tr494:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st133;
st133:
	if ( ++p == pe )
		goto _test_eof133;
case 133:
#line 7095 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st133;
		case 32: goto st133;
		case 34: goto tr383;
		case 40: goto tr384;
		case 41: goto tr385;
		case 92: goto tr386;
		case 522: goto tr387;
		case 571: goto st159;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr381;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr381;
	} else
		goto tr381;
	goto st0;
tr381:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st134;
tr453:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st134;
st134:
	if ( ++p == pe )
		goto _test_eof134;
case 134:
#line 7141 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr390;
		case 32: goto tr390;
		case 40: goto tr391;
		case 41: goto tr392;
		case 92: goto st149;
		case 522: goto tr394;
		case 571: goto tr395;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st134;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st134;
		} else if ( _widec >= 35 )
			goto st134;
	} else
		goto st134;
	goto st0;
tr399:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st135;
tr400:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st135;
tr403:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st135;
tr390:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st135;
tr391:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st135;
tr392:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st135;
tr394:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st135;
tr454:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st135;
tr455:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st135;
tr456:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st135;
tr458:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st135;
tr465:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st135;
tr466:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st135;
tr467:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st135;
tr468:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st135;
st135:
	if ( ++p == pe )
		goto _test_eof135;
case 135:
#line 7267 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st135;
		case 32: goto st135;
		case 34: goto tr398;
		case 40: goto tr399;
		case 41: goto tr400;
		case 64: goto tr401;
		case 92: goto tr402;
		case 522: goto tr403;
		case 571: goto st148;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr396;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr396;
		} else if ( _widec >= 37 )
			goto tr396;
	} else
		goto tr396;
	goto st0;
tr396:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st136;
tr416:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st136;
st136:
	if ( ++p == pe )
		goto _test_eof136;
case 136:
#line 7317 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr406;
		case 32: goto tr406;
		case 40: goto tr407;
		case 41: goto tr408;
		case 92: goto st137;
		case 266: goto tr410;
		case 315: goto tr411;
		case 522: goto tr412;
		case 571: goto tr413;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st136;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st136;
		} else if ( _widec >= 35 )
			goto st136;
	} else
		goto st136;
	goto st0;
tr402:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st137;
tr420:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st137;
st137:
	if ( ++p == pe )
		goto _test_eof137;
case 137:
#line 7367 "./zscan.c"
	if ( (*p) == 10 )
		goto st138;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st139;
	goto st136;
st138:
	if ( ++p == pe )
		goto _test_eof138;
case 138:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr417;
		case 32: goto tr417;
		case 40: goto tr418;
		case 41: goto tr419;
		case 92: goto tr420;
		case 266: goto tr421;
		case 315: goto tr422;
		case 522: goto tr423;
		case 571: goto tr424;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr416;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr416;
		} else if ( _widec >= 35 )
			goto tr416;
	} else
		goto tr416;
	goto st0;
st139:
	if ( ++p == pe )
		goto _test_eof139;
case 139:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st140;
	goto st0;
st140:
	if ( ++p == pe )
		goto _test_eof140;
case 140:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st136;
	goto st0;
tr427:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st141;
tr398:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st141;
tr439:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st141;
tr440:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st141;
st141:
	if ( ++p == pe )
		goto _test_eof141;
case 141:
#line 7453 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr427;
		case 34: goto st142;
		case 92: goto st143;
	}
	goto st141;
tr441:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st142;
st142:
	if ( ++p == pe )
		goto _test_eof142;
case 142:
#line 7468 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr430;
		case 32: goto tr430;
		case 40: goto tr431;
		case 41: goto tr432;
		case 266: goto tr433;
		case 315: goto tr434;
		case 522: goto tr435;
		case 571: goto tr436;
	}
	goto st0;
tr442:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st143;
st143:
	if ( ++p == pe )
		goto _test_eof143;
case 143:
#line 7502 "./zscan.c"
	if ( (*p) == 10 )
		goto st144;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st145;
	goto st141;
st144:
	if ( ++p == pe )
		goto _test_eof144;
case 144:
	switch( (*p) ) {
		case 10: goto tr440;
		case 34: goto tr441;
		case 92: goto tr442;
	}
	goto tr439;
st145:
	if ( ++p == pe )
		goto _test_eof145;
case 145:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st146;
	goto st0;
st146:
	if ( ++p == pe )
		goto _test_eof146;
case 146:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st141;
	goto st0;
tr401:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st147;
st147:
	if ( ++p == pe )
		goto _test_eof147;
case 147:
#line 7543 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr444;
		case 32: goto tr444;
		case 40: goto tr445;
		case 41: goto tr446;
		case 266: goto tr447;
		case 315: goto tr448;
		case 522: goto tr449;
		case 571: goto tr450;
	}
	goto st0;
tr395:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st148;
tr459:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st148;
tr469:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st148;
st148:
	if ( ++p == pe )
		goto _test_eof148;
case 148:
#line 7587 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr403;
	if ( 384 <= _widec && _widec <= 639 )
		goto st148;
	goto st0;
tr386:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st149;
tr457:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st149;
st149:
	if ( ++p == pe )
		goto _test_eof149;
case 149:
#line 7626 "./zscan.c"
	if ( (*p) == 10 )
		goto st150;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st151;
	goto st134;
st150:
	if ( ++p == pe )
		goto _test_eof150;
case 150:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr454;
		case 32: goto tr454;
		case 40: goto tr455;
		case 41: goto tr456;
		case 92: goto tr457;
		case 522: goto tr458;
		case 571: goto tr459;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr453;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr453;
		} else if ( _widec >= 35 )
			goto tr453;
	} else
		goto tr453;
	goto st0;
st151:
	if ( ++p == pe )
		goto _test_eof151;
case 151:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st152;
	goto st0;
st152:
	if ( ++p == pe )
		goto _test_eof152;
case 152:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st134;
	goto st0;
tr462:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st153;
tr383:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st153;
tr472:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st153;
tr473:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st153;
st153:
	if ( ++p == pe )
		goto _test_eof153;
case 153:
#line 7710 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr462;
		case 34: goto st154;
		case 92: goto st155;
	}
	goto st153;
tr474:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st154;
st154:
	if ( ++p == pe )
		goto _test_eof154;
case 154:
#line 7725 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr465;
		case 32: goto tr465;
		case 40: goto tr466;
		case 41: goto tr467;
		case 522: goto tr468;
		case 571: goto tr469;
	}
	goto st0;
tr475:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st155;
st155:
	if ( ++p == pe )
		goto _test_eof155;
case 155:
#line 7757 "./zscan.c"
	if ( (*p) == 10 )
		goto st156;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st157;
	goto st153;
st156:
	if ( ++p == pe )
		goto _test_eof156;
case 156:
	switch( (*p) ) {
		case 10: goto tr473;
		case 34: goto tr474;
		case 92: goto tr475;
	}
	goto tr472;
st157:
	if ( ++p == pe )
		goto _test_eof157;
case 157:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st158;
	goto st0;
st158:
	if ( ++p == pe )
		goto _test_eof158;
case 158:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st153;
	goto st0;
tr380:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st159;
tr485:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st159;
tr495:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st159;
st159:
	if ( ++p == pe )
		goto _test_eof159;
case 159:
#line 7808 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr387;
	if ( 384 <= _widec && _widec <= 639 )
		goto st159;
	goto st0;
tr371:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st160;
tr483:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st160;
st160:
	if ( ++p == pe )
		goto _test_eof160;
case 160:
#line 7847 "./zscan.c"
	if ( (*p) == 10 )
		goto st161;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st162;
	goto st132;
st161:
	if ( ++p == pe )
		goto _test_eof161;
case 161:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr480;
		case 32: goto tr480;
		case 40: goto tr481;
		case 41: goto tr482;
		case 92: goto tr483;
		case 522: goto tr484;
		case 571: goto tr485;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr479;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr479;
		} else if ( _widec >= 35 )
			goto tr479;
	} else
		goto tr479;
	goto st0;
st162:
	if ( ++p == pe )
		goto _test_eof162;
case 162:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st163;
	goto st0;
st163:
	if ( ++p == pe )
		goto _test_eof163;
case 163:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st132;
	goto st0;
tr488:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st164;
tr368:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st164;
tr498:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st164;
tr499:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st164;
st164:
	if ( ++p == pe )
		goto _test_eof164;
case 164:
#line 7931 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr488;
		case 34: goto st165;
		case 92: goto st166;
	}
	goto st164;
tr500:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st165;
st165:
	if ( ++p == pe )
		goto _test_eof165;
case 165:
#line 7946 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr491;
		case 32: goto tr491;
		case 40: goto tr492;
		case 41: goto tr493;
		case 522: goto tr494;
		case 571: goto tr495;
	}
	goto st0;
tr501:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st166;
st166:
	if ( ++p == pe )
		goto _test_eof166;
case 166:
#line 7978 "./zscan.c"
	if ( (*p) == 10 )
		goto st167;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st168;
	goto st164;
st167:
	if ( ++p == pe )
		goto _test_eof167;
case 167:
	switch( (*p) ) {
		case 10: goto tr499;
		case 34: goto tr500;
		case 92: goto tr501;
	}
	goto tr498;
st168:
	if ( ++p == pe )
		goto _test_eof168;
case 168:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st169;
	goto st0;
st169:
	if ( ++p == pe )
		goto _test_eof169;
case 169:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st164;
	goto st0;
tr365:
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st170;
tr511:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 490 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, false); }
	goto st170;
tr521:
#line 491 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, false); }
	goto st170;
st170:
	if ( ++p == pe )
		goto _test_eof170;
case 170:
#line 8029 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr372;
	if ( 384 <= _widec && _widec <= 639 )
		goto st170;
	goto st0;
tr509:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st171;
tr356:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st171;
st171:
	if ( ++p == pe )
		goto _test_eof171;
case 171:
#line 8070 "./zscan.c"
	if ( (*p) == 10 )
		goto st172;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st173;
	goto st130;
st172:
	if ( ++p == pe )
		goto _test_eof172;
case 172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr506;
		case 32: goto tr506;
		case 40: goto tr507;
		case 41: goto tr508;
		case 92: goto tr509;
		case 522: goto tr510;
		case 571: goto tr511;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr505;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr505;
		} else if ( _widec >= 35 )
			goto tr505;
	} else
		goto tr505;
	goto st0;
st173:
	if ( ++p == pe )
		goto _test_eof173;
case 173:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st174;
	goto st0;
st174:
	if ( ++p == pe )
		goto _test_eof174;
case 174:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st130;
	goto st0;
tr514:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st175;
tr524:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st175;
tr525:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st175;
tr353:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st175;
st175:
	if ( ++p == pe )
		goto _test_eof175;
case 175:
#line 8156 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr514;
		case 34: goto st176;
		case 92: goto st177;
	}
	goto st175;
tr526:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st176;
st176:
	if ( ++p == pe )
		goto _test_eof176;
case 176:
#line 8171 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr517;
		case 32: goto tr517;
		case 40: goto tr518;
		case 41: goto tr519;
		case 522: goto tr520;
		case 571: goto tr521;
	}
	goto st0;
tr527:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st177;
st177:
	if ( ++p == pe )
		goto _test_eof177;
case 177:
#line 8203 "./zscan.c"
	if ( (*p) == 10 )
		goto st178;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st179;
	goto st175;
st178:
	if ( ++p == pe )
		goto _test_eof178;
case 178:
	switch( (*p) ) {
		case 10: goto tr525;
		case 34: goto tr526;
		case 92: goto tr527;
	}
	goto tr524;
st179:
	if ( ++p == pe )
		goto _test_eof179;
case 179:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st180;
	goto st0;
st180:
	if ( ++p == pe )
		goto _test_eof180;
case 180:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st175;
	goto st0;
tr350:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st181;
st181:
	if ( ++p == pe )
		goto _test_eof181;
case 181:
#line 8246 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr357;
	if ( 384 <= _widec && _widec <= 639 )
		goto st181;
	goto st0;
tr338:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st182;
st182:
	if ( ++p == pe )
		goto _test_eof182;
case 182:
#line 8283 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr343;
	if ( 384 <= _widec && _widec <= 639 )
		goto st182;
	goto st0;
st183:
	if ( ++p == pe )
		goto _test_eof183;
case 183:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr330;
	if ( 384 <= _widec && _widec <= 639 )
		goto st183;
	goto st0;
st184:
	if ( ++p == pe )
		goto _test_eof184;
case 184:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st185;
		case 32: goto st185;
		case 40: goto tr530;
		case 41: goto tr531;
		case 522: goto tr532;
		case 571: goto st198;
	}
	goto st0;
tr530:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st185;
tr531:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st185;
tr532:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st185;
st185:
	if ( ++p == pe )
		goto _test_eof185;
case 185:
#line 8383 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st185;
		case 32: goto st185;
		case 34: goto tr535;
		case 40: goto tr530;
		case 41: goto tr531;
		case 64: goto tr536;
		case 92: goto tr537;
		case 522: goto tr532;
		case 571: goto st198;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr534;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr534;
		} else if ( _widec >= 37 )
			goto tr534;
	} else
		goto tr534;
	goto st0;
tr534:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st186;
tr549:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st186;
st186:
	if ( ++p == pe )
		goto _test_eof186;
case 186:
#line 8433 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr539;
		case 32: goto tr539;
		case 40: goto tr540;
		case 41: goto tr541;
		case 92: goto st187;
		case 266: goto tr543;
		case 315: goto tr544;
		case 522: goto tr545;
		case 571: goto tr546;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st186;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st186;
		} else if ( _widec >= 35 )
			goto st186;
	} else
		goto st186;
	goto st0;
tr537:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st187;
tr553:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st187;
st187:
	if ( ++p == pe )
		goto _test_eof187;
case 187:
#line 8483 "./zscan.c"
	if ( (*p) == 10 )
		goto st188;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st189;
	goto st186;
st188:
	if ( ++p == pe )
		goto _test_eof188;
case 188:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr550;
		case 32: goto tr550;
		case 40: goto tr551;
		case 41: goto tr552;
		case 92: goto tr553;
		case 266: goto tr554;
		case 315: goto tr555;
		case 522: goto tr556;
		case 571: goto tr557;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr549;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr549;
		} else if ( _widec >= 35 )
			goto tr549;
	} else
		goto tr549;
	goto st0;
st189:
	if ( ++p == pe )
		goto _test_eof189;
case 189:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st190;
	goto st0;
st190:
	if ( ++p == pe )
		goto _test_eof190;
case 190:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st186;
	goto st0;
tr560:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st191;
tr535:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st191;
tr572:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st191;
tr573:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st191;
st191:
	if ( ++p == pe )
		goto _test_eof191;
case 191:
#line 8569 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr560;
		case 34: goto st192;
		case 92: goto st193;
	}
	goto st191;
tr574:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st192;
st192:
	if ( ++p == pe )
		goto _test_eof192;
case 192:
#line 8584 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr563;
		case 32: goto tr563;
		case 40: goto tr564;
		case 41: goto tr565;
		case 266: goto tr566;
		case 315: goto tr567;
		case 522: goto tr568;
		case 571: goto tr569;
	}
	goto st0;
tr575:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st193;
st193:
	if ( ++p == pe )
		goto _test_eof193;
case 193:
#line 8618 "./zscan.c"
	if ( (*p) == 10 )
		goto st194;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st195;
	goto st191;
st194:
	if ( ++p == pe )
		goto _test_eof194;
case 194:
	switch( (*p) ) {
		case 10: goto tr573;
		case 34: goto tr574;
		case 92: goto tr575;
	}
	goto tr572;
st195:
	if ( ++p == pe )
		goto _test_eof195;
case 195:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st196;
	goto st0;
st196:
	if ( ++p == pe )
		goto _test_eof196;
case 196:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st191;
	goto st0;
tr536:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st197;
st197:
	if ( ++p == pe )
		goto _test_eof197;
case 197:
#line 8659 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr577;
		case 32: goto tr577;
		case 40: goto tr578;
		case 41: goto tr579;
		case 266: goto tr580;
		case 315: goto tr581;
		case 522: goto tr582;
		case 571: goto tr583;
	}
	goto st0;
st198:
	if ( ++p == pe )
		goto _test_eof198;
case 198:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr532;
	if ( 384 <= _widec && _widec <= 639 )
		goto st198;
	goto st0;
tr18:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st199;
st199:
	if ( ++p == pe )
		goto _test_eof199;
case 199:
#line 8723 "./zscan.c"
	switch( (*p) ) {
		case 84: goto st200;
		case 116: goto st200;
	}
	goto st0;
st200:
	if ( ++p == pe )
		goto _test_eof200;
case 200:
	switch( (*p) ) {
		case 82: goto st201;
		case 114: goto st201;
	}
	goto st0;
st201:
	if ( ++p == pe )
		goto _test_eof201;
case 201:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st202;
		case 32: goto st202;
		case 40: goto tr587;
		case 41: goto tr588;
		case 522: goto tr589;
		case 571: goto st215;
	}
	goto st0;
tr587:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st202;
tr588:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st202;
tr589:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st202;
st202:
	if ( ++p == pe )
		goto _test_eof202;
case 202:
#line 8781 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st202;
		case 32: goto st202;
		case 34: goto tr592;
		case 40: goto tr587;
		case 41: goto tr588;
		case 64: goto tr593;
		case 92: goto tr594;
		case 522: goto tr589;
		case 571: goto st215;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr591;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr591;
		} else if ( _widec >= 37 )
			goto tr591;
	} else
		goto tr591;
	goto st0;
tr591:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st203;
tr606:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st203;
st203:
	if ( ++p == pe )
		goto _test_eof203;
case 203:
#line 8831 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr596;
		case 32: goto tr596;
		case 40: goto tr597;
		case 41: goto tr598;
		case 92: goto st204;
		case 266: goto tr600;
		case 315: goto tr601;
		case 522: goto tr602;
		case 571: goto tr603;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st203;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st203;
		} else if ( _widec >= 35 )
			goto st203;
	} else
		goto st203;
	goto st0;
tr594:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st204;
tr610:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st204;
st204:
	if ( ++p == pe )
		goto _test_eof204;
case 204:
#line 8881 "./zscan.c"
	if ( (*p) == 10 )
		goto st205;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st206;
	goto st203;
st205:
	if ( ++p == pe )
		goto _test_eof205;
case 205:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr607;
		case 32: goto tr607;
		case 40: goto tr608;
		case 41: goto tr609;
		case 92: goto tr610;
		case 266: goto tr611;
		case 315: goto tr612;
		case 522: goto tr613;
		case 571: goto tr614;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr606;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr606;
		} else if ( _widec >= 35 )
			goto tr606;
	} else
		goto tr606;
	goto st0;
st206:
	if ( ++p == pe )
		goto _test_eof206;
case 206:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st207;
	goto st0;
st207:
	if ( ++p == pe )
		goto _test_eof207;
case 207:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st203;
	goto st0;
tr617:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st208;
tr592:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st208;
tr629:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st208;
tr630:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st208;
st208:
	if ( ++p == pe )
		goto _test_eof208;
case 208:
#line 8967 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr617;
		case 34: goto st209;
		case 92: goto st210;
	}
	goto st208;
tr631:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st209;
st209:
	if ( ++p == pe )
		goto _test_eof209;
case 209:
#line 8982 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr620;
		case 32: goto tr620;
		case 40: goto tr621;
		case 41: goto tr622;
		case 266: goto tr623;
		case 315: goto tr624;
		case 522: goto tr625;
		case 571: goto tr626;
	}
	goto st0;
tr632:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st210;
st210:
	if ( ++p == pe )
		goto _test_eof210;
case 210:
#line 9016 "./zscan.c"
	if ( (*p) == 10 )
		goto st211;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st212;
	goto st208;
st211:
	if ( ++p == pe )
		goto _test_eof211;
case 211:
	switch( (*p) ) {
		case 10: goto tr630;
		case 34: goto tr631;
		case 92: goto tr632;
	}
	goto tr629;
st212:
	if ( ++p == pe )
		goto _test_eof212;
case 212:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st213;
	goto st0;
st213:
	if ( ++p == pe )
		goto _test_eof213;
case 213:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st208;
	goto st0;
tr593:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st214;
st214:
	if ( ++p == pe )
		goto _test_eof214;
case 214:
#line 9057 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr634;
		case 32: goto tr634;
		case 40: goto tr635;
		case 41: goto tr636;
		case 266: goto tr637;
		case 315: goto tr638;
		case 522: goto tr639;
		case 571: goto tr640;
	}
	goto st0;
st215:
	if ( ++p == pe )
		goto _test_eof215;
case 215:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr589;
	if ( 384 <= _widec && _widec <= 639 )
		goto st215;
	goto st0;
tr19:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st216;
st216:
	if ( ++p == pe )
		goto _test_eof216;
case 216:
#line 9121 "./zscan.c"
	switch( (*p) ) {
		case 79: goto st217;
		case 80: goto st266;
		case 82: goto st303;
		case 111: goto st217;
		case 112: goto st266;
		case 114: goto st303;
	}
	goto st0;
st217:
	if ( ++p == pe )
		goto _test_eof217;
case 217:
	switch( (*p) ) {
		case 65: goto st218;
		case 97: goto st218;
	}
	goto st0;
st218:
	if ( ++p == pe )
		goto _test_eof218;
case 218:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st219;
		case 32: goto st219;
		case 40: goto tr646;
		case 41: goto tr647;
		case 522: goto tr648;
		case 571: goto st265;
	}
	goto st0;
tr646:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st219;
tr647:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st219;
tr648:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st219;
st219:
	if ( ++p == pe )
		goto _test_eof219;
case 219:
#line 9183 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st219;
		case 32: goto st219;
		case 34: goto tr651;
		case 40: goto tr646;
		case 41: goto tr647;
		case 64: goto tr652;
		case 92: goto tr653;
		case 522: goto tr648;
		case 571: goto st265;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr650;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr650;
		} else if ( _widec >= 37 )
			goto tr650;
	} else
		goto tr650;
	goto st0;
tr650:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st220;
tr798:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st220;
st220:
	if ( ++p == pe )
		goto _test_eof220;
case 220:
#line 9233 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr655;
		case 32: goto tr655;
		case 40: goto tr656;
		case 41: goto tr657;
		case 92: goto st254;
		case 522: goto tr659;
		case 571: goto tr660;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st220;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st220;
		} else if ( _widec >= 35 )
			goto st220;
	} else
		goto st220;
	goto st0;
tr664:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st221;
tr665:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st221;
tr667:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st221;
tr655:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
	goto st221;
tr656:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st221;
tr657:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st221;
tr659:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st221;
tr799:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
	goto st221;
tr800:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st221;
tr801:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st221;
tr803:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st221;
tr810:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
	goto st221;
tr811:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st221;
tr812:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st221;
tr813:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st221;
tr822:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
	goto st221;
tr823:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st221;
tr824:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st221;
tr825:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st221;
st221:
	if ( ++p == pe )
		goto _test_eof221;
case 221:
#line 9381 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st221;
		case 32: goto st221;
		case 34: goto tr663;
		case 40: goto tr664;
		case 41: goto tr665;
		case 92: goto tr666;
		case 522: goto tr667;
		case 571: goto st253;
	}
	if ( _widec < 37 ) {
		if ( _widec > 8 ) {
			if ( 11 <= _widec && _widec <= 35 )
				goto tr661;
		} else
			goto tr661;
	} else if ( _widec > 58 ) {
		if ( _widec > 63 ) {
			if ( 65 <= _widec )
				goto tr661;
		} else if ( _widec >= 60 )
			goto tr661;
	} else
		goto tr661;
	goto st0;
tr661:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st222;
tr772:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st222;
st222:
	if ( ++p == pe )
		goto _test_eof222;
case 222:
#line 9433 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr670;
		case 32: goto tr670;
		case 40: goto tr671;
		case 41: goto tr672;
		case 92: goto st243;
		case 522: goto tr674;
		case 571: goto tr675;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st222;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st222;
		} else if ( _widec >= 35 )
			goto st222;
	} else
		goto st222;
	goto st0;
tr677:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st223;
tr678:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st223;
tr680:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st223;
tr670:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
	goto st223;
tr671:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st223;
tr672:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st223;
tr674:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st223;
tr773:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
	goto st223;
tr774:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st223;
tr775:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st223;
tr777:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st223;
tr784:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
	goto st223;
tr785:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st223;
tr786:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st223;
tr787:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st223;
st223:
	if ( ++p == pe )
		goto _test_eof223;
case 223:
#line 9559 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st223;
		case 32: goto st223;
		case 40: goto tr677;
		case 41: goto tr678;
		case 522: goto tr680;
		case 571: goto st242;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr679;
	goto st0;
tr679:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st224;
st224:
	if ( ++p == pe )
		goto _test_eof224;
case 224:
#line 9593 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr682;
		case 32: goto tr682;
		case 40: goto tr683;
		case 41: goto tr684;
		case 68: goto tr686;
		case 72: goto tr686;
		case 77: goto tr686;
		case 87: goto tr686;
		case 100: goto tr686;
		case 104: goto tr686;
		case 109: goto tr686;
		case 119: goto tr686;
		case 522: goto tr687;
		case 571: goto tr688;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st224;
	goto st0;
tr690:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st225;
tr691:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st225;
tr693:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st225;
tr682:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st225;
tr683:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st225;
tr684:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st225;
tr687:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st225;
tr765:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st225;
tr766:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st225;
tr767:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st225;
tr768:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st225;
st225:
	if ( ++p == pe )
		goto _test_eof225;
case 225:
#line 9695 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st225;
		case 32: goto st225;
		case 40: goto tr690;
		case 41: goto tr691;
		case 522: goto tr693;
		case 571: goto st240;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr692;
	goto st0;
tr692:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st226;
st226:
	if ( ++p == pe )
		goto _test_eof226;
case 226:
#line 9729 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr695;
		case 32: goto tr695;
		case 40: goto tr696;
		case 41: goto tr697;
		case 68: goto tr699;
		case 72: goto tr699;
		case 77: goto tr699;
		case 87: goto tr699;
		case 100: goto tr699;
		case 104: goto tr699;
		case 109: goto tr699;
		case 119: goto tr699;
		case 522: goto tr700;
		case 571: goto tr701;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st226;
	goto st0;
tr703:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st227;
tr704:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st227;
tr706:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st227;
tr695:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st227;
tr696:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st227;
tr697:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st227;
tr700:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st227;
tr760:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st227;
tr761:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st227;
tr762:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st227;
tr763:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st227;
st227:
	if ( ++p == pe )
		goto _test_eof227;
case 227:
#line 9831 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st227;
		case 32: goto st227;
		case 40: goto tr703;
		case 41: goto tr704;
		case 522: goto tr706;
		case 571: goto st238;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr705;
	goto st0;
tr705:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st228;
st228:
	if ( ++p == pe )
		goto _test_eof228;
case 228:
#line 9865 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr708;
		case 32: goto tr708;
		case 40: goto tr709;
		case 41: goto tr710;
		case 68: goto tr712;
		case 72: goto tr712;
		case 77: goto tr712;
		case 87: goto tr712;
		case 100: goto tr712;
		case 104: goto tr712;
		case 109: goto tr712;
		case 119: goto tr712;
		case 522: goto tr713;
		case 571: goto tr714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st228;
	goto st0;
tr716:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st229;
tr717:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st229;
tr719:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st229;
tr708:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st229;
tr709:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st229;
tr710:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st229;
tr713:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st229;
tr755:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st229;
tr756:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st229;
tr757:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st229;
tr758:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st229;
st229:
	if ( ++p == pe )
		goto _test_eof229;
case 229:
#line 9967 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st229;
		case 32: goto st229;
		case 40: goto tr716;
		case 41: goto tr717;
		case 522: goto tr719;
		case 571: goto st236;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr718;
	goto st0;
tr718:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st230;
st230:
	if ( ++p == pe )
		goto _test_eof230;
case 230:
#line 10001 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr721;
		case 32: goto tr721;
		case 40: goto tr722;
		case 41: goto tr723;
		case 68: goto tr725;
		case 72: goto tr725;
		case 77: goto tr725;
		case 87: goto tr725;
		case 100: goto tr725;
		case 104: goto tr725;
		case 109: goto tr725;
		case 119: goto tr725;
		case 522: goto tr726;
		case 571: goto tr727;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st230;
	goto st0;
tr729:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st231;
tr730:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st231;
tr732:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st231;
tr721:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
	goto st231;
tr722:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st231;
tr723:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st231;
tr726:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st231;
tr750:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
	goto st231;
tr751:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st231;
tr752:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st231;
tr753:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st231;
st231:
	if ( ++p == pe )
		goto _test_eof231;
case 231:
#line 10103 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st231;
		case 32: goto st231;
		case 40: goto tr729;
		case 41: goto tr730;
		case 522: goto tr732;
		case 571: goto st234;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr731;
	goto st0;
tr731:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st232;
st232:
	if ( ++p == pe )
		goto _test_eof232;
case 232:
#line 10137 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr734;
		case 32: goto tr734;
		case 40: goto tr735;
		case 41: goto tr736;
		case 68: goto tr738;
		case 72: goto tr738;
		case 77: goto tr738;
		case 87: goto tr738;
		case 100: goto tr738;
		case 104: goto tr738;
		case 109: goto tr738;
		case 119: goto tr738;
		case 266: goto tr739;
		case 315: goto tr740;
		case 522: goto tr741;
		case 571: goto tr742;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st232;
	goto st0;
tr738:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st233;
st233:
	if ( ++p == pe )
		goto _test_eof233;
case 233:
#line 10183 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr743;
		case 32: goto tr743;
		case 40: goto tr744;
		case 41: goto tr745;
		case 266: goto tr746;
		case 315: goto tr747;
		case 522: goto tr748;
		case 571: goto tr749;
	}
	goto st0;
tr727:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
	goto st234;
tr754:
#line 504 "./zscan.rl"
	{ z->uv_4 = z->uval; }
	goto st234;
st234:
	if ( ++p == pe )
		goto _test_eof234;
case 234:
#line 10223 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr732;
	if ( 384 <= _widec && _widec <= 639 )
		goto st234;
	goto st0;
tr725:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st235;
st235:
	if ( ++p == pe )
		goto _test_eof235;
case 235:
#line 10260 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr750;
		case 32: goto tr750;
		case 40: goto tr751;
		case 41: goto tr752;
		case 522: goto tr753;
		case 571: goto tr754;
	}
	goto st0;
tr714:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st236;
tr759:
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st236;
st236:
	if ( ++p == pe )
		goto _test_eof236;
case 236:
#line 10298 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr719;
	if ( 384 <= _widec && _widec <= 639 )
		goto st236;
	goto st0;
tr712:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st237;
st237:
	if ( ++p == pe )
		goto _test_eof237;
case 237:
#line 10335 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr755;
		case 32: goto tr755;
		case 40: goto tr756;
		case 41: goto tr757;
		case 522: goto tr758;
		case 571: goto tr759;
	}
	goto st0;
tr701:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st238;
tr764:
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st238;
st238:
	if ( ++p == pe )
		goto _test_eof238;
case 238:
#line 10373 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr706;
	if ( 384 <= _widec && _widec <= 639 )
		goto st238;
	goto st0;
tr699:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st239;
st239:
	if ( ++p == pe )
		goto _test_eof239;
case 239:
#line 10410 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr760;
		case 32: goto tr760;
		case 40: goto tr761;
		case 41: goto tr762;
		case 522: goto tr763;
		case 571: goto tr764;
	}
	goto st0;
tr688:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st240;
tr769:
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st240;
st240:
	if ( ++p == pe )
		goto _test_eof240;
case 240:
#line 10448 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr693;
	if ( 384 <= _widec && _widec <= 639 )
		goto st240;
	goto st0;
tr686:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st241;
st241:
	if ( ++p == pe )
		goto _test_eof241;
case 241:
#line 10485 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr765;
		case 32: goto tr765;
		case 40: goto tr766;
		case 41: goto tr767;
		case 522: goto tr768;
		case 571: goto tr769;
	}
	goto st0;
tr675:
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
	goto st242;
tr778:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 474 "./zscan.rl"
	{ dname_set(z, z->eml_dname, p - z->tstart, false); }
	goto st242;
tr788:
#line 475 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->eml_dname, p - z->tstart - 1, false); }
	goto st242;
st242:
	if ( ++p == pe )
		goto _test_eof242;
case 242:
#line 10527 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr680;
	if ( 384 <= _widec && _widec <= 639 )
		goto st242;
	goto st0;
tr666:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st243;
tr776:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st243;
st243:
	if ( ++p == pe )
		goto _test_eof243;
case 243:
#line 10566 "./zscan.c"
	if ( (*p) == 10 )
		goto st244;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st245;
	goto st222;
st244:
	if ( ++p == pe )
		goto _test_eof244;
case 244:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr773;
		case 32: goto tr773;
		case 40: goto tr774;
		case 41: goto tr775;
		case 92: goto tr776;
		case 522: goto tr777;
		case 571: goto tr778;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr772;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr772;
		} else if ( _widec >= 35 )
			goto tr772;
	} else
		goto tr772;
	goto st0;
st245:
	if ( ++p == pe )
		goto _test_eof245;
case 245:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st246;
	goto st0;
st246:
	if ( ++p == pe )
		goto _test_eof246;
case 246:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st222;
	goto st0;
tr781:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st247;
tr663:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st247;
tr791:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st247;
tr792:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st247;
st247:
	if ( ++p == pe )
		goto _test_eof247;
case 247:
#line 10650 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr781;
		case 34: goto st248;
		case 92: goto st249;
	}
	goto st247;
tr793:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st248;
st248:
	if ( ++p == pe )
		goto _test_eof248;
case 248:
#line 10665 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr784;
		case 32: goto tr784;
		case 40: goto tr785;
		case 41: goto tr786;
		case 522: goto tr787;
		case 571: goto tr788;
	}
	goto st0;
tr794:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st249;
st249:
	if ( ++p == pe )
		goto _test_eof249;
case 249:
#line 10697 "./zscan.c"
	if ( (*p) == 10 )
		goto st250;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st251;
	goto st247;
st250:
	if ( ++p == pe )
		goto _test_eof250;
case 250:
	switch( (*p) ) {
		case 10: goto tr792;
		case 34: goto tr793;
		case 92: goto tr794;
	}
	goto tr791;
st251:
	if ( ++p == pe )
		goto _test_eof251;
case 251:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st252;
	goto st0;
st252:
	if ( ++p == pe )
		goto _test_eof252;
case 252:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st247;
	goto st0;
tr660:
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
	goto st253;
tr804:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 472 "./zscan.rl"
	{ dname_set(z, z->rhs_dname, p - z->tstart, false); }
	goto st253;
tr814:
#line 473 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->rhs_dname, p - z->tstart - 1, false); }
	goto st253;
tr826:
#line 471 "./zscan.rl"
	{ dname_copy(z->rhs_dname, z->origin); }
	goto st253;
st253:
	if ( ++p == pe )
		goto _test_eof253;
case 253:
#line 10752 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr667;
	if ( 384 <= _widec && _widec <= 639 )
		goto st253;
	goto st0;
tr653:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st254;
tr802:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st254;
st254:
	if ( ++p == pe )
		goto _test_eof254;
case 254:
#line 10791 "./zscan.c"
	if ( (*p) == 10 )
		goto st255;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st256;
	goto st220;
st255:
	if ( ++p == pe )
		goto _test_eof255;
case 255:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr799;
		case 32: goto tr799;
		case 40: goto tr800;
		case 41: goto tr801;
		case 92: goto tr802;
		case 522: goto tr803;
		case 571: goto tr804;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr798;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr798;
		} else if ( _widec >= 35 )
			goto tr798;
	} else
		goto tr798;
	goto st0;
st256:
	if ( ++p == pe )
		goto _test_eof256;
case 256:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st257;
	goto st0;
st257:
	if ( ++p == pe )
		goto _test_eof257;
case 257:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st220;
	goto st0;
tr807:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st258;
tr651:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st258;
tr817:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st258;
tr818:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st258;
st258:
	if ( ++p == pe )
		goto _test_eof258;
case 258:
#line 10875 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr807;
		case 34: goto st259;
		case 92: goto st260;
	}
	goto st258;
tr819:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st259;
st259:
	if ( ++p == pe )
		goto _test_eof259;
case 259:
#line 10890 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr810;
		case 32: goto tr810;
		case 40: goto tr811;
		case 41: goto tr812;
		case 522: goto tr813;
		case 571: goto tr814;
	}
	goto st0;
tr820:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st260;
st260:
	if ( ++p == pe )
		goto _test_eof260;
case 260:
#line 10922 "./zscan.c"
	if ( (*p) == 10 )
		goto st261;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st262;
	goto st258;
st261:
	if ( ++p == pe )
		goto _test_eof261;
case 261:
	switch( (*p) ) {
		case 10: goto tr818;
		case 34: goto tr819;
		case 92: goto tr820;
	}
	goto tr817;
st262:
	if ( ++p == pe )
		goto _test_eof262;
case 262:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st263;
	goto st0;
st263:
	if ( ++p == pe )
		goto _test_eof263;
case 263:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st258;
	goto st0;
tr652:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st264;
st264:
	if ( ++p == pe )
		goto _test_eof264;
case 264:
#line 10963 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr822;
		case 32: goto tr822;
		case 40: goto tr823;
		case 41: goto tr824;
		case 522: goto tr825;
		case 571: goto tr826;
	}
	goto st0;
st265:
	if ( ++p == pe )
		goto _test_eof265;
case 265:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr648;
	if ( 384 <= _widec && _widec <= 639 )
		goto st265;
	goto st0;
st266:
	if ( ++p == pe )
		goto _test_eof266;
case 266:
	switch( (*p) ) {
		case 70: goto st267;
		case 102: goto st267;
	}
	goto st0;
st267:
	if ( ++p == pe )
		goto _test_eof267;
case 267:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st268;
		case 32: goto st268;
		case 40: goto tr829;
		case 41: goto tr830;
		case 43: goto st284;
		case 522: goto tr832;
		case 571: goto st302;
	}
	goto st0;
tr829:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st268;
tr830:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st268;
tr832:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st268;
st268:
	if ( ++p == pe )
		goto _test_eof268;
case 268:
#line 11070 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr835;
		case 32: goto tr835;
		case 34: goto tr836;
		case 40: goto tr837;
		case 41: goto tr838;
		case 92: goto tr839;
		case 522: goto tr840;
		case 571: goto tr841;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr834;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr834;
	} else
		goto tr834;
	goto st0;
tr852:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st269;
tr878:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st269;
tr834:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st269;
tr866:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st269;
st269:
	if ( ++p == pe )
		goto _test_eof269;
case 269:
#line 11128 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr843;
		case 32: goto tr843;
		case 34: goto tr844;
		case 40: goto tr845;
		case 41: goto tr846;
		case 92: goto st273;
		case 266: goto tr848;
		case 315: goto tr849;
		case 522: goto tr850;
		case 571: goto tr851;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st269;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st269;
	} else
		goto st269;
	goto st0;
tr855:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st270;
tr856:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st270;
tr860:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st270;
tr843:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st270;
tr845:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st270;
tr846:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st270;
tr850:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st270;
tr867:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st270;
tr869:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st270;
tr870:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st270;
tr874:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st270;
tr879:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st270;
tr881:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st270;
tr882:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st270;
tr886:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st270;
st270:
	if ( ++p == pe )
		goto _test_eof270;
case 270:
#line 11254 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st270;
		case 32: goto st270;
		case 34: goto tr854;
		case 40: goto tr855;
		case 41: goto tr856;
		case 92: goto tr857;
		case 266: goto tr858;
		case 315: goto tr859;
		case 522: goto tr860;
		case 571: goto st275;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr852;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr852;
	} else
		goto tr852;
	goto st0;
tr863:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st271;
tr854:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st271;
tr891:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st271;
tr892:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st271;
tr836:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st271;
tr844:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st271;
tr868:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st271;
tr880:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st271;
st271:
	if ( ++p == pe )
		goto _test_eof271;
case 271:
#line 11338 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr863;
		case 34: goto st272;
		case 92: goto st278;
	}
	goto st271;
tr893:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st272;
st272:
	if ( ++p == pe )
		goto _test_eof272;
case 272:
#line 11353 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr867;
		case 32: goto tr867;
		case 34: goto tr868;
		case 40: goto tr869;
		case 41: goto tr870;
		case 92: goto tr871;
		case 266: goto tr872;
		case 315: goto tr873;
		case 522: goto tr874;
		case 571: goto tr875;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr866;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr866;
	} else
		goto tr866;
	goto st0;
tr857:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st273;
tr883:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st273;
tr839:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st273;
tr871:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st273;
st273:
	if ( ++p == pe )
		goto _test_eof273;
case 273:
#line 11413 "./zscan.c"
	if ( (*p) == 10 )
		goto st274;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st276;
	goto st269;
st274:
	if ( ++p == pe )
		goto _test_eof274;
case 274:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr879;
		case 32: goto tr879;
		case 34: goto tr880;
		case 40: goto tr881;
		case 41: goto tr882;
		case 92: goto tr883;
		case 266: goto tr884;
		case 315: goto tr885;
		case 522: goto tr886;
		case 571: goto tr887;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr878;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr878;
	} else
		goto tr878;
	goto st0;
tr851:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st275;
tr875:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st275;
tr887:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st275;
st275:
	if ( ++p == pe )
		goto _test_eof275;
case 275:
#line 11479 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr860;
	if ( 384 <= _widec && _widec <= 639 )
		goto st275;
	goto st0;
st276:
	if ( ++p == pe )
		goto _test_eof276;
case 276:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st277;
	goto st0;
st277:
	if ( ++p == pe )
		goto _test_eof277;
case 277:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st269;
	goto st0;
tr894:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st278;
st278:
	if ( ++p == pe )
		goto _test_eof278;
case 278:
#line 11528 "./zscan.c"
	if ( (*p) == 10 )
		goto st279;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st280;
	goto st271;
st279:
	if ( ++p == pe )
		goto _test_eof279;
case 279:
	switch( (*p) ) {
		case 10: goto tr892;
		case 34: goto tr893;
		case 92: goto tr894;
	}
	goto tr891;
st280:
	if ( ++p == pe )
		goto _test_eof280;
case 280:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st281;
	goto st0;
st281:
	if ( ++p == pe )
		goto _test_eof281;
case 281:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st271;
	goto st0;
tr897:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st282;
tr835:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st282;
tr837:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st282;
tr838:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st282;
tr840:
#line 546 "./zscan.rl"
	{z->lcount++;}
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st282;
st282:
	if ( ++p == pe )
		goto _test_eof282;
case 282:
#line 11591 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr835;
		case 32: goto tr835;
		case 34: goto tr836;
		case 40: goto tr837;
		case 41: goto tr838;
		case 92: goto tr839;
		case 266: goto tr858;
		case 315: goto tr859;
		case 522: goto tr840;
		case 571: goto tr841;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr834;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr834;
	} else
		goto tr834;
	goto st0;
tr841:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st283;
st283:
	if ( ++p == pe )
		goto _test_eof283;
case 283:
#line 11635 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr897;
	if ( 384 <= _widec && _widec <= 639 )
		goto st283;
	goto st0;
st284:
	if ( ++p == pe )
		goto _test_eof284;
case 284:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st285;
		case 32: goto st285;
		case 40: goto tr899;
		case 41: goto tr900;
		case 522: goto tr901;
		case 571: goto st301;
	}
	goto st0;
tr899:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st285;
tr900:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st285;
tr901:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st285;
st285:
	if ( ++p == pe )
		goto _test_eof285;
case 285:
#line 11705 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr904;
		case 32: goto tr904;
		case 34: goto tr905;
		case 40: goto tr906;
		case 41: goto tr907;
		case 92: goto tr908;
		case 522: goto tr909;
		case 571: goto tr910;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr903;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr903;
	} else
		goto tr903;
	goto st0;
tr921:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st286;
tr947:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st286;
tr903:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st286;
tr935:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st286;
st286:
	if ( ++p == pe )
		goto _test_eof286;
case 286:
#line 11763 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr912;
		case 32: goto tr912;
		case 34: goto tr913;
		case 40: goto tr914;
		case 41: goto tr915;
		case 92: goto st290;
		case 266: goto tr917;
		case 315: goto tr918;
		case 522: goto tr919;
		case 571: goto tr920;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st286;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st286;
	} else
		goto st286;
	goto st0;
tr924:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st287;
tr925:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st287;
tr929:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st287;
tr912:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st287;
tr914:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st287;
tr915:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st287;
tr919:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st287;
tr936:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st287;
tr938:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st287;
tr939:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st287;
tr943:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st287;
tr948:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st287;
tr950:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st287;
tr951:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st287;
tr955:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st287;
st287:
	if ( ++p == pe )
		goto _test_eof287;
case 287:
#line 11889 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st287;
		case 32: goto st287;
		case 34: goto tr923;
		case 40: goto tr924;
		case 41: goto tr925;
		case 92: goto tr926;
		case 266: goto tr927;
		case 315: goto tr928;
		case 522: goto tr929;
		case 571: goto st292;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr921;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr921;
	} else
		goto tr921;
	goto st0;
tr932:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st288;
tr923:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st288;
tr960:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st288;
tr961:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st288;
tr905:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st288;
tr913:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st288;
tr937:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st288;
tr949:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st288;
st288:
	if ( ++p == pe )
		goto _test_eof288;
case 288:
#line 11973 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr932;
		case 34: goto st289;
		case 92: goto st295;
	}
	goto st288;
tr962:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st289;
st289:
	if ( ++p == pe )
		goto _test_eof289;
case 289:
#line 11988 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr936;
		case 32: goto tr936;
		case 34: goto tr937;
		case 40: goto tr938;
		case 41: goto tr939;
		case 92: goto tr940;
		case 266: goto tr941;
		case 315: goto tr942;
		case 522: goto tr943;
		case 571: goto tr944;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr935;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr935;
	} else
		goto tr935;
	goto st0;
tr926:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st290;
tr952:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st290;
tr908:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st290;
tr940:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st290;
st290:
	if ( ++p == pe )
		goto _test_eof290;
case 290:
#line 12048 "./zscan.c"
	if ( (*p) == 10 )
		goto st291;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st293;
	goto st286;
st291:
	if ( ++p == pe )
		goto _test_eof291;
case 291:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr948;
		case 32: goto tr948;
		case 34: goto tr949;
		case 40: goto tr950;
		case 41: goto tr951;
		case 92: goto tr952;
		case 266: goto tr953;
		case 315: goto tr954;
		case 522: goto tr955;
		case 571: goto tr956;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr947;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr947;
	} else
		goto tr947;
	goto st0;
tr920:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st292;
tr944:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st292;
tr956:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st292;
st292:
	if ( ++p == pe )
		goto _test_eof292;
case 292:
#line 12114 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr929;
	if ( 384 <= _widec && _widec <= 639 )
		goto st292;
	goto st0;
st293:
	if ( ++p == pe )
		goto _test_eof293;
case 293:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st294;
	goto st0;
st294:
	if ( ++p == pe )
		goto _test_eof294;
case 294:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st286;
	goto st0;
tr963:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st295;
st295:
	if ( ++p == pe )
		goto _test_eof295;
case 295:
#line 12163 "./zscan.c"
	if ( (*p) == 10 )
		goto st296;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st297;
	goto st288;
st296:
	if ( ++p == pe )
		goto _test_eof296;
case 296:
	switch( (*p) ) {
		case 10: goto tr961;
		case 34: goto tr962;
		case 92: goto tr963;
	}
	goto tr960;
st297:
	if ( ++p == pe )
		goto _test_eof297;
case 297:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st298;
	goto st0;
st298:
	if ( ++p == pe )
		goto _test_eof298;
case 298:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st288;
	goto st0;
tr966:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st299;
tr904:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st299;
tr906:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st299;
tr907:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st299;
tr909:
#line 546 "./zscan.rl"
	{z->lcount++;}
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st299;
st299:
	if ( ++p == pe )
		goto _test_eof299;
case 299:
#line 12226 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr904;
		case 32: goto tr904;
		case 34: goto tr905;
		case 40: goto tr906;
		case 41: goto tr907;
		case 92: goto tr908;
		case 266: goto tr927;
		case 315: goto tr928;
		case 522: goto tr909;
		case 571: goto tr910;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr903;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr903;
	} else
		goto tr903;
	goto st0;
tr910:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st300;
st300:
	if ( ++p == pe )
		goto _test_eof300;
case 300:
#line 12270 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr966;
	if ( 384 <= _widec && _widec <= 639 )
		goto st300;
	goto st0;
st301:
	if ( ++p == pe )
		goto _test_eof301;
case 301:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr901;
	if ( 384 <= _widec && _widec <= 639 )
		goto st301;
	goto st0;
st302:
	if ( ++p == pe )
		goto _test_eof302;
case 302:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr832;
	if ( 384 <= _widec && _widec <= 639 )
		goto st302;
	goto st0;
st303:
	if ( ++p == pe )
		goto _test_eof303;
case 303:
	switch( (*p) ) {
		case 86: goto st304;
		case 118: goto st304;
	}
	goto st0;
st304:
	if ( ++p == pe )
		goto _test_eof304;
case 304:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st305;
		case 32: goto st305;
		case 40: goto tr969;
		case 41: goto tr970;
		case 522: goto tr971;
		case 571: goto st327;
	}
	goto st0;
tr969:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st305;
tr970:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st305;
tr971:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st305;
st305:
	if ( ++p == pe )
		goto _test_eof305;
case 305:
#line 12409 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st305;
		case 32: goto st305;
		case 40: goto tr969;
		case 41: goto tr970;
		case 522: goto tr971;
		case 571: goto st327;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr973;
	goto st0;
tr973:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st306;
st306:
	if ( ++p == pe )
		goto _test_eof306;
case 306:
#line 12443 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr974;
		case 32: goto tr974;
		case 40: goto tr975;
		case 41: goto tr976;
		case 522: goto tr978;
		case 571: goto tr979;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st306;
	goto st0;
tr981:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st307;
tr982:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st307;
tr984:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st307;
tr974:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st307;
tr975:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st307;
tr976:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st307;
tr978:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st307;
st307:
	if ( ++p == pe )
		goto _test_eof307;
case 307:
#line 12515 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st307;
		case 32: goto st307;
		case 40: goto tr981;
		case 41: goto tr982;
		case 522: goto tr984;
		case 571: goto st326;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr983;
	goto st0;
tr983:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st308;
st308:
	if ( ++p == pe )
		goto _test_eof308;
case 308:
#line 12549 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr986;
		case 32: goto tr986;
		case 40: goto tr987;
		case 41: goto tr988;
		case 522: goto tr990;
		case 571: goto tr991;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st308;
	goto st0;
tr993:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st309;
tr994:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st309;
tr996:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st309;
tr986:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st309;
tr987:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st309;
tr988:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st309;
tr990:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st309;
st309:
	if ( ++p == pe )
		goto _test_eof309;
case 309:
#line 12621 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st309;
		case 32: goto st309;
		case 40: goto tr993;
		case 41: goto tr994;
		case 522: goto tr996;
		case 571: goto st325;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr995;
	goto st0;
tr995:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st310;
st310:
	if ( ++p == pe )
		goto _test_eof310;
case 310:
#line 12655 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr998;
		case 32: goto tr998;
		case 40: goto tr999;
		case 41: goto tr1000;
		case 522: goto tr1002;
		case 571: goto tr1003;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st310;
	goto st0;
tr1007:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st311;
tr1008:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st311;
tr1011:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st311;
tr998:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st311;
tr999:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st311;
tr1000:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st311;
tr1002:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st311;
st311:
	if ( ++p == pe )
		goto _test_eof311;
case 311:
#line 12727 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st311;
		case 32: goto st311;
		case 34: goto tr1006;
		case 40: goto tr1007;
		case 41: goto tr1008;
		case 64: goto tr1009;
		case 92: goto tr1010;
		case 522: goto tr1011;
		case 571: goto st324;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1004;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1004;
		} else if ( _widec >= 37 )
			goto tr1004;
	} else
		goto tr1004;
	goto st0;
tr1004:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st312;
tr1024:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st312;
st312:
	if ( ++p == pe )
		goto _test_eof312;
case 312:
#line 12777 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1014;
		case 32: goto tr1014;
		case 40: goto tr1015;
		case 41: goto tr1016;
		case 92: goto st313;
		case 266: goto tr1018;
		case 315: goto tr1019;
		case 522: goto tr1020;
		case 571: goto tr1021;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st312;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st312;
		} else if ( _widec >= 35 )
			goto st312;
	} else
		goto st312;
	goto st0;
tr1010:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st313;
tr1028:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st313;
st313:
	if ( ++p == pe )
		goto _test_eof313;
case 313:
#line 12827 "./zscan.c"
	if ( (*p) == 10 )
		goto st314;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st315;
	goto st312;
st314:
	if ( ++p == pe )
		goto _test_eof314;
case 314:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1025;
		case 32: goto tr1025;
		case 40: goto tr1026;
		case 41: goto tr1027;
		case 92: goto tr1028;
		case 266: goto tr1029;
		case 315: goto tr1030;
		case 522: goto tr1031;
		case 571: goto tr1032;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1024;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1024;
		} else if ( _widec >= 35 )
			goto tr1024;
	} else
		goto tr1024;
	goto st0;
st315:
	if ( ++p == pe )
		goto _test_eof315;
case 315:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st316;
	goto st0;
st316:
	if ( ++p == pe )
		goto _test_eof316;
case 316:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st312;
	goto st0;
tr1035:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st317;
tr1006:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st317;
tr1047:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st317;
tr1048:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st317;
st317:
	if ( ++p == pe )
		goto _test_eof317;
case 317:
#line 12913 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1035;
		case 34: goto st318;
		case 92: goto st319;
	}
	goto st317;
tr1049:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st318;
st318:
	if ( ++p == pe )
		goto _test_eof318;
case 318:
#line 12928 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1038;
		case 32: goto tr1038;
		case 40: goto tr1039;
		case 41: goto tr1040;
		case 266: goto tr1041;
		case 315: goto tr1042;
		case 522: goto tr1043;
		case 571: goto tr1044;
	}
	goto st0;
tr1050:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st319;
st319:
	if ( ++p == pe )
		goto _test_eof319;
case 319:
#line 12962 "./zscan.c"
	if ( (*p) == 10 )
		goto st320;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st321;
	goto st317;
st320:
	if ( ++p == pe )
		goto _test_eof320;
case 320:
	switch( (*p) ) {
		case 10: goto tr1048;
		case 34: goto tr1049;
		case 92: goto tr1050;
	}
	goto tr1047;
st321:
	if ( ++p == pe )
		goto _test_eof321;
case 321:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st322;
	goto st0;
st322:
	if ( ++p == pe )
		goto _test_eof322;
case 322:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st317;
	goto st0;
tr1009:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st323;
st323:
	if ( ++p == pe )
		goto _test_eof323;
case 323:
#line 13003 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1052;
		case 32: goto tr1052;
		case 40: goto tr1053;
		case 41: goto tr1054;
		case 266: goto tr1055;
		case 315: goto tr1056;
		case 522: goto tr1057;
		case 571: goto tr1058;
	}
	goto st0;
tr1003:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 503 "./zscan.rl"
	{ z->uv_3 = z->uval; }
	goto st324;
st324:
	if ( ++p == pe )
		goto _test_eof324;
case 324:
#line 13039 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1011;
	if ( 384 <= _widec && _widec <= 639 )
		goto st324;
	goto st0;
tr991:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 502 "./zscan.rl"
	{ z->uv_2 = z->uval; }
	goto st325;
st325:
	if ( ++p == pe )
		goto _test_eof325;
case 325:
#line 13076 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr996;
	if ( 384 <= _widec && _widec <= 639 )
		goto st325;
	goto st0;
tr979:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st326;
st326:
	if ( ++p == pe )
		goto _test_eof326;
case 326:
#line 13113 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr984;
	if ( 384 <= _widec && _widec <= 639 )
		goto st326;
	goto st0;
st327:
	if ( ++p == pe )
		goto _test_eof327;
case 327:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr971;
	if ( 384 <= _widec && _widec <= 639 )
		goto st327;
	goto st0;
tr20:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st328;
st328:
	if ( ++p == pe )
		goto _test_eof328;
case 328:
#line 13178 "./zscan.c"
	switch( (*p) ) {
		case 88: goto st329;
		case 89: goto st348;
		case 120: goto st329;
		case 121: goto st348;
	}
	goto st0;
st329:
	if ( ++p == pe )
		goto _test_eof329;
case 329:
	switch( (*p) ) {
		case 84: goto st330;
		case 116: goto st330;
	}
	goto st0;
st330:
	if ( ++p == pe )
		goto _test_eof330;
case 330:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st331;
		case 32: goto st331;
		case 40: goto tr1063;
		case 41: goto tr1064;
		case 522: goto tr1065;
		case 571: goto st347;
	}
	goto st0;
tr1063:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st331;
tr1064:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st331;
tr1065:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st331;
st331:
	if ( ++p == pe )
		goto _test_eof331;
case 331:
#line 13238 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1068;
		case 32: goto tr1068;
		case 34: goto tr1069;
		case 40: goto tr1070;
		case 41: goto tr1071;
		case 92: goto tr1072;
		case 522: goto tr1073;
		case 571: goto tr1074;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1067;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1067;
	} else
		goto tr1067;
	goto st0;
tr1085:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st332;
tr1111:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st332;
tr1067:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st332;
tr1099:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st332;
st332:
	if ( ++p == pe )
		goto _test_eof332;
case 332:
#line 13296 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1076;
		case 32: goto tr1076;
		case 34: goto tr1077;
		case 40: goto tr1078;
		case 41: goto tr1079;
		case 92: goto st336;
		case 266: goto tr1081;
		case 315: goto tr1082;
		case 522: goto tr1083;
		case 571: goto tr1084;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st332;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st332;
	} else
		goto st332;
	goto st0;
tr1088:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st333;
tr1089:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st333;
tr1093:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st333;
tr1076:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st333;
tr1078:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st333;
tr1079:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st333;
tr1083:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st333;
tr1100:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st333;
tr1102:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st333;
tr1103:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st333;
tr1107:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st333;
tr1112:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st333;
tr1114:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st333;
tr1115:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st333;
tr1119:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st333;
st333:
	if ( ++p == pe )
		goto _test_eof333;
case 333:
#line 13422 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st333;
		case 32: goto st333;
		case 34: goto tr1087;
		case 40: goto tr1088;
		case 41: goto tr1089;
		case 92: goto tr1090;
		case 266: goto tr1091;
		case 315: goto tr1092;
		case 522: goto tr1093;
		case 571: goto st338;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1085;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1085;
	} else
		goto tr1085;
	goto st0;
tr1096:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st334;
tr1087:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st334;
tr1124:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st334;
tr1125:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st334;
tr1069:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st334;
tr1077:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st334;
tr1101:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st334;
tr1113:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st334;
st334:
	if ( ++p == pe )
		goto _test_eof334;
case 334:
#line 13506 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1096;
		case 34: goto st335;
		case 92: goto st341;
	}
	goto st334;
tr1126:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st335;
st335:
	if ( ++p == pe )
		goto _test_eof335;
case 335:
#line 13521 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1100;
		case 32: goto tr1100;
		case 34: goto tr1101;
		case 40: goto tr1102;
		case 41: goto tr1103;
		case 92: goto tr1104;
		case 266: goto tr1105;
		case 315: goto tr1106;
		case 522: goto tr1107;
		case 571: goto tr1108;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1099;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1099;
	} else
		goto tr1099;
	goto st0;
tr1090:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st336;
tr1116:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st336;
tr1072:
#line 487 "./zscan.rl"
	{ text_start(z); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st336;
tr1104:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st336;
st336:
	if ( ++p == pe )
		goto _test_eof336;
case 336:
#line 13581 "./zscan.c"
	if ( (*p) == 10 )
		goto st337;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st339;
	goto st332;
st337:
	if ( ++p == pe )
		goto _test_eof337;
case 337:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1112;
		case 32: goto tr1112;
		case 34: goto tr1113;
		case 40: goto tr1114;
		case 41: goto tr1115;
		case 92: goto tr1116;
		case 266: goto tr1117;
		case 315: goto tr1118;
		case 522: goto tr1119;
		case 571: goto tr1120;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1111;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1111;
	} else
		goto tr1111;
	goto st0;
tr1084:
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st338;
tr1108:
#line 489 "./zscan.rl"
	{ z->tstart++; text_add_tok(z, p - z->tstart - 1, true); }
	goto st338;
tr1120:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 488 "./zscan.rl"
	{ text_add_tok(z, p - z->tstart, true); }
	goto st338;
st338:
	if ( ++p == pe )
		goto _test_eof338;
case 338:
#line 13647 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1093;
	if ( 384 <= _widec && _widec <= 639 )
		goto st338;
	goto st0;
st339:
	if ( ++p == pe )
		goto _test_eof339;
case 339:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st340;
	goto st0;
st340:
	if ( ++p == pe )
		goto _test_eof340;
case 340:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st332;
	goto st0;
tr1127:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st341;
st341:
	if ( ++p == pe )
		goto _test_eof341;
case 341:
#line 13696 "./zscan.c"
	if ( (*p) == 10 )
		goto st342;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st343;
	goto st334;
st342:
	if ( ++p == pe )
		goto _test_eof342;
case 342:
	switch( (*p) ) {
		case 10: goto tr1125;
		case 34: goto tr1126;
		case 92: goto tr1127;
	}
	goto tr1124;
st343:
	if ( ++p == pe )
		goto _test_eof343;
case 343:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st344;
	goto st0;
st344:
	if ( ++p == pe )
		goto _test_eof344;
case 344:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st334;
	goto st0;
tr1130:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st345;
tr1068:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st345;
tr1070:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st345;
tr1071:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st345;
tr1073:
#line 546 "./zscan.rl"
	{z->lcount++;}
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st345;
st345:
	if ( ++p == pe )
		goto _test_eof345;
case 345:
#line 13759 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1068;
		case 32: goto tr1068;
		case 34: goto tr1069;
		case 40: goto tr1070;
		case 41: goto tr1071;
		case 92: goto tr1072;
		case 266: goto tr1091;
		case 315: goto tr1092;
		case 522: goto tr1073;
		case 571: goto tr1074;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1067;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1067;
	} else
		goto tr1067;
	goto st0;
tr1074:
#line 487 "./zscan.rl"
	{ text_start(z); }
	goto st346;
st346:
	if ( ++p == pe )
		goto _test_eof346;
case 346:
#line 13803 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1130;
	if ( 384 <= _widec && _widec <= 639 )
		goto st346;
	goto st0;
st347:
	if ( ++p == pe )
		goto _test_eof347;
case 347:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1065;
	if ( 384 <= _widec && _widec <= 639 )
		goto st347;
	goto st0;
st348:
	if ( ++p == pe )
		goto _test_eof348;
case 348:
	switch( (*p) ) {
		case 80: goto st349;
		case 112: goto st349;
	}
	goto st0;
st349:
	if ( ++p == pe )
		goto _test_eof349;
case 349:
	switch( (*p) ) {
		case 69: goto st350;
		case 101: goto st350;
	}
	goto st0;
st350:
	if ( ++p == pe )
		goto _test_eof350;
case 350:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1133;
	goto st0;
tr1133:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st351;
st351:
	if ( ++p == pe )
		goto _test_eof351;
case 351:
#line 13893 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1134;
		case 32: goto tr1134;
		case 40: goto tr1135;
		case 41: goto tr1136;
		case 522: goto tr1138;
		case 571: goto tr1139;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st351;
	goto st0;
tr1141:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st352;
tr1142:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st352;
tr1144:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st352;
tr1134:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st352;
tr1135:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st352;
tr1136:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st352;
tr1138:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st352;
st352:
	if ( ++p == pe )
		goto _test_eof352;
case 352:
#line 13965 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st352;
		case 32: goto st352;
		case 40: goto tr1141;
		case 41: goto tr1142;
		case 92: goto st353;
		case 522: goto tr1144;
		case 571: goto st362;
	}
	goto st0;
st353:
	if ( ++p == pe )
		goto _test_eof353;
case 353:
	if ( (*p) == 35 )
		goto st354;
	goto st0;
st354:
	if ( ++p == pe )
		goto _test_eof354;
case 354:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st355;
		case 32: goto st355;
		case 40: goto tr1148;
		case 41: goto tr1149;
		case 522: goto tr1150;
		case 571: goto st361;
	}
	goto st0;
tr1148:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st355;
tr1149:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st355;
tr1150:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st355;
st355:
	if ( ++p == pe )
		goto _test_eof355;
case 355:
#line 14040 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st355;
		case 32: goto st355;
		case 40: goto tr1148;
		case 41: goto tr1149;
		case 522: goto tr1150;
		case 571: goto st361;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1152;
	goto st0;
tr1152:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st356;
st356:
	if ( ++p == pe )
		goto _test_eof356;
case 356:
#line 14074 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1153;
		case 32: goto tr1153;
		case 40: goto tr1154;
		case 41: goto tr1155;
		case 266: goto tr1157;
		case 315: goto tr1158;
		case 522: goto tr1159;
		case 571: goto tr1160;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st356;
	goto st0;
tr1162:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st357;
tr1163:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st357;
tr1165:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st357;
tr1153:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
	goto st357;
tr1154:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st357;
tr1155:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st357;
tr1159:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st357;
tr1168:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
	goto st357;
tr1169:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st357;
tr1170:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st357;
tr1174:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st357;
st357:
	if ( ++p == pe )
		goto _test_eof357;
case 357:
#line 14186 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st357;
		case 32: goto st357;
		case 40: goto tr1162;
		case 41: goto tr1163;
		case 522: goto tr1165;
		case 571: goto st360;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1164;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1164;
	} else
		goto tr1164;
	goto st0;
tr1164:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st358;
tr1171:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st358;
st358:
	if ( ++p == pe )
		goto _test_eof358;
case 358:
#line 14236 "./zscan.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto st359;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto st359;
	} else
		goto st359;
	goto st0;
st359:
	if ( ++p == pe )
		goto _test_eof359;
case 359:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1168;
		case 32: goto tr1168;
		case 40: goto tr1169;
		case 41: goto tr1170;
		case 266: goto tr1172;
		case 315: goto tr1173;
		case 522: goto tr1174;
		case 571: goto tr1175;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1171;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1171;
	} else
		goto tr1171;
	goto st0;
tr1160:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 538 "./zscan.rl"
	{ z->rfc3597_data_len = z->uval; z->rfc3597_data_written = 0; z->rfc3597_data = malloc(z->uval); }
	goto st360;
tr1175:
#line 633 "./zscan.rl"
	{
        if(z->rfc3597_data_written == z->rfc3597_data_len)
           parse_error_noargs("RFC3597 generic RR: more rdata is present than the indicated length");
        z->rfc3597_data[z->rfc3597_data_written++] = hexbyte(z->tstart);
    }
	goto st360;
st360:
	if ( ++p == pe )
		goto _test_eof360;
case 360:
#line 14301 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1165;
	if ( 384 <= _widec && _widec <= 639 )
		goto st360;
	goto st0;
st361:
	if ( ++p == pe )
		goto _test_eof361;
case 361:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1150;
	if ( 384 <= _widec && _widec <= 639 )
		goto st361;
	goto st0;
tr1139:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 501 "./zscan.rl"
	{ z->uv_1 = z->uval; }
	goto st362;
st362:
	if ( ++p == pe )
		goto _test_eof362;
case 362:
#line 14368 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1144;
	if ( 384 <= _widec && _widec <= 639 )
		goto st362;
	goto st0;
tr1194:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st363;
tr1199:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st363;
st363:
	if ( ++p == pe )
		goto _test_eof363;
case 363:
#line 14409 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr252;
	if ( 384 <= _widec && _widec <= 639 )
		goto st363;
	goto st0;
tr29:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st364;
tr1180:
#line 498 "./zscan.rl"
	{ z->ttl  = z->uval; }
	goto st364;
st364:
	if ( ++p == pe )
		goto _test_eof364;
case 364:
#line 14450 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr42;
	if ( 384 <= _widec && _widec <= 639 )
		goto st364;
	goto st0;
tr27:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st365;
st365:
	if ( ++p == pe )
		goto _test_eof365;
case 365:
#line 14487 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1176;
		case 32: goto tr1176;
		case 40: goto tr1177;
		case 41: goto tr1178;
		case 522: goto tr1179;
		case 571: goto tr1180;
	}
	goto st0;
tr15:
#line 500 "./zscan.rl"
	{ z->ttl  = z->def_ttl; }
	goto st366;
st366:
	if ( ++p == pe )
		goto _test_eof366;
case 366:
#line 14519 "./zscan.c"
	switch( (*p) ) {
		case 78: goto st367;
		case 110: goto st367;
	}
	goto st0;
st367:
	if ( ++p == pe )
		goto _test_eof367;
case 367:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st368;
		case 32: goto st368;
		case 40: goto tr1183;
		case 41: goto tr1184;
		case 522: goto tr1185;
		case 571: goto st371;
	}
	goto st0;
tr1183:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st368;
tr1184:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st368;
tr1185:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st368;
st368:
	if ( ++p == pe )
		goto _test_eof368;
case 368:
#line 14568 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st368;
		case 32: goto st368;
		case 40: goto tr1183;
		case 41: goto tr1184;
		case 65: goto st5;
		case 67: goto st48;
		case 68: goto st67;
		case 77: goto st101;
		case 78: goto st120;
		case 80: goto st199;
		case 83: goto st216;
		case 84: goto st328;
		case 97: goto st5;
		case 99: goto st48;
		case 100: goto st67;
		case 109: goto st101;
		case 110: goto st120;
		case 112: goto st199;
		case 115: goto st216;
		case 116: goto st328;
		case 522: goto tr1185;
		case 571: goto st371;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1187;
	goto st0;
tr1187:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st369;
st369:
	if ( ++p == pe )
		goto _test_eof369;
case 369:
#line 14618 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1188;
		case 32: goto tr1188;
		case 40: goto tr1189;
		case 41: goto tr1190;
		case 68: goto tr1192;
		case 72: goto tr1192;
		case 77: goto tr1192;
		case 87: goto tr1192;
		case 100: goto tr1192;
		case 104: goto tr1192;
		case 109: goto tr1192;
		case 119: goto tr1192;
		case 522: goto tr1193;
		case 571: goto tr1194;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st369;
	goto st0;
tr1192:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st370;
st370:
	if ( ++p == pe )
		goto _test_eof370;
case 370:
#line 14662 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1195;
		case 32: goto tr1195;
		case 40: goto tr1196;
		case 41: goto tr1197;
		case 522: goto tr1198;
		case 571: goto tr1199;
	}
	goto st0;
st371:
	if ( ++p == pe )
		goto _test_eof371;
case 371:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1185;
	if ( 384 <= _widec && _widec <= 639 )
		goto st371;
	goto st0;
tr7:
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
	goto st372;
tr1208:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 469 "./zscan.rl"
	{ dname_set(z, z->lhs_dname, p - z->tstart, true); }
	goto st372;
tr1223:
#line 470 "./zscan.rl"
	{ z->tstart++; dname_set(z, z->lhs_dname, p - z->tstart - 1, true); }
	goto st372;
tr1472:
#line 468 "./zscan.rl"
	{ dname_copy(z->lhs_dname, z->origin); }
	goto st372;
st372:
	if ( ++p == pe )
		goto _test_eof372;
case 372:
#line 14738 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr21;
	if ( 384 <= _widec && _widec <= 639 )
		goto st372;
	goto st0;
tr1477:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st373;
tr1206:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st373;
st373:
	if ( ++p == pe )
		goto _test_eof373;
case 373:
#line 14777 "./zscan.c"
	if ( (*p) == 10 )
		goto st374;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st375;
	goto st1;
st374:
	if ( ++p == pe )
		goto _test_eof374;
case 374:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1203;
		case 32: goto tr1203;
		case 40: goto tr1204;
		case 41: goto tr1205;
		case 92: goto tr1206;
		case 522: goto tr1207;
		case 571: goto tr1208;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1202;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1202;
		} else if ( _widec >= 35 )
			goto tr1202;
	} else
		goto tr1202;
	goto st0;
st375:
	if ( ++p == pe )
		goto _test_eof375;
case 375:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st376;
	goto st0;
st376:
	if ( ++p == pe )
		goto _test_eof376;
case 376:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st1;
	goto st0;
tr1211:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st377;
tr1212:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st377;
tr1213:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st377;
st377:
	if ( ++p == pe )
		goto _test_eof377;
case 377:
#line 14855 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st377;
		case 32: goto st377;
		case 40: goto tr1211;
		case 41: goto tr1212;
		case 65: goto tr12;
		case 67: goto tr13;
		case 68: goto tr14;
		case 73: goto tr15;
		case 77: goto tr16;
		case 78: goto tr17;
		case 80: goto tr18;
		case 83: goto tr19;
		case 84: goto tr20;
		case 97: goto tr12;
		case 99: goto tr13;
		case 100: goto tr14;
		case 105: goto tr15;
		case 109: goto tr16;
		case 110: goto tr17;
		case 112: goto tr18;
		case 115: goto tr19;
		case 116: goto tr20;
		case 266: goto tr71;
		case 315: goto st15;
		case 522: goto tr1213;
		case 571: goto st378;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr11;
	goto st0;
st378:
	if ( ++p == pe )
		goto _test_eof378;
case 378:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1213;
	if ( 384 <= _widec && _widec <= 639 )
		goto st378;
	goto st0;
tr1216:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st379;
tr1474:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st379;
tr1226:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st379;
tr1227:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st379;
st379:
	if ( ++p == pe )
		goto _test_eof379;
case 379:
#line 14953 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1216;
		case 34: goto st380;
		case 92: goto st381;
	}
	goto st379;
tr1228:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st380;
st380:
	if ( ++p == pe )
		goto _test_eof380;
case 380:
#line 14968 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1219;
		case 32: goto tr1219;
		case 40: goto tr1220;
		case 41: goto tr1221;
		case 522: goto tr1222;
		case 571: goto tr1223;
	}
	goto st0;
tr1229:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st381;
st381:
	if ( ++p == pe )
		goto _test_eof381;
case 381:
#line 15000 "./zscan.c"
	if ( (*p) == 10 )
		goto st382;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st383;
	goto st379;
st382:
	if ( ++p == pe )
		goto _test_eof382;
case 382:
	switch( (*p) ) {
		case 10: goto tr1227;
		case 34: goto tr1228;
		case 92: goto tr1229;
	}
	goto tr1226;
st383:
	if ( ++p == pe )
		goto _test_eof383;
case 383:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st384;
	goto st0;
st384:
	if ( ++p == pe )
		goto _test_eof384;
case 384:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st379;
	goto st0;
st385:
	if ( ++p == pe )
		goto _test_eof385;
case 385:
	switch( (*p) ) {
		case 65: goto st386;
		case 73: goto st406;
		case 79: goto st440;
		case 84: goto st460;
		case 97: goto st386;
		case 105: goto st406;
		case 111: goto st440;
		case 116: goto st460;
	}
	goto st0;
st386:
	if ( ++p == pe )
		goto _test_eof386;
case 386:
	switch( (*p) ) {
		case 68: goto st387;
		case 100: goto st387;
	}
	goto st0;
st387:
	if ( ++p == pe )
		goto _test_eof387;
case 387:
	switch( (*p) ) {
		case 68: goto st388;
		case 100: goto st388;
	}
	goto st0;
st388:
	if ( ++p == pe )
		goto _test_eof388;
case 388:
	switch( (*p) ) {
		case 82: goto st389;
		case 114: goto st389;
	}
	goto st0;
st389:
	if ( ++p == pe )
		goto _test_eof389;
case 389:
	if ( (*p) == 95 )
		goto st390;
	goto st0;
st390:
	if ( ++p == pe )
		goto _test_eof390;
case 390:
	switch( (*p) ) {
		case 76: goto st391;
		case 108: goto st391;
	}
	goto st0;
st391:
	if ( ++p == pe )
		goto _test_eof391;
case 391:
	switch( (*p) ) {
		case 73: goto st392;
		case 105: goto st392;
	}
	goto st0;
st392:
	if ( ++p == pe )
		goto _test_eof392;
case 392:
	switch( (*p) ) {
		case 77: goto st393;
		case 109: goto st393;
	}
	goto st0;
st393:
	if ( ++p == pe )
		goto _test_eof393;
case 393:
	switch( (*p) ) {
		case 73: goto st394;
		case 105: goto st394;
	}
	goto st0;
st394:
	if ( ++p == pe )
		goto _test_eof394;
case 394:
	switch( (*p) ) {
		case 84: goto st395;
		case 116: goto st395;
	}
	goto st0;
st395:
	if ( ++p == pe )
		goto _test_eof395;
case 395:
	if ( (*p) == 95 )
		goto st396;
	goto st0;
st396:
	if ( ++p == pe )
		goto _test_eof396;
case 396:
	switch( (*p) ) {
		case 86: goto st397;
		case 118: goto st397;
	}
	goto st0;
st397:
	if ( ++p == pe )
		goto _test_eof397;
case 397:
	switch( (*p) ) {
		case 52: goto st398;
		case 54: goto st402;
	}
	goto st0;
st398:
	if ( ++p == pe )
		goto _test_eof398;
case 398:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st399;
		case 32: goto st399;
		case 40: goto tr1249;
		case 41: goto tr1250;
		case 522: goto tr1251;
		case 571: goto st401;
	}
	goto st0;
tr1249:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st399;
tr1250:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st399;
tr1251:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st399;
st399:
	if ( ++p == pe )
		goto _test_eof399;
case 399:
#line 15195 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st399;
		case 32: goto st399;
		case 40: goto tr1249;
		case 41: goto tr1250;
		case 522: goto tr1251;
		case 571: goto st401;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1253;
	goto st0;
tr1253:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st400;
st400:
	if ( ++p == pe )
		goto _test_eof400;
case 400:
#line 15229 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1254;
		case 32: goto tr1254;
		case 40: goto tr1255;
		case 41: goto tr1256;
		case 266: goto tr1258;
		case 315: goto tr1259;
		case 522: goto tr1260;
		case 571: goto tr1261;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st400;
	goto st0;
st401:
	if ( ++p == pe )
		goto _test_eof401;
case 401:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1251;
	if ( 384 <= _widec && _widec <= 639 )
		goto st401;
	goto st0;
st402:
	if ( ++p == pe )
		goto _test_eof402;
case 402:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st403;
		case 32: goto st403;
		case 40: goto tr1263;
		case 41: goto tr1264;
		case 522: goto tr1265;
		case 571: goto st405;
	}
	goto st0;
tr1263:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st403;
tr1264:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st403;
tr1265:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st403;
st403:
	if ( ++p == pe )
		goto _test_eof403;
case 403:
#line 15330 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st403;
		case 32: goto st403;
		case 40: goto tr1263;
		case 41: goto tr1264;
		case 522: goto tr1265;
		case 571: goto st405;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1267;
	goto st0;
tr1267:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st404;
st404:
	if ( ++p == pe )
		goto _test_eof404;
case 404:
#line 15364 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1268;
		case 32: goto tr1268;
		case 40: goto tr1269;
		case 41: goto tr1270;
		case 266: goto tr1272;
		case 315: goto tr1273;
		case 522: goto tr1274;
		case 571: goto tr1275;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st404;
	goto st0;
st405:
	if ( ++p == pe )
		goto _test_eof405;
case 405:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1265;
	if ( 384 <= _widec && _widec <= 639 )
		goto st405;
	goto st0;
st406:
	if ( ++p == pe )
		goto _test_eof406;
case 406:
	switch( (*p) ) {
		case 78: goto st407;
		case 110: goto st407;
	}
	goto st0;
st407:
	if ( ++p == pe )
		goto _test_eof407;
case 407:
	switch( (*p) ) {
		case 67: goto st408;
		case 99: goto st408;
	}
	goto st0;
st408:
	if ( ++p == pe )
		goto _test_eof408;
case 408:
	switch( (*p) ) {
		case 76: goto st409;
		case 108: goto st409;
	}
	goto st0;
st409:
	if ( ++p == pe )
		goto _test_eof409;
case 409:
	switch( (*p) ) {
		case 85: goto st410;
		case 117: goto st410;
	}
	goto st0;
st410:
	if ( ++p == pe )
		goto _test_eof410;
case 410:
	switch( (*p) ) {
		case 68: goto st411;
		case 100: goto st411;
	}
	goto st0;
st411:
	if ( ++p == pe )
		goto _test_eof411;
case 411:
	switch( (*p) ) {
		case 69: goto st412;
		case 101: goto st412;
	}
	goto st0;
st412:
	if ( ++p == pe )
		goto _test_eof412;
case 412:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st413;
		case 32: goto st413;
		case 40: goto tr1283;
		case 41: goto tr1284;
		case 522: goto tr1285;
		case 571: goto st439;
	}
	goto st0;
tr1283:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st413;
tr1284:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st413;
tr1285:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st413;
st413:
	if ( ++p == pe )
		goto _test_eof413;
case 413:
#line 15519 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st413;
		case 32: goto st413;
		case 34: goto tr1288;
		case 40: goto tr1283;
		case 41: goto tr1284;
		case 92: goto tr1289;
		case 522: goto tr1285;
		case 571: goto st439;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1287;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1287;
	} else
		goto tr1287;
	goto st0;
tr1287:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st414;
tr1356:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st414;
st414:
	if ( ++p == pe )
		goto _test_eof414;
case 414:
#line 15565 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1291;
		case 32: goto tr1291;
		case 40: goto tr1292;
		case 41: goto tr1293;
		case 92: goto st429;
		case 266: goto tr1295;
		case 315: goto tr1296;
		case 522: goto tr1297;
		case 571: goto tr1298;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st414;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st414;
		} else if ( _widec >= 35 )
			goto st414;
	} else
		goto st414;
	goto st0;
tr1302:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st415;
tr1303:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st415;
tr1306:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st415;
tr1291:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
	goto st415;
tr1292:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st415;
tr1293:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st415;
tr1297:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st415;
tr1357:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
	goto st415;
tr1358:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st415;
tr1359:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st415;
tr1363:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st415;
tr1370:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
	goto st415;
tr1371:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st415;
tr1372:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st415;
tr1375:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st415;
st415:
	if ( ++p == pe )
		goto _test_eof415;
case 415:
#line 15693 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st415;
		case 32: goto st415;
		case 34: goto tr1301;
		case 40: goto tr1302;
		case 41: goto tr1303;
		case 64: goto tr1304;
		case 92: goto tr1305;
		case 522: goto tr1306;
		case 571: goto st428;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1299;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1299;
		} else if ( _widec >= 37 )
			goto tr1299;
	} else
		goto tr1299;
	goto st0;
tr1299:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st416;
tr1319:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st416;
st416:
	if ( ++p == pe )
		goto _test_eof416;
case 416:
#line 15743 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1309;
		case 32: goto tr1309;
		case 40: goto tr1310;
		case 41: goto tr1311;
		case 92: goto st417;
		case 266: goto tr1313;
		case 315: goto tr1314;
		case 522: goto tr1315;
		case 571: goto tr1316;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st416;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st416;
		} else if ( _widec >= 35 )
			goto st416;
	} else
		goto st416;
	goto st0;
tr1305:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st417;
tr1323:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st417;
st417:
	if ( ++p == pe )
		goto _test_eof417;
case 417:
#line 15793 "./zscan.c"
	if ( (*p) == 10 )
		goto st418;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st419;
	goto st416;
st418:
	if ( ++p == pe )
		goto _test_eof418;
case 418:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1320;
		case 32: goto tr1320;
		case 40: goto tr1321;
		case 41: goto tr1322;
		case 92: goto tr1323;
		case 266: goto tr1324;
		case 315: goto tr1325;
		case 522: goto tr1326;
		case 571: goto tr1327;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1319;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1319;
		} else if ( _widec >= 35 )
			goto tr1319;
	} else
		goto tr1319;
	goto st0;
st419:
	if ( ++p == pe )
		goto _test_eof419;
case 419:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st420;
	goto st0;
st420:
	if ( ++p == pe )
		goto _test_eof420;
case 420:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st416;
	goto st0;
tr1330:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st421;
tr1301:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st421;
tr1342:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st421;
tr1343:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st421;
st421:
	if ( ++p == pe )
		goto _test_eof421;
case 421:
#line 15879 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1330;
		case 34: goto st422;
		case 92: goto st423;
	}
	goto st421;
tr1344:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st422;
st422:
	if ( ++p == pe )
		goto _test_eof422;
case 422:
#line 15894 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1333;
		case 32: goto tr1333;
		case 40: goto tr1334;
		case 41: goto tr1335;
		case 266: goto tr1336;
		case 315: goto tr1337;
		case 522: goto tr1338;
		case 571: goto tr1339;
	}
	goto st0;
tr1345:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st423;
st423:
	if ( ++p == pe )
		goto _test_eof423;
case 423:
#line 15928 "./zscan.c"
	if ( (*p) == 10 )
		goto st424;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st425;
	goto st421;
st424:
	if ( ++p == pe )
		goto _test_eof424;
case 424:
	switch( (*p) ) {
		case 10: goto tr1343;
		case 34: goto tr1344;
		case 92: goto tr1345;
	}
	goto tr1342;
st425:
	if ( ++p == pe )
		goto _test_eof425;
case 425:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st426;
	goto st0;
st426:
	if ( ++p == pe )
		goto _test_eof426;
case 426:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st421;
	goto st0;
tr1304:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st427;
st427:
	if ( ++p == pe )
		goto _test_eof427;
case 427:
#line 15969 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1347;
		case 32: goto tr1347;
		case 40: goto tr1348;
		case 41: goto tr1349;
		case 266: goto tr1350;
		case 315: goto tr1351;
		case 522: goto tr1352;
		case 571: goto tr1353;
	}
	goto st0;
tr1298:
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
	goto st428;
tr1364:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 482 "./zscan.rl"
	{ set_filename(z, p - z->tstart); }
	goto st428;
tr1376:
#line 483 "./zscan.rl"
	{ z->tstart++; set_filename(z, p - z->tstart - 1); }
	goto st428;
st428:
	if ( ++p == pe )
		goto _test_eof428;
case 428:
#line 16013 "./zscan.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1306;
	if ( 384 <= _widec && _widec <= 639 )
		goto st428;
	goto st0;
tr1289:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st429;
tr1360:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st429;
st429:
	if ( ++p == pe )
		goto _test_eof429;
case 429:
#line 16052 "./zscan.c"
	if ( (*p) == 10 )
		goto st430;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st431;
	goto st414;
st430:
	if ( ++p == pe )
		goto _test_eof430;
case 430:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1357;
		case 32: goto tr1357;
		case 40: goto tr1358;
		case 41: goto tr1359;
		case 92: goto tr1360;
		case 266: goto tr1361;
		case 315: goto tr1362;
		case 522: goto tr1363;
		case 571: goto tr1364;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1356;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1356;
		} else if ( _widec >= 35 )
			goto tr1356;
	} else
		goto tr1356;
	goto st0;
st431:
	if ( ++p == pe )
		goto _test_eof431;
case 431:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st432;
	goto st0;
st432:
	if ( ++p == pe )
		goto _test_eof432;
case 432:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st414;
	goto st0;
tr1367:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st433;
tr1288:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st433;
tr1379:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st433;
tr1380:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st433;
st433:
	if ( ++p == pe )
		goto _test_eof433;
case 433:
#line 16138 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1367;
		case 34: goto st434;
		case 92: goto st435;
	}
	goto st433;
tr1381:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st434;
st434:
	if ( ++p == pe )
		goto _test_eof434;
case 434:
#line 16153 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1370;
		case 32: goto tr1370;
		case 40: goto tr1371;
		case 41: goto tr1372;
		case 266: goto tr1373;
		case 315: goto tr1374;
		case 522: goto tr1375;
		case 571: goto tr1376;
	}
	goto st0;
tr1382:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st435;
st435:
	if ( ++p == pe )
		goto _test_eof435;
case 435:
#line 16187 "./zscan.c"
	if ( (*p) == 10 )
		goto st436;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st437;
	goto st433;
st436:
	if ( ++p == pe )
		goto _test_eof436;
case 436:
	switch( (*p) ) {
		case 10: goto tr1380;
		case 34: goto tr1381;
		case 92: goto tr1382;
	}
	goto tr1379;
st437:
	if ( ++p == pe )
		goto _test_eof437;
case 437:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st438;
	goto st0;
st438:
	if ( ++p == pe )
		goto _test_eof438;
case 438:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st433;
	goto st0;
st439:
	if ( ++p == pe )
		goto _test_eof439;
case 439:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1285;
	if ( 384 <= _widec && _widec <= 639 )
		goto st439;
	goto st0;
st440:
	if ( ++p == pe )
		goto _test_eof440;
case 440:
	switch( (*p) ) {
		case 82: goto st441;
		case 114: goto st441;
	}
	goto st0;
st441:
	if ( ++p == pe )
		goto _test_eof441;
case 441:
	switch( (*p) ) {
		case 73: goto st442;
		case 105: goto st442;
	}
	goto st0;
st442:
	if ( ++p == pe )
		goto _test_eof442;
case 442:
	switch( (*p) ) {
		case 71: goto st443;
		case 103: goto st443;
	}
	goto st0;
st443:
	if ( ++p == pe )
		goto _test_eof443;
case 443:
	switch( (*p) ) {
		case 73: goto st444;
		case 105: goto st444;
	}
	goto st0;
st444:
	if ( ++p == pe )
		goto _test_eof444;
case 444:
	switch( (*p) ) {
		case 78: goto st445;
		case 110: goto st445;
	}
	goto st0;
st445:
	if ( ++p == pe )
		goto _test_eof445;
case 445:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st446;
		case 32: goto st446;
		case 40: goto tr1390;
		case 41: goto tr1391;
		case 522: goto tr1392;
		case 571: goto st459;
	}
	goto st0;
tr1390:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st446;
tr1391:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st446;
tr1392:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st446;
st446:
	if ( ++p == pe )
		goto _test_eof446;
case 446:
#line 16338 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st446;
		case 32: goto st446;
		case 34: goto tr1395;
		case 40: goto tr1390;
		case 41: goto tr1391;
		case 64: goto tr1396;
		case 92: goto tr1397;
		case 522: goto tr1392;
		case 571: goto st459;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1394;
	} else if ( _widec > 35 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1394;
		} else if ( _widec >= 37 )
			goto tr1394;
	} else
		goto tr1394;
	goto st0;
tr1394:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st447;
tr1409:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st447;
st447:
	if ( ++p == pe )
		goto _test_eof447;
case 447:
#line 16388 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1399;
		case 32: goto tr1399;
		case 40: goto tr1400;
		case 41: goto tr1401;
		case 92: goto st448;
		case 266: goto tr1403;
		case 315: goto tr1404;
		case 522: goto tr1405;
		case 571: goto tr1406;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st447;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto st447;
		} else if ( _widec >= 35 )
			goto st447;
	} else
		goto st447;
	goto st0;
tr1397:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st448;
tr1413:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st448;
st448:
	if ( ++p == pe )
		goto _test_eof448;
case 448:
#line 16438 "./zscan.c"
	if ( (*p) == 10 )
		goto st449;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st450;
	goto st447;
st449:
	if ( ++p == pe )
		goto _test_eof449;
case 449:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1410;
		case 32: goto tr1410;
		case 40: goto tr1411;
		case 41: goto tr1412;
		case 92: goto tr1413;
		case 266: goto tr1414;
		case 315: goto tr1415;
		case 522: goto tr1416;
		case 571: goto tr1417;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1409;
	} else if ( _widec > 33 ) {
		if ( _widec > 58 ) {
			if ( 60 <= _widec )
				goto tr1409;
		} else if ( _widec >= 35 )
			goto tr1409;
	} else
		goto tr1409;
	goto st0;
st450:
	if ( ++p == pe )
		goto _test_eof450;
case 450:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st451;
	goto st0;
st451:
	if ( ++p == pe )
		goto _test_eof451;
case 451:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st447;
	goto st0;
tr1420:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st452;
tr1395:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st452;
tr1432:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st452;
tr1433:
#line 562 "./zscan.rl"
	{z->lcount++;}
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st452;
st452:
	if ( ++p == pe )
		goto _test_eof452;
case 452:
#line 16524 "./zscan.c"
	switch( (*p) ) {
		case 10: goto tr1420;
		case 34: goto st453;
		case 92: goto st454;
	}
	goto st452;
tr1434:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st453;
st453:
	if ( ++p == pe )
		goto _test_eof453;
case 453:
#line 16539 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1423;
		case 32: goto tr1423;
		case 40: goto tr1424;
		case 41: goto tr1425;
		case 266: goto tr1426;
		case 315: goto tr1427;
		case 522: goto tr1428;
		case 571: goto tr1429;
	}
	goto st0;
tr1435:
#line 562 "./zscan.rl"
	{z->lcount++;}
	goto st454;
st454:
	if ( ++p == pe )
		goto _test_eof454;
case 454:
#line 16573 "./zscan.c"
	if ( (*p) == 10 )
		goto st455;
	if ( (*p) > 49 ) {
		if ( 50 <= (*p) && (*p) <= 57 )
			goto st0;
	} else if ( (*p) >= 48 )
		goto st456;
	goto st452;
st455:
	if ( ++p == pe )
		goto _test_eof455;
case 455:
	switch( (*p) ) {
		case 10: goto tr1433;
		case 34: goto tr1434;
		case 92: goto tr1435;
	}
	goto tr1432;
st456:
	if ( ++p == pe )
		goto _test_eof456;
case 456:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st457;
	goto st0;
st457:
	if ( ++p == pe )
		goto _test_eof457;
case 457:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto st452;
	goto st0;
tr1396:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st458;
st458:
	if ( ++p == pe )
		goto _test_eof458;
case 458:
#line 16614 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1437;
		case 32: goto tr1437;
		case 40: goto tr1438;
		case 41: goto tr1439;
		case 266: goto tr1440;
		case 315: goto tr1441;
		case 522: goto tr1442;
		case 571: goto tr1443;
	}
	goto st0;
st459:
	if ( ++p == pe )
		goto _test_eof459;
case 459:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1392;
	if ( 384 <= _widec && _widec <= 639 )
		goto st459;
	goto st0;
st460:
	if ( ++p == pe )
		goto _test_eof460;
case 460:
	switch( (*p) ) {
		case 84: goto st461;
		case 116: goto st461;
	}
	goto st0;
st461:
	if ( ++p == pe )
		goto _test_eof461;
case 461:
	switch( (*p) ) {
		case 76: goto st462;
		case 108: goto st462;
	}
	goto st0;
st462:
	if ( ++p == pe )
		goto _test_eof462;
case 462:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st463;
		case 32: goto st463;
		case 40: goto tr1447;
		case 41: goto tr1448;
		case 522: goto tr1449;
		case 571: goto st466;
	}
	goto st0;
tr1447:
#line 541 "./zscan.rl"
	{ if(z->in_paren) parse_error_noargs("Parenthetical error: double-open"); z->in_paren = true; }
	goto st463;
tr1448:
#line 542 "./zscan.rl"
	{ if(!z->in_paren) parse_error_noargs("Parenthetical error: unneccesary close"); z->in_paren = false; }
	goto st463;
tr1449:
#line 546 "./zscan.rl"
	{z->lcount++;}
	goto st463;
st463:
	if ( ++p == pe )
		goto _test_eof463;
case 463:
#line 16731 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st463;
		case 32: goto st463;
		case 40: goto tr1447;
		case 41: goto tr1448;
		case 522: goto tr1449;
		case 571: goto st466;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1451;
	goto st0;
tr1451:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st464;
st464:
	if ( ++p == pe )
		goto _test_eof464;
case 464:
#line 16765 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1452;
		case 32: goto tr1452;
		case 40: goto tr1453;
		case 41: goto tr1454;
		case 68: goto tr1456;
		case 72: goto tr1456;
		case 77: goto tr1456;
		case 87: goto tr1456;
		case 100: goto tr1456;
		case 104: goto tr1456;
		case 109: goto tr1456;
		case 119: goto tr1456;
		case 266: goto tr1457;
		case 315: goto tr1458;
		case 522: goto tr1459;
		case 571: goto tr1460;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto st464;
	goto st0;
tr1456:
#line 495 "./zscan.rl"
	{ z->uval = atoi(z->tstart); z->tstart = NULL; }
#line 496 "./zscan.rl"
	{ mult_uval(z, (*p)); }
	goto st465;
st465:
	if ( ++p == pe )
		goto _test_eof465;
case 465:
#line 16811 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1461;
		case 32: goto tr1461;
		case 40: goto tr1462;
		case 41: goto tr1463;
		case 266: goto tr1464;
		case 315: goto tr1465;
		case 522: goto tr1466;
		case 571: goto tr1467;
	}
	goto st0;
st466:
	if ( ++p == pe )
		goto _test_eof466;
case 466:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	if ( _widec == 522 )
		goto tr1449;
	if ( 384 <= _widec && _widec <= 639 )
		goto st466;
	goto st0;
tr1476:
#line 466 "./zscan.rl"
	{ z->tstart = p; }
	goto st467;
st467:
	if ( ++p == pe )
		goto _test_eof467;
case 467:
#line 16875 "./zscan.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(128 + ((*p) - -128));
			if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
#line 543 "./zscan.rl"
 z->in_paren  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1468;
		case 32: goto tr1468;
		case 40: goto tr1469;
		case 41: goto tr1470;
		case 522: goto tr1471;
		case 571: goto tr1472;
	}
	goto st0;
	}
	_test_eof468: cs = 468; goto _test_eof; 
	_test_eof1: cs = 1; goto _test_eof; 
	_test_eof2: cs = 2; goto _test_eof; 
	_test_eof3: cs = 3; goto _test_eof; 
	_test_eof4: cs = 4; goto _test_eof; 
	_test_eof5: cs = 5; goto _test_eof; 
	_test_eof6: cs = 6; goto _test_eof; 
	_test_eof7: cs = 7; goto _test_eof; 
	_test_eof8: cs = 8; goto _test_eof; 
	_test_eof9: cs = 9; goto _test_eof; 
	_test_eof10: cs = 10; goto _test_eof; 
	_test_eof11: cs = 11; goto _test_eof; 
	_test_eof12: cs = 12; goto _test_eof; 
	_test_eof13: cs = 13; goto _test_eof; 
	_test_eof14: cs = 14; goto _test_eof; 
	_test_eof15: cs = 15; goto _test_eof; 
	_test_eof16: cs = 16; goto _test_eof; 
	_test_eof17: cs = 17; goto _test_eof; 
	_test_eof18: cs = 18; goto _test_eof; 
	_test_eof19: cs = 19; goto _test_eof; 
	_test_eof20: cs = 20; goto _test_eof; 
	_test_eof21: cs = 21; goto _test_eof; 
	_test_eof22: cs = 22; goto _test_eof; 
	_test_eof23: cs = 23; goto _test_eof; 
	_test_eof24: cs = 24; goto _test_eof; 
	_test_eof25: cs = 25; goto _test_eof; 
	_test_eof26: cs = 26; goto _test_eof; 
	_test_eof27: cs = 27; goto _test_eof; 
	_test_eof28: cs = 28; goto _test_eof; 
	_test_eof29: cs = 29; goto _test_eof; 
	_test_eof30: cs = 30; goto _test_eof; 
	_test_eof31: cs = 31; goto _test_eof; 
	_test_eof32: cs = 32; goto _test_eof; 
	_test_eof33: cs = 33; goto _test_eof; 
	_test_eof34: cs = 34; goto _test_eof; 
	_test_eof35: cs = 35; goto _test_eof; 
	_test_eof36: cs = 36; goto _test_eof; 
	_test_eof37: cs = 37; goto _test_eof; 
	_test_eof38: cs = 38; goto _test_eof; 
	_test_eof39: cs = 39; goto _test_eof; 
	_test_eof40: cs = 40; goto _test_eof; 
	_test_eof41: cs = 41; goto _test_eof; 
	_test_eof42: cs = 42; goto _test_eof; 
	_test_eof43: cs = 43; goto _test_eof; 
	_test_eof44: cs = 44; goto _test_eof; 
	_test_eof45: cs = 45; goto _test_eof; 
	_test_eof46: cs = 46; goto _test_eof; 
	_test_eof47: cs = 47; goto _test_eof; 
	_test_eof48: cs = 48; goto _test_eof; 
	_test_eof49: cs = 49; goto _test_eof; 
	_test_eof50: cs = 50; goto _test_eof; 
	_test_eof51: cs = 51; goto _test_eof; 
	_test_eof52: cs = 52; goto _test_eof; 
	_test_eof53: cs = 53; goto _test_eof; 
	_test_eof54: cs = 54; goto _test_eof; 
	_test_eof55: cs = 55; goto _test_eof; 
	_test_eof56: cs = 56; goto _test_eof; 
	_test_eof57: cs = 57; goto _test_eof; 
	_test_eof58: cs = 58; goto _test_eof; 
	_test_eof59: cs = 59; goto _test_eof; 
	_test_eof60: cs = 60; goto _test_eof; 
	_test_eof61: cs = 61; goto _test_eof; 
	_test_eof62: cs = 62; goto _test_eof; 
	_test_eof63: cs = 63; goto _test_eof; 
	_test_eof64: cs = 64; goto _test_eof; 
	_test_eof65: cs = 65; goto _test_eof; 
	_test_eof66: cs = 66; goto _test_eof; 
	_test_eof67: cs = 67; goto _test_eof; 
	_test_eof68: cs = 68; goto _test_eof; 
	_test_eof69: cs = 69; goto _test_eof; 
	_test_eof70: cs = 70; goto _test_eof; 
	_test_eof71: cs = 71; goto _test_eof; 
	_test_eof72: cs = 72; goto _test_eof; 
	_test_eof73: cs = 73; goto _test_eof; 
	_test_eof74: cs = 74; goto _test_eof; 
	_test_eof75: cs = 75; goto _test_eof; 
	_test_eof76: cs = 76; goto _test_eof; 
	_test_eof77: cs = 77; goto _test_eof; 
	_test_eof78: cs = 78; goto _test_eof; 
	_test_eof79: cs = 79; goto _test_eof; 
	_test_eof80: cs = 80; goto _test_eof; 
	_test_eof81: cs = 81; goto _test_eof; 
	_test_eof82: cs = 82; goto _test_eof; 
	_test_eof83: cs = 83; goto _test_eof; 
	_test_eof84: cs = 84; goto _test_eof; 
	_test_eof85: cs = 85; goto _test_eof; 
	_test_eof86: cs = 86; goto _test_eof; 
	_test_eof87: cs = 87; goto _test_eof; 
	_test_eof88: cs = 88; goto _test_eof; 
	_test_eof89: cs = 89; goto _test_eof; 
	_test_eof90: cs = 90; goto _test_eof; 
	_test_eof91: cs = 91; goto _test_eof; 
	_test_eof92: cs = 92; goto _test_eof; 
	_test_eof93: cs = 93; goto _test_eof; 
	_test_eof94: cs = 94; goto _test_eof; 
	_test_eof95: cs = 95; goto _test_eof; 
	_test_eof96: cs = 96; goto _test_eof; 
	_test_eof97: cs = 97; goto _test_eof; 
	_test_eof98: cs = 98; goto _test_eof; 
	_test_eof99: cs = 99; goto _test_eof; 
	_test_eof100: cs = 100; goto _test_eof; 
	_test_eof101: cs = 101; goto _test_eof; 
	_test_eof102: cs = 102; goto _test_eof; 
	_test_eof103: cs = 103; goto _test_eof; 
	_test_eof104: cs = 104; goto _test_eof; 
	_test_eof105: cs = 105; goto _test_eof; 
	_test_eof106: cs = 106; goto _test_eof; 
	_test_eof107: cs = 107; goto _test_eof; 
	_test_eof108: cs = 108; goto _test_eof; 
	_test_eof109: cs = 109; goto _test_eof; 
	_test_eof110: cs = 110; goto _test_eof; 
	_test_eof111: cs = 111; goto _test_eof; 
	_test_eof112: cs = 112; goto _test_eof; 
	_test_eof113: cs = 113; goto _test_eof; 
	_test_eof114: cs = 114; goto _test_eof; 
	_test_eof115: cs = 115; goto _test_eof; 
	_test_eof116: cs = 116; goto _test_eof; 
	_test_eof117: cs = 117; goto _test_eof; 
	_test_eof118: cs = 118; goto _test_eof; 
	_test_eof119: cs = 119; goto _test_eof; 
	_test_eof120: cs = 120; goto _test_eof; 
	_test_eof121: cs = 121; goto _test_eof; 
	_test_eof122: cs = 122; goto _test_eof; 
	_test_eof123: cs = 123; goto _test_eof; 
	_test_eof124: cs = 124; goto _test_eof; 
	_test_eof125: cs = 125; goto _test_eof; 
	_test_eof126: cs = 126; goto _test_eof; 
	_test_eof127: cs = 127; goto _test_eof; 
	_test_eof128: cs = 128; goto _test_eof; 
	_test_eof129: cs = 129; goto _test_eof; 
	_test_eof130: cs = 130; goto _test_eof; 
	_test_eof131: cs = 131; goto _test_eof; 
	_test_eof132: cs = 132; goto _test_eof; 
	_test_eof133: cs = 133; goto _test_eof; 
	_test_eof134: cs = 134; goto _test_eof; 
	_test_eof135: cs = 135; goto _test_eof; 
	_test_eof136: cs = 136; goto _test_eof; 
	_test_eof137: cs = 137; goto _test_eof; 
	_test_eof138: cs = 138; goto _test_eof; 
	_test_eof139: cs = 139; goto _test_eof; 
	_test_eof140: cs = 140; goto _test_eof; 
	_test_eof141: cs = 141; goto _test_eof; 
	_test_eof142: cs = 142; goto _test_eof; 
	_test_eof143: cs = 143; goto _test_eof; 
	_test_eof144: cs = 144; goto _test_eof; 
	_test_eof145: cs = 145; goto _test_eof; 
	_test_eof146: cs = 146; goto _test_eof; 
	_test_eof147: cs = 147; goto _test_eof; 
	_test_eof148: cs = 148; goto _test_eof; 
	_test_eof149: cs = 149; goto _test_eof; 
	_test_eof150: cs = 150; goto _test_eof; 
	_test_eof151: cs = 151; goto _test_eof; 
	_test_eof152: cs = 152; goto _test_eof; 
	_test_eof153: cs = 153; goto _test_eof; 
	_test_eof154: cs = 154; goto _test_eof; 
	_test_eof155: cs = 155; goto _test_eof; 
	_test_eof156: cs = 156; goto _test_eof; 
	_test_eof157: cs = 157; goto _test_eof; 
	_test_eof158: cs = 158; goto _test_eof; 
	_test_eof159: cs = 159; goto _test_eof; 
	_test_eof160: cs = 160; goto _test_eof; 
	_test_eof161: cs = 161; goto _test_eof; 
	_test_eof162: cs = 162; goto _test_eof; 
	_test_eof163: cs = 163; goto _test_eof; 
	_test_eof164: cs = 164; goto _test_eof; 
	_test_eof165: cs = 165; goto _test_eof; 
	_test_eof166: cs = 166; goto _test_eof; 
	_test_eof167: cs = 167; goto _test_eof; 
	_test_eof168: cs = 168; goto _test_eof; 
	_test_eof169: cs = 169; goto _test_eof; 
	_test_eof170: cs = 170; goto _test_eof; 
	_test_eof171: cs = 171; goto _test_eof; 
	_test_eof172: cs = 172; goto _test_eof; 
	_test_eof173: cs = 173; goto _test_eof; 
	_test_eof174: cs = 174; goto _test_eof; 
	_test_eof175: cs = 175; goto _test_eof; 
	_test_eof176: cs = 176; goto _test_eof; 
	_test_eof177: cs = 177; goto _test_eof; 
	_test_eof178: cs = 178; goto _test_eof; 
	_test_eof179: cs = 179; goto _test_eof; 
	_test_eof180: cs = 180; goto _test_eof; 
	_test_eof181: cs = 181; goto _test_eof; 
	_test_eof182: cs = 182; goto _test_eof; 
	_test_eof183: cs = 183; goto _test_eof; 
	_test_eof184: cs = 184; goto _test_eof; 
	_test_eof185: cs = 185; goto _test_eof; 
	_test_eof186: cs = 186; goto _test_eof; 
	_test_eof187: cs = 187; goto _test_eof; 
	_test_eof188: cs = 188; goto _test_eof; 
	_test_eof189: cs = 189; goto _test_eof; 
	_test_eof190: cs = 190; goto _test_eof; 
	_test_eof191: cs = 191; goto _test_eof; 
	_test_eof192: cs = 192; goto _test_eof; 
	_test_eof193: cs = 193; goto _test_eof; 
	_test_eof194: cs = 194; goto _test_eof; 
	_test_eof195: cs = 195; goto _test_eof; 
	_test_eof196: cs = 196; goto _test_eof; 
	_test_eof197: cs = 197; goto _test_eof; 
	_test_eof198: cs = 198; goto _test_eof; 
	_test_eof199: cs = 199; goto _test_eof; 
	_test_eof200: cs = 200; goto _test_eof; 
	_test_eof201: cs = 201; goto _test_eof; 
	_test_eof202: cs = 202; goto _test_eof; 
	_test_eof203: cs = 203; goto _test_eof; 
	_test_eof204: cs = 204; goto _test_eof; 
	_test_eof205: cs = 205; goto _test_eof; 
	_test_eof206: cs = 206; goto _test_eof; 
	_test_eof207: cs = 207; goto _test_eof; 
	_test_eof208: cs = 208; goto _test_eof; 
	_test_eof209: cs = 209; goto _test_eof; 
	_test_eof210: cs = 210; goto _test_eof; 
	_test_eof211: cs = 211; goto _test_eof; 
	_test_eof212: cs = 212; goto _test_eof; 
	_test_eof213: cs = 213; goto _test_eof; 
	_test_eof214: cs = 214; goto _test_eof; 
	_test_eof215: cs = 215; goto _test_eof; 
	_test_eof216: cs = 216; goto _test_eof; 
	_test_eof217: cs = 217; goto _test_eof; 
	_test_eof218: cs = 218; goto _test_eof; 
	_test_eof219: cs = 219; goto _test_eof; 
	_test_eof220: cs = 220; goto _test_eof; 
	_test_eof221: cs = 221; goto _test_eof; 
	_test_eof222: cs = 222; goto _test_eof; 
	_test_eof223: cs = 223; goto _test_eof; 
	_test_eof224: cs = 224; goto _test_eof; 
	_test_eof225: cs = 225; goto _test_eof; 
	_test_eof226: cs = 226; goto _test_eof; 
	_test_eof227: cs = 227; goto _test_eof; 
	_test_eof228: cs = 228; goto _test_eof; 
	_test_eof229: cs = 229; goto _test_eof; 
	_test_eof230: cs = 230; goto _test_eof; 
	_test_eof231: cs = 231; goto _test_eof; 
	_test_eof232: cs = 232; goto _test_eof; 
	_test_eof233: cs = 233; goto _test_eof; 
	_test_eof234: cs = 234; goto _test_eof; 
	_test_eof235: cs = 235; goto _test_eof; 
	_test_eof236: cs = 236; goto _test_eof; 
	_test_eof237: cs = 237; goto _test_eof; 
	_test_eof238: cs = 238; goto _test_eof; 
	_test_eof239: cs = 239; goto _test_eof; 
	_test_eof240: cs = 240; goto _test_eof; 
	_test_eof241: cs = 241; goto _test_eof; 
	_test_eof242: cs = 242; goto _test_eof; 
	_test_eof243: cs = 243; goto _test_eof; 
	_test_eof244: cs = 244; goto _test_eof; 
	_test_eof245: cs = 245; goto _test_eof; 
	_test_eof246: cs = 246; goto _test_eof; 
	_test_eof247: cs = 247; goto _test_eof; 
	_test_eof248: cs = 248; goto _test_eof; 
	_test_eof249: cs = 249; goto _test_eof; 
	_test_eof250: cs = 250; goto _test_eof; 
	_test_eof251: cs = 251; goto _test_eof; 
	_test_eof252: cs = 252; goto _test_eof; 
	_test_eof253: cs = 253; goto _test_eof; 
	_test_eof254: cs = 254; goto _test_eof; 
	_test_eof255: cs = 255; goto _test_eof; 
	_test_eof256: cs = 256; goto _test_eof; 
	_test_eof257: cs = 257; goto _test_eof; 
	_test_eof258: cs = 258; goto _test_eof; 
	_test_eof259: cs = 259; goto _test_eof; 
	_test_eof260: cs = 260; goto _test_eof; 
	_test_eof261: cs = 261; goto _test_eof; 
	_test_eof262: cs = 262; goto _test_eof; 
	_test_eof263: cs = 263; goto _test_eof; 
	_test_eof264: cs = 264; goto _test_eof; 
	_test_eof265: cs = 265; goto _test_eof; 
	_test_eof266: cs = 266; goto _test_eof; 
	_test_eof267: cs = 267; goto _test_eof; 
	_test_eof268: cs = 268; goto _test_eof; 
	_test_eof269: cs = 269; goto _test_eof; 
	_test_eof270: cs = 270; goto _test_eof; 
	_test_eof271: cs = 271; goto _test_eof; 
	_test_eof272: cs = 272; goto _test_eof; 
	_test_eof273: cs = 273; goto _test_eof; 
	_test_eof274: cs = 274; goto _test_eof; 
	_test_eof275: cs = 275; goto _test_eof; 
	_test_eof276: cs = 276; goto _test_eof; 
	_test_eof277: cs = 277; goto _test_eof; 
	_test_eof278: cs = 278; goto _test_eof; 
	_test_eof279: cs = 279; goto _test_eof; 
	_test_eof280: cs = 280; goto _test_eof; 
	_test_eof281: cs = 281; goto _test_eof; 
	_test_eof282: cs = 282; goto _test_eof; 
	_test_eof283: cs = 283; goto _test_eof; 
	_test_eof284: cs = 284; goto _test_eof; 
	_test_eof285: cs = 285; goto _test_eof; 
	_test_eof286: cs = 286; goto _test_eof; 
	_test_eof287: cs = 287; goto _test_eof; 
	_test_eof288: cs = 288; goto _test_eof; 
	_test_eof289: cs = 289; goto _test_eof; 
	_test_eof290: cs = 290; goto _test_eof; 
	_test_eof291: cs = 291; goto _test_eof; 
	_test_eof292: cs = 292; goto _test_eof; 
	_test_eof293: cs = 293; goto _test_eof; 
	_test_eof294: cs = 294; goto _test_eof; 
	_test_eof295: cs = 295; goto _test_eof; 
	_test_eof296: cs = 296; goto _test_eof; 
	_test_eof297: cs = 297; goto _test_eof; 
	_test_eof298: cs = 298; goto _test_eof; 
	_test_eof299: cs = 299; goto _test_eof; 
	_test_eof300: cs = 300; goto _test_eof; 
	_test_eof301: cs = 301; goto _test_eof; 
	_test_eof302: cs = 302; goto _test_eof; 
	_test_eof303: cs = 303; goto _test_eof; 
	_test_eof304: cs = 304; goto _test_eof; 
	_test_eof305: cs = 305; goto _test_eof; 
	_test_eof306: cs = 306; goto _test_eof; 
	_test_eof307: cs = 307; goto _test_eof; 
	_test_eof308: cs = 308; goto _test_eof; 
	_test_eof309: cs = 309; goto _test_eof; 
	_test_eof310: cs = 310; goto _test_eof; 
	_test_eof311: cs = 311; goto _test_eof; 
	_test_eof312: cs = 312; goto _test_eof; 
	_test_eof313: cs = 313; goto _test_eof; 
	_test_eof314: cs = 314; goto _test_eof; 
	_test_eof315: cs = 315; goto _test_eof; 
	_test_eof316: cs = 316; goto _test_eof; 
	_test_eof317: cs = 317; goto _test_eof; 
	_test_eof318: cs = 318; goto _test_eof; 
	_test_eof319: cs = 319; goto _test_eof; 
	_test_eof320: cs = 320; goto _test_eof; 
	_test_eof321: cs = 321; goto _test_eof; 
	_test_eof322: cs = 322; goto _test_eof; 
	_test_eof323: cs = 323; goto _test_eof; 
	_test_eof324: cs = 324; goto _test_eof; 
	_test_eof325: cs = 325; goto _test_eof; 
	_test_eof326: cs = 326; goto _test_eof; 
	_test_eof327: cs = 327; goto _test_eof; 
	_test_eof328: cs = 328; goto _test_eof; 
	_test_eof329: cs = 329; goto _test_eof; 
	_test_eof330: cs = 330; goto _test_eof; 
	_test_eof331: cs = 331; goto _test_eof; 
	_test_eof332: cs = 332; goto _test_eof; 
	_test_eof333: cs = 333; goto _test_eof; 
	_test_eof334: cs = 334; goto _test_eof; 
	_test_eof335: cs = 335; goto _test_eof; 
	_test_eof336: cs = 336; goto _test_eof; 
	_test_eof337: cs = 337; goto _test_eof; 
	_test_eof338: cs = 338; goto _test_eof; 
	_test_eof339: cs = 339; goto _test_eof; 
	_test_eof340: cs = 340; goto _test_eof; 
	_test_eof341: cs = 341; goto _test_eof; 
	_test_eof342: cs = 342; goto _test_eof; 
	_test_eof343: cs = 343; goto _test_eof; 
	_test_eof344: cs = 344; goto _test_eof; 
	_test_eof345: cs = 345; goto _test_eof; 
	_test_eof346: cs = 346; goto _test_eof; 
	_test_eof347: cs = 347; goto _test_eof; 
	_test_eof348: cs = 348; goto _test_eof; 
	_test_eof349: cs = 349; goto _test_eof; 
	_test_eof350: cs = 350; goto _test_eof; 
	_test_eof351: cs = 351; goto _test_eof; 
	_test_eof352: cs = 352; goto _test_eof; 
	_test_eof353: cs = 353; goto _test_eof; 
	_test_eof354: cs = 354; goto _test_eof; 
	_test_eof355: cs = 355; goto _test_eof; 
	_test_eof356: cs = 356; goto _test_eof; 
	_test_eof357: cs = 357; goto _test_eof; 
	_test_eof358: cs = 358; goto _test_eof; 
	_test_eof359: cs = 359; goto _test_eof; 
	_test_eof360: cs = 360; goto _test_eof; 
	_test_eof361: cs = 361; goto _test_eof; 
	_test_eof362: cs = 362; goto _test_eof; 
	_test_eof363: cs = 363; goto _test_eof; 
	_test_eof364: cs = 364; goto _test_eof; 
	_test_eof365: cs = 365; goto _test_eof; 
	_test_eof366: cs = 366; goto _test_eof; 
	_test_eof367: cs = 367; goto _test_eof; 
	_test_eof368: cs = 368; goto _test_eof; 
	_test_eof369: cs = 369; goto _test_eof; 
	_test_eof370: cs = 370; goto _test_eof; 
	_test_eof371: cs = 371; goto _test_eof; 
	_test_eof372: cs = 372; goto _test_eof; 
	_test_eof373: cs = 373; goto _test_eof; 
	_test_eof374: cs = 374; goto _test_eof; 
	_test_eof375: cs = 375; goto _test_eof; 
	_test_eof376: cs = 376; goto _test_eof; 
	_test_eof377: cs = 377; goto _test_eof; 
	_test_eof378: cs = 378; goto _test_eof; 
	_test_eof379: cs = 379; goto _test_eof; 
	_test_eof380: cs = 380; goto _test_eof; 
	_test_eof381: cs = 381; goto _test_eof; 
	_test_eof382: cs = 382; goto _test_eof; 
	_test_eof383: cs = 383; goto _test_eof; 
	_test_eof384: cs = 384; goto _test_eof; 
	_test_eof385: cs = 385; goto _test_eof; 
	_test_eof386: cs = 386; goto _test_eof; 
	_test_eof387: cs = 387; goto _test_eof; 
	_test_eof388: cs = 388; goto _test_eof; 
	_test_eof389: cs = 389; goto _test_eof; 
	_test_eof390: cs = 390; goto _test_eof; 
	_test_eof391: cs = 391; goto _test_eof; 
	_test_eof392: cs = 392; goto _test_eof; 
	_test_eof393: cs = 393; goto _test_eof; 
	_test_eof394: cs = 394; goto _test_eof; 
	_test_eof395: cs = 395; goto _test_eof; 
	_test_eof396: cs = 396; goto _test_eof; 
	_test_eof397: cs = 397; goto _test_eof; 
	_test_eof398: cs = 398; goto _test_eof; 
	_test_eof399: cs = 399; goto _test_eof; 
	_test_eof400: cs = 400; goto _test_eof; 
	_test_eof401: cs = 401; goto _test_eof; 
	_test_eof402: cs = 402; goto _test_eof; 
	_test_eof403: cs = 403; goto _test_eof; 
	_test_eof404: cs = 404; goto _test_eof; 
	_test_eof405: cs = 405; goto _test_eof; 
	_test_eof406: cs = 406; goto _test_eof; 
	_test_eof407: cs = 407; goto _test_eof; 
	_test_eof408: cs = 408; goto _test_eof; 
	_test_eof409: cs = 409; goto _test_eof; 
	_test_eof410: cs = 410; goto _test_eof; 
	_test_eof411: cs = 411; goto _test_eof; 
	_test_eof412: cs = 412; goto _test_eof; 
	_test_eof413: cs = 413; goto _test_eof; 
	_test_eof414: cs = 414; goto _test_eof; 
	_test_eof415: cs = 415; goto _test_eof; 
	_test_eof416: cs = 416; goto _test_eof; 
	_test_eof417: cs = 417; goto _test_eof; 
	_test_eof418: cs = 418; goto _test_eof; 
	_test_eof419: cs = 419; goto _test_eof; 
	_test_eof420: cs = 420; goto _test_eof; 
	_test_eof421: cs = 421; goto _test_eof; 
	_test_eof422: cs = 422; goto _test_eof; 
	_test_eof423: cs = 423; goto _test_eof; 
	_test_eof424: cs = 424; goto _test_eof; 
	_test_eof425: cs = 425; goto _test_eof; 
	_test_eof426: cs = 426; goto _test_eof; 
	_test_eof427: cs = 427; goto _test_eof; 
	_test_eof428: cs = 428; goto _test_eof; 
	_test_eof429: cs = 429; goto _test_eof; 
	_test_eof430: cs = 430; goto _test_eof; 
	_test_eof431: cs = 431; goto _test_eof; 
	_test_eof432: cs = 432; goto _test_eof; 
	_test_eof433: cs = 433; goto _test_eof; 
	_test_eof434: cs = 434; goto _test_eof; 
	_test_eof435: cs = 435; goto _test_eof; 
	_test_eof436: cs = 436; goto _test_eof; 
	_test_eof437: cs = 437; goto _test_eof; 
	_test_eof438: cs = 438; goto _test_eof; 
	_test_eof439: cs = 439; goto _test_eof; 
	_test_eof440: cs = 440; goto _test_eof; 
	_test_eof441: cs = 441; goto _test_eof; 
	_test_eof442: cs = 442; goto _test_eof; 
	_test_eof443: cs = 443; goto _test_eof; 
	_test_eof444: cs = 444; goto _test_eof; 
	_test_eof445: cs = 445; goto _test_eof; 
	_test_eof446: cs = 446; goto _test_eof; 
	_test_eof447: cs = 447; goto _test_eof; 
	_test_eof448: cs = 448; goto _test_eof; 
	_test_eof449: cs = 449; goto _test_eof; 
	_test_eof450: cs = 450; goto _test_eof; 
	_test_eof451: cs = 451; goto _test_eof; 
	_test_eof452: cs = 452; goto _test_eof; 
	_test_eof453: cs = 453; goto _test_eof; 
	_test_eof454: cs = 454; goto _test_eof; 
	_test_eof455: cs = 455; goto _test_eof; 
	_test_eof456: cs = 456; goto _test_eof; 
	_test_eof457: cs = 457; goto _test_eof; 
	_test_eof458: cs = 458; goto _test_eof; 
	_test_eof459: cs = 459; goto _test_eof; 
	_test_eof460: cs = 460; goto _test_eof; 
	_test_eof461: cs = 461; goto _test_eof; 
	_test_eof462: cs = 462; goto _test_eof; 
	_test_eof463: cs = 463; goto _test_eof; 
	_test_eof464: cs = 464; goto _test_eof; 
	_test_eof465: cs = 465; goto _test_eof; 
	_test_eof466: cs = 466; goto _test_eof; 
	_test_eof467: cs = 467; goto _test_eof; 

	_test_eof: {}
	_out: {}
	}

#line 756 "./zscan.rl"


        if(cs == zone_error)
            parse_error_noargs("unparseable");
    }

    free(buf);
}

void scan_zone(const zoneinfo_t* zone) {
    dmn_assert(zone);

    log_debug("Scanning zone '%s'", logf_dname(zone->dname));

    zscan_t* z = malloc(sizeof(zscan_t));
    zscan_init(z, zone->zones_dir, zone->dname, zone->dname, zone->file, zone->def_ttl, 0, 0, zone->n_subzones, zone->subzones);
    int fd = open(zone->file, O_RDONLY);
    if(fd < 0)
        log_fatal("Cannot open zone file '%s' for reading: %s", zone->file, logf_errno());
    scanner(z, fd);
    if(close(fd))
        log_fatal("Cannot close zone file '%s': %s", zone->file, logf_errno());
    free(z);
}
